/*
 * Copyright (C) 1999
 * CSIRO Mathematical and Information Sciences Australia
 * Spatial Information Systems Project
 *
 * http://www.cmis.csiro.au/svg
 */


/**
 * CircleElement.java
 *
 * $Id: CircleElement.java,v 1.7 2000/05/06 20:08:22 gjb Exp $
 *
 */

package org.csiro.svgv.reader;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import java.awt.geom.GeneralPath;
import java.awt.geom.Ellipse2D;
import org.csiro.svgv.graphic.*;
import java.awt.geom.AffineTransform;

import EDU.Washington.grad.gjb.cassowary.*;
/**
 * The CircleElement class describes an SVG circle element.
 */
public class CircleElement extends SvgStdGraphic implements Definable {

    private DescElement _desc = null;
    private TitleElement _title = null;

    // default attribute values
    private static final String defaultCx = "0";
    private static final String defaultCy = "0";

    // circle element attributes
    private String _cx = null;
    private String _cy = null;
    private String _r;

    /**
     * Constructs a new CircleElement. The radius is initialized to 0.
     */
    public CircleElement() {
        super();
        _r = "0";    // attribute r is required so initialise it to a default value
    }

    private SvgDocument m_svg_document;
    public CircleElement(SvgDocument svg_document) {
	this();
        m_svg_document = svg_document;
    }

    /**
     * Constructs and initializes a CircleElement with the specified radius.
     * @param radius The radius of the circle.
     */
    public CircleElement(String radius) {
        super();
        _r = radius;
    }

    /**
     * Constructs and initializes a CircleElement with the specified location and radius.
     * @param cx The X coordinate of the circle's center.
     * @param cy The Y coordinate of the circle's center.
     * @param radius The radius of the circle.
     */
    public CircleElement(String cx, String cy, String radius) {
        super();
        _cx = cx;
        _cy = cy;
        _r = radius;
    }

    /**
     * Constructs and initializes a CircleElement with the specified radius.
     * @param radius The radius of the circle.
     */
    public CircleElement(float radius) {
        super();
        _r = Float.toString(radius);
    }

    /**
     * Constructs and initializes a CircleElement with the specified location and radius.
     * @param cx The X coordinate of the circle's center.
     * @param cy The Y coordinate of the circle's center.
     * @param radius The radius of the circle.
     */
    public CircleElement(float cx, float cy, float radius) {
        super();
        _cx = Float.toString(cx);
        _cy = Float.toString(cy);
        _r = Float.toString(radius);
    }

    /**
     * Sets the desc element.
     * @param desc The value to use when setting the desc element.
     */
    public void setDesc(DescElement desc) {
        _desc = desc;
    }

    /**
     * Sets the title element.
     * @param title The value to use when setting the title element.
     */
    public void setTitle(TitleElement title) {
        _title = title;
    }

    /**
     * Returns the desc element.
     * @return The desc element.
     */
    public DescElement getDesc() {
        return _desc;
    }

    /**
     * Returns the title element.
     * @return The title element.
     */
    public TitleElement getTitle() {
        return _title;
    }

    /**
     * Returns the X coordinate of the center of the circle.
     * @return The X coordinate of the center of the circle.
     */
    public String getCenterX() {
        if (_cx == null) {
            return defaultCx;
        }       

	return m_svg_document.handlePossibleVariable(_cx);
    }

    /**
     * Returns the Y coordinate of the center of the circle.
     * @return The Y coordinate of the center of the circle.
     */
    public String getCenterY() {
        if (_cy == null) {
            return defaultCy;
        }

	return m_svg_document.handlePossibleVariable(_cy);
    }

    /**
     * Returns the radius of the circle.
     * @return The radius of the circle.
     */
    public String getRadius() {

	return m_svg_document.handlePossibleVariable(_r);
    }

    /**
     * Sets the X coordinate of the circle's center to the specified value.
     * @param cx The value to use to set the X coordinate.
     */
    public void setCenterX(String cx) {
        _cx = cx;
    }

    /**
     * Sets the Y coordinate of the circle's center to the specified value.
     * @param cy The value to use to set the Y coordinate.
     */
    public void setCenterY(String cy) {
        _cy = cy;
    }

    /**
     * Sets the radius of the circle to the specified value.
     * @param r The value to use to set the radius.
     */
    public void setRadius(String r) {
        if (r != null)
            _r = r;
        else
            _r = "0";
    }

    /**
     * Sets the X coordinate of the circle's center to the specified value.
     * @param cx The value to use to set the X coordinate.
     */
    public void setCenterX(float cx) {
        _cx = Float.toString(cx);
    }

    /**
     * Sets the Y coordinate of the circle's center to the specified value.
     * @param cy The value to use to set the Y coordinate.
     */
    public void setCenterY(float cy) {
        _cy = Float.toString(cy);
    }

    /**
     * Sets the radius of the circle to the specified value.
     * @param r The value to use to set the radius.
     */
    public void setRadius(float r) {
        _r = Float.toString(r);
    }

    /**
     * Sets the circle's center to the specified location.
     * @param cx The value to use to set the X coordinate.
     * @param cy The value to use to set the Y coordinate.
     */
    public void setCenter(float cx, float cy) {
         _cx = Float.toString(cx);
         _cy = Float.toString(cy);
    }

    /**
     * Returns the Graphic object that represents this circle.
     * @return The Graphic object that represents this circle.
     */
    public Graphic getGraphic() {
	System.out.println("In Circle::getGraphic()");
        setStyleProperties();

        Path graphic = new Path();
        if (getId() != null)
            graphic.setId(getId());

        GeneralPath generalPath = new GeneralPath();
        float cx = Float.valueOf(getCenterX()).floatValue();
        float cy = Float.valueOf(getCenterY()).floatValue();
        float r = Float.valueOf(getRadius()).floatValue();
        Ellipse2D circle = new Ellipse2D.Float(cx - r, cy - r, 2 * r, 2 * r);
        generalPath.append(circle, false);

        graphic.setGeneralPath(generalPath);

        // get transform matrix and apply to graphic
        AffineTransform transformMatrix = getTransformMatrix(getTransform());
        if (transformMatrix != null)
            graphic.transform(transformMatrix);

        setGraphicStyle(graphic);
        return graphic;
    }



    // XmlElement stuff

    /**
     * Returns the tag that identifies a circle element in an SVG/XML file.
     * @return The tag that identifies a circle element in an SVG/XML file.
     */
    public String getTag() {
        return "circle";
    }

    /**
     * Reads in CircleElement data from the given XML Element.
     * @param elem The XML Element that contains data to be read.
     */
    public void read(Element elem) {

         // read the standard graphic attributes
         readStandardGraphicAttributes(elem);

         // read cx attribute
         _cx = elem.getAttribute("cx");

		System.out.println("_cx = " + _cx);
         // read cy attribute
         _cy = elem.getAttribute("cy");

		System.out.println("_cy = " + _cy);
         // read r attribute
         _r = elem.getAttribute("r");

		System.out.println("_r = " + _r);
        // read each child
        NodeList childNodes = elem.getChildNodes();
        int numChildren = childNodes.getLength();
        for (int i = 0; i < numChildren; i++) {
            Node child = childNodes.item(i);
            if (child.getNodeType() == Node.ELEMENT_NODE) { // it is an element
                Element childElement = (Element)child;
                String childTag = childElement.getTagName();

                if (childTag.equals("desc")) {
                    _desc = new DescElement();
                    _desc.read(childElement);

                } else if (childTag.equals("title")) {
                    _title = new TitleElement();
                    _title.read(childElement);
                }
             }
         }

    }

    /**
     * Inserts circle data into the given Document at the specified position.
     * The circle data will become a child of the given parent node.
     * @param doc The Document to write to.
     * @param parent The parent node of the circle data to be written.
     */
    public void write(Document doc, Node parent) {

        // create new element for this circle
        Element circle = doc.createElement(getTag());

        // set the standard graphic attributes
        setStandardGraphicAttributes(circle);

        // set the attributes, only set ones that have defaults if they are not null
        if (_cx != null) {
            circle.setAttribute("cx", _cx);
        }
        if (_cy != null) {
            circle.setAttribute("cy", _cy);
        }
        circle.setAttribute("r", _r);

        if (_desc != null)
            _desc.write(doc, circle);
        if (_title != null)
            _title.write(doc, circle);

        // append new element to the parent node
        parent.appendChild(circle);
    }

}
