/*
 * Copyright (C) 1999
 * CSIRO Mathematical and Information Sciences Australia
 * Spatial Information Systems Project
 *
 * http://www.cmis.csiro.au/svg
 */


/**
 * DefsElement.java
 *
 * $Id: DefsElement.java,v 1.4 2000/05/04 23:24:21 jwnichls Exp $
 *
 */

package org.csiro.svgv.reader;


import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import java.util.Vector;

/**
 * The DefsElement class describes an SVG defs element.
 */
public class DefsElement implements XmlElement {

    // list of Definable children elements
    private Vector _children;

    // attributes
    private String _id = null;
    private String _xmlLang = null;
    private String _xmlSpace = null;
    private String _class = null;
    private String _style = null;

    private SvgDocument m_svg_document;

    /**
     * Constructs a new DefsElement with no children elements.
     */
    public DefsElement(SvgDocument svg_document) {
        _children = new Vector();
	m_svg_document = svg_document;
    }

    /**
     * Adds a Definable child element.
     * @param child The child to be added.
     */
    public void addChild(Definable child) {
        if (child != null)
            _children.addElement(child);
    }

    /**
     * Returns the child element at the specified index.
     * @param index The position of the child to be returned.
     * @return The child element at the specified index. Or null if
     * the specified index is invalid.
     */
    public XmlElement getChild(int index) {
        if (index >= 0 && index < _children.size())
            return (XmlElement)_children.elementAt(index);
        return null;
    }

    /**
     * Returns the child defined element with the specified id.
     * @param if The id of the defined element reuqired.
     * @return The definable child element with the specified id. Or null if
     * no child was found with the specified id.
     */
    public Definable getDefinedElement(String id) {
        int numChildren = _children.size();
        for (int i = 0; i < numChildren; i++) {
            Definable child = (Definable)_children.elementAt(i);
            String childId = child.getId();
            if (childId != null && childId.equals(id))
                return child;
        }
        return null;
    }



    // XmlElement stuff

    /**
     * Returns the tag that identifies a defs element in an SVG/XML file.
     * @return The tag that identifies a defs element in an SVG/XML file.
     */
    public String getTag() {
        return "defs";
    }

    /**
     * Reads in DefsElement data from the given XML Element.
     * @param elem The XML Element that contains data to be read.
     */
    public void read(Element elem) {

        // read attributes
        String att = elem.getAttribute("id");
        if (att.length() > 0)
            _id = att;

        att = elem.getAttribute("xml:lang");
        if (att.length() > 0)
            _xmlLang = att;

        att = elem.getAttribute("xml:space");
        if (att.length() > 0)
            _xmlSpace = att;

        att = elem.getAttribute("class");
        if (att.length() > 0)
            _class = att;

        att = elem.getAttribute("style");
        if (att.length() > 0)
            _style = att;

         // read each child
         NodeList childNodes = elem.getChildNodes();
         int numChildren = childNodes.getLength();
         for (int i = 0; i < numChildren; i++) {
             Node child = childNodes.item(i);
             if (child.getNodeType() == Node.ELEMENT_NODE) { // it is an element
                Element childElement = (Element)child;
                String childTag = childElement.getTagName();

                if (childTag.equals("script")) {
                    // <script> element not yet supported

                } else if (childTag.equals("style")) {
                    // <style> element not yet supported

                } else if (childTag.equals("symbol")) {
                    SymbolElement symbol = new SymbolElement(m_svg_document);
                    symbol.read(childElement);
                    addChild(symbol);

                } else if (childTag.equals("marker")) {
                    MarkerElement marker = new MarkerElement(m_svg_document);
                    marker.read(childElement);
                    addChild(marker);

                } else if (childTag.equals("clipPath")) {
                    // <clipPath> element not yet supported

                } else if (childTag.equals("mask")) {
                    // <mask> element not yet supported

                } else if (childTag.equals("linearGradient")) {
                    LinearGradientElement gradient = new LinearGradientElement();
                    gradient.read(childElement);
                    addChild(gradient);

                } else if (childTag.equals("radialGradient")) {
                    RadialGradientElement gradient = new RadialGradientElement();
                    gradient.read(childElement);
                    addChild(gradient);
           /*
                } else if (childTag.equals("pattern")) {
                    // <pattern> element not yet supported

                } else if (childTag.equals("filter")) {
                    // <filter> element not yet supported

                } else if (childTag.equals("cursor")) {
                   // <cursor> element not yet supported

                } else if (childTag.equals("font")) {
                    // <font> element not yet supported

                } else if (childTag.equals("animate")) {
                    // <animate> element not yet supported

                } else if (childTag.equals("animateMotion")) {
                    // <animateMotion> element not yet supported

                } else if (childTag.equals("animateTransform")) {
                    // <animateTransform> element not yet supported

                } else if (childTag.equals("animateColor")) {
                    // <animateColor> element not yet supported

                } else if (childTag.equals("animateFlipbook")) {
                    // <animateFlipbook> element not yet supported
             */
                } else if (childTag.equals("path")) {
                    PathElement path = new PathElement(m_svg_document);
                    path.read(childElement);
                    addChild(path);

                } else if (childTag.equals("text")) {
                    TextElement text = new TextElement(m_svg_document);
                    text.read(childElement);
                    addChild(text);

                } else if (childTag.equals("rect")) {
                    RectElement rect = new RectElement(m_svg_document);
                    rect.read(childElement);
                    addChild(rect);

                } else if (childTag.equals("circle")) {
                    CircleElement circle = new CircleElement(m_svg_document);
                    circle.read(childElement);
                    addChild(circle);

                } else if (childTag.equals("ellipse")) {
                    EllipseElement ellipse = new EllipseElement(m_svg_document);
                    ellipse.read(childElement);
                    addChild(ellipse);

                } else if (childTag.equals("line")) {
                    LineElement line = new LineElement(m_svg_document);
                    line.read(childElement);
                    addChild(line);

                } else if (childTag.equals("polyline")) {
                    PolylineElement polyline = new PolylineElement(m_svg_document);
                    polyline.read(childElement);
                    addChild(polyline);

                } else if (childTag.equals("polygon")) {
                    PolygonElement polygon = new PolygonElement(m_svg_document);
                    polygon.read(childElement);
                    addChild(polygon);

                } else if (childTag.equals("use")) {
                    UseElement use = new UseElement(m_svg_document);
                    use.read(childElement);
                    addChild(use);

                } else if (childTag.equals("image")) {
                    ImageElement image = new ImageElement(m_svg_document);
                    image.read(childElement);
                    addChild(image);

                } else if (childTag.equals("svg")) {
                    SvgElement svg = new SvgElement(m_svg_document);
                    svg.read(childElement);
                    addChild(svg);

                } else if (childTag.equals("g")) {
                    GElement g = new GElement(m_svg_document);
                    g.read(childElement);
                    addChild(g);

                } else if (childTag.equals("switch")) {
                    // <switch> element not yet supported

                } else if (childTag.equals("constraint")) {
					System.out.println("in DefsElement::read with childTag == constraint");

				}
            }
        }
    }

    /**
     * Inserts data into the given Document.
     * The data will become a child of the given parent element.
     *
     * @param doc The Document to write to.
     * @param parent Indicates the parent element of the data to be written.
     */
    public void write(Document doc, Node parent) {

        // create new element for this defs element
        Element defs = doc.createElement(getTag());

        // set attributes
        if (_id != null)
            defs.setAttribute("id", _id);
        if (_xmlLang != null)
            defs.setAttribute("xml:lang", _xmlLang);
        if (_xmlSpace != null)
            defs.setAttribute("xml:space", _xmlSpace);
        if (_class != null)
            defs.setAttribute("class", _class);
        if (_style != null)
            defs.setAttribute("style", _style);

         // write child elements
        int numChildren = _children.size();
        for (int i = 0; i < numChildren; i++) {
            XmlElement child = getChild(i);
            child.write(doc, defs);
        }

        // append new element to the parent element
        parent.appendChild(defs);
    }

}
