/*
 * Copyright (C) 1999
 * CSIRO Mathematical and Information Sciences Australia
 * Spatial Information Systems Project
 *
 * http://www.cmis.csiro.au/svg
 */


/**
 * EllipseElement.java
 *
 * $Id: EllipseElement.java,v 1.5 2000/05/04 23:58:04 jwnichls Exp $
 *
 */

package org.csiro.svgv.reader;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import java.awt.geom.GeneralPath;
import java.awt.geom.Ellipse2D;
import java.lang.Math;
import java.awt.geom.AffineTransform;
import org.csiro.svgv.graphic.*;

/**
 * The EllipseElement class describes an SVG ellipse element.
 */
public class EllipseElement extends SvgStdGraphic implements Definable {

    private DescElement _desc = null;
    private TitleElement _title = null;

    // default attribute values
    private static final String defaultCx = "0";
    private static final String defaultCy = "0";

    // ellipse element attributes
    private String _cx = null;
    private String _cy = null;
    private String _rx;
    private String _ry;

    /**
     * Constructs a new EllipseElement.
     * rx and ry attriubtes are initialized to 0.
     */
    public EllipseElement() {
        super();
        _rx = "0";       // rx and ry attributes are required
        _ry = "0";       // so initialize them to some default values
    }

    private SvgDocument m_svg_document;
    public EllipseElement(SvgDocument svg_document) {
	this();
	m_svg_document = svg_document;
    }

    /**
     * Constructs and initializes an EllipseElement with the specified major and
     * minor attributes.
     * @param rx The radius of the ellipse along the X axis.
     * @param ry The radius of the ellipse along the Y axis.
     */
    public EllipseElement(String rx, String ry) {
        super();
        _rx = rx;
        _ry = ry;
    }

    /**
     * Constructs and initializes an EllipseElement with the specified attributes.
     * @param cx The X coordinate of the circle's center.
     * @param cy The Y coordinate of the circle's center.
     * @param rx The radius of the ellipse along the X axis.
     * @param ry The radius of the ellipse along the Y axis.
     */
    public EllipseElement(String cx, String cy, String rx, String ry) {
        super();
        _cx = cx;
        _cy = cy;
        _rx = rx;
        _ry = ry;
    }

    /**
     * Constructs and initializes an EllipseElement with the specified rx and
     * ry attributes.
     * @param rx The radius of the ellipse along the X axis.
     * @param ry The radius of the ellipse along the Y axis.
     */
    public EllipseElement(float rx, float ry) {
        super();
        _rx = Float.toString(rx);
        _ry = Float.toString(ry);
    }

    /**
     * Constructs and initializes an EllipseElement with the specified attributes.
     * @param cx The X coordinate of the circle's center.
     * @param cy The Y coordinate of the circle's center.
     * @param rx The radius of the ellipse along the X axis.
     * @param ry The radius of the ellipse along the Y axis.
     */
    public EllipseElement(float cx, float cy, float rx, float ry) {
        super();
        _cx = Float.toString(cx);
        _cy = Float.toString(cy);
        _rx = Float.toString(rx);
        _ry = Float.toString(ry);
    }

    /**
     * Sets the desc element.
     * @param desc The value to use when setting the desc element.
     */
    public void setDesc(DescElement desc) {
        _desc = desc;
    }

    /**
     * Sets the title element.
     * @param title The value to use when setting the title element.
     */
    public void setTitle(TitleElement title) {
        _title = title;
    }

    /**
     * Returns the desc element.
     * @return The desc element.
     */
    public DescElement getDesc() {
        return _desc;
    }

    /**
     * Returns the title element.
     * @return The title element.
     */
    public TitleElement getTitle() {
        return _title;
    }


    /**
     * Returns the X coordinate of the center of the ellipse.
     * @return The X coordinate of the center of the ellipse.
     */
    public String getCenterX() {
        if (_cx == null) {
            return defaultCx;
        }        
	return m_svg_document.handlePossibleVariable(_cx);
    }

    /**
     * Returns the Y coordinate of the center of the ellipse.
     * @return The Y coordinate of the center of the ellipse.
     */
    public String getCenterY() {
        if (_cy == null) {
            return defaultCy;        }        
	return m_svg_document.handlePossibleVariable(_cy);
    }

    /**
     * Returns the rx attribute of the ellipse.
     * @return The rx attribute of the ellipse.
     */
    public String getRx() {
	return m_svg_document.handlePossibleVariable(_rx);
    }

    /**
     * Returns the ry attribute of the ellipse.
     * @return The ry attribute of the ellipse.
     */
    public String getRy() {		
	return m_svg_document.handlePossibleVariable(_ry);
    }

    /**
     * Sets the X coordinate of the ellipse's center to the specified value.
     * @param cx The value to use to set the X coordinate.
     */
    public void setCenterX(String cx) {
        _cx = cx;
    }

    /**
     * Sets the Y coordinate of the ellipse's center to the specified value.
     * @param cy The value to use to set the Y coordinate.
     */
    public void setCenterY(String cy) {
        _cy = cy;
    }

    /**
     * Sets the ellipse's rx attribute to the specified value.
     * @param major The value to use to set the rx attribute.
     */
    public void setRx(String rx) {
        if (rx != null)
            _rx = rx;
        else
            _rx = "0";
    }

    /**
     * Sets the ellipse's minor (height) attribute to the specified value.
     * @param minor The value to use to set the minor attribute.
     */
    public void setRy(String ry) {
        if (ry != null)
            _ry = ry;
        else
            _ry = "0";
    }

    /**
     * Sets the X coordinate of the ellipse's center to the specified value.
     * @param cx The value to use to set the X coordinate.
     */
    public void setCenterX(float cx) {
        _cx = Float.toString(cx);
    }

    /**
     * Sets the Y coordinate of the ellipse's center to the specified value.
     * @param cy The value to use to set the Y coordinate.
     */
    public void setCenterY(float cy) {
        _cy = Float.toString(cy);
    }

    /**
     * Sets the ellipse's rx attribute to the specified value.
     * @param major The value to use to set the rx attribute.
     */
    public void setRx(float rx) {
        _rx = Float.toString(rx);
    }

    /**
     * Sets the ellipse's ry attribute to the specified value.
     * @param minor The value to use to set the ry attribute.
     */
    public void setRy(float ry) {
        _ry = Float.toString(ry);
    }


    /**
     * Sets the ellipse's center to the specified location.
     * @param cx The value to use to set the X coordinate.
     * @param cy The value to use to set the Y coordinate.
     */
    public void setCenter(float cx, float cy) {
        _cx = Float.toString(cx);
        _cy = Float.toString(cy);
    }

    /**
     * Returns the Graphic object that represents this ellipse.
     * @return The Graphic object that represents this ellipse.
     */
    public Graphic getGraphic() {

		System.out.println("In EllipseElement::getGraphic()");
        setStyleProperties();

        Path graphic = new Path();
        if (getId() != null)
            graphic.setId(getId());

        GeneralPath generalPath = new GeneralPath();
        float cx = Float.valueOf(getCenterX()).floatValue();
        float cy = Float.valueOf(getCenterY()).floatValue();
        float rx = Float.valueOf(getRx()).floatValue();
        float ry = Float.valueOf(getRy()).floatValue();
        Ellipse2D ellipse = new Ellipse2D.Float(cx - rx, cy - ry, rx*2, ry*2);
        generalPath.append(ellipse, false);

        graphic.setGeneralPath(generalPath);

        // get transform matrix and apply to graphic
        AffineTransform transformMatrix = getTransformMatrix(getTransform());
        if (transformMatrix != null)
            graphic.transform(transformMatrix);

        setGraphicStyle(graphic);

        return graphic;
    }



    // XmlElement stuff

    /**
     * Returns the tag that identifies an ellipse element in an SVG/XML file.
     * @return The tag that identifies an ellipse element in an SVG/XML file.
     */
    public String getTag() {
        return "ellipse";
    }

    /**
     * Reads in EllipseElement data from the given XML Element.
     * @param elem The XML Element that contains data to be read.
     */
    public void read(Element elem) {

         // read the standard graphic attributes
         readStandardGraphicAttributes(elem);

         // read cx attribute
         _cx = elem.getAttribute("cx");

		System.out.println("_cx = " + _cx);
         // read cy attribute
         _cy = elem.getAttribute("cy");

		System.out.println("_cy = " + _cy);
         // read rx attribute
         _rx = elem.getAttribute("rx");

		System.out.println("_rx = " + _rx);
         // read ry attribute
         _ry = elem.getAttribute("ry");

		System.out.println("_ry = " + _ry);        // read each child
        NodeList childNodes = elem.getChildNodes();
        int numChildren = childNodes.getLength();
        for (int i = 0; i < numChildren; i++) {
            Node child = childNodes.item(i);
            if (child.getNodeType() == Node.ELEMENT_NODE) { // it is an element
                Element childElement = (Element)child;
                String childTag = childElement.getTagName();

                if (childTag.equals("desc")) {
                    _desc = new DescElement();
                    _desc.read(childElement);

                } else if (childTag.equals("title")) {
                    _title = new TitleElement();
                    _title.read(childElement);
                }
             }
         }

    }

    /**
     * Inserts ellipse data into the given Document at the specified position.
     * The ellipse data will become a child of the given parent node.
     *
     * @param doc The Document to write to.
     * @param parent The parent node of the circle data to be written.
     */
    public void write(Document doc, Node parent) {

        // create new element for this ellipse
        Element ellipse = doc.createElement(getTag());

        // set the standard graphic attributes
        setStandardGraphicAttributes(ellipse);

        // set the attributes, only set ones that have defaults if they are not null
        if (_cx != null) {
            ellipse.setAttribute("cx", _cx);
        }
        if (_cy != null) {
            ellipse.setAttribute("cy", _cy);
        }
        ellipse.setAttribute("rx", _rx);
        ellipse.setAttribute("ry", _ry);

        if (_desc != null)
            _desc.write(doc, ellipse);
        if (_title != null)
            _title.write(doc, ellipse);

        // append new element to the parent element
        parent.appendChild(ellipse);
    }

}
