/*
 * Copyright (C) 1999
 * CSIRO Mathematical and Information Sciences Australia
 * Spatial Information Systems Project
 *
 * http://www.cmis.csiro.au/svg
 */


/**
 * GElement.java
 *
 * $Id: GElement.java,v 1.5 2000/05/06 20:20:10 gjb Exp $
 *
 */

package org.csiro.svgv.reader;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import java.util.Vector;
import org.csiro.svgv.graphic.*;
import java.awt.geom.AffineTransform;

/**
 * The GElement class describes an SVG g element.
 */
public class GElement extends SvgStdGraphic implements Definable {

    // children elements
    private DefsElement _defs = null;
    private DescElement _desc = null;
    private TitleElement _title = null;
    private Vector _children;

    /**
     * Cosntructs a new GElement with no children.
     */
    public GElement() {
        super();
        _children = new Vector();
    }

    private SvgDocument m_svg_document;
    public GElement(SvgDocument svg_document) {
	this();
	m_svg_document = svg_document;
    }

    /**
     * Sets the defs element.
     * @param defs The value to use when setting the defs element.
     */
    public void setDefs(DefsElement defs) {
        _defs = defs;
    }

    /**
     * Sets the title element.
     * @param title The value to use when setting the title element.
     */
    public void setTitle(TitleElement title) {
        _title = title;
    }

    /**
     * Sets the desc element.
     * @param desc The value to use when setting the desc element.
     */
    public void setDesc(DescElement desc) {
        _desc = desc;
    }

    /**
     * Add a child graphic to this g element.
     * @param graphic The child graphic element to add.
     */
    public void addChild(SvgGraphic graphic) {
        if (graphic != null) {
            _children.addElement(graphic);
            graphic.setParent(this);
        }
    }

    /**
     * Returns the defs element.
     * @return The defs element.
     */
    public DefsElement getDefs() {
        return _defs;
    }

    /**
     * Returns the title element.
     * @return The title element.
     */
    public TitleElement getTitle() {
        return _title;
    }

    /**
     * Returns the desc element.
     * @return The desc element.
     */
    public DescElement getDesc() {
        return _desc;
    }

    /**
     * Returns the child graphic element at the specified index.
     * @param The index of the required child element.
     * @return The child graphic element at the specified index. Note: will be
     * null if the index is invalid.
     */
    public SvgGraphic getChild(int index) {
        if (index >= 0 && index < _children.size())
            return (SvgGraphic)_children.elementAt(index);
        return null;
    }


    /**
     * Returns the Graphic object that represents this g element.
     * @return The Graphic object that represents this g element.
     */
    public Graphic getGraphic() {

        setStyleProperties(m_svg_document);

        Group group = new Group();
        int numChildren = _children.size();
        for (int i = 0; i < numChildren; i++) {
            SvgGraphic child = getChild(i);
            Graphic graphic = child.getGraphic();
            if (graphic != null)
                group.addElement(graphic);
        }

        // get transform matrix and apply to graphic
        AffineTransform transformMatrix = getTransformMatrix(getTransform());
        if (transformMatrix != null)
            group.transform(transformMatrix);

        setGraphicStyle(group);

        return group;
    }



    // XmlElement stuff

    /**
     * Returns the tag that identifies a g element in an SVG/XML file.
     * @return The tag that identifies a g element in an SVG/XML file.
     */
    public String getTag() {
        return "g";
    }

    /**
     * Reads in GElement data from the given XML Element.
     * @param elem The XML Element that contains data to be read.
     */
    public void read(Element elem) {
      
        // read the standard graphic attributes
         readStandardGraphicAttributes(elem);

         // read each child
         NodeList childNodes = elem.getChildNodes();
         int numChildren = childNodes.getLength();
         for (int i = 0; i < numChildren; i++) {
             Node child = childNodes.item(i);
             if (child.getNodeType() == Node.ELEMENT_NODE) { // it is an element
                Element childElement = (Element)child;
                String childTag = childElement.getTagName();

                if (childTag.equals("defs")) {
                    _defs = new DefsElement(m_svg_document);
                    _defs.read(childElement);

                } else if (childTag.equals("desc")) {
                    _desc = new DescElement();
                    _desc.read(childElement);

                } else if (childTag.equals("title")) {
                    _title = new TitleElement();
                    _title.read(childElement);

                } else if (childTag.equals("path")) {
                    PathElement path = new PathElement(m_svg_document);
                    path.read(childElement);
                    addChild(path);

                } else if (childTag.equals("text")) {
                    TextElement text = new TextElement(m_svg_document);
                    text.read(childElement);
                    addChild(text);

                } else if (childTag.equals("rect")) {
                    RectElement rect = new RectElement(m_svg_document);
                    rect.read(childElement);
                    addChild(rect);

                } else if (childTag.equals("circle")) {
                    CircleElement circle = new CircleElement(m_svg_document);
                    circle.read(childElement);
                    addChild(circle);

                } else if (childTag.equals("ellipse")) {
                    EllipseElement ellipse = new EllipseElement(m_svg_document);
                    ellipse.read(childElement);
                    addChild(ellipse);

                } else if (childTag.equals("line")) {
                    LineElement line = new LineElement(m_svg_document);
                    line.read(childElement);
                    addChild(line);

                } else if (childTag.equals("polyline")) {
                    PolylineElement polyline = new PolylineElement(m_svg_document);
                    polyline.read(childElement);
                    addChild(polyline);

                } else if (childTag.equals("polygon")) {
                    PolygonElement polygon = new PolygonElement(m_svg_document);
                    polygon.read(childElement);
                    addChild(polygon);

                } else if (childTag.equals("use")) {
                    UseElement use = new UseElement(m_svg_document);
                    use.read(childElement);
                    addChild(use);

                } else if (childTag.equals("image")) {
                    ImageElement image = new ImageElement(m_svg_document);
                    image.read(childElement);
                    addChild(image);

                } else if (childTag.equals("svg")) {
                    SvgElement svg = new SvgElement(m_svg_document);
                    svg.read(childElement);
                    addChild(svg);

                } else if (childTag.equals("g")) {
                    GElement g = new GElement(m_svg_document);
                    g.read(childElement);
                    addChild(g);

                } else if (childTag.equals("switch")) {
                    // <switch> element not yet supported

                } else if (childTag.equals("a")) {
                    // <a> element not yet supported

                } else if (childTag.equals("animate")) {
                    // <animate> element not yet supported

                } else if (childTag.equals("animateTransform")) {
                    // <animateTransform> element not yet supported

                } else if (childTag.equals("animateColor")) {
                    // <animateColor> element not yet supported

                } else if (childTag.equals("constraint")) {
					System.out.println("in GElement::read with childTag == constraint");

				}
            }
         }
    }

    /**
     * Inserts data into the given Document.
     * The data will become a child of the given parent element.
     *
     * @param doc The Document to write to.
     * @param parent Indicates the parent element of the data to be written.
     */
    public void write(Document doc, Node parent) {

        // create new element for this g element
        Element g = doc.createElement(getTag());

        // set the standard graphic attributes
        setStandardGraphicAttributes(g);

        if (_defs != null)
            _defs.write(doc, g);
        if (_desc != null)
            _desc.write(doc, g);
        if (_title != null)
            _title.write(doc, g);

        // write children elements
        int numChildren = _children.size();
        for (int i = 0; i < numChildren; i++) {
            SvgGraphic child = getChild(i);
            child.write(doc, g);
        }

        // append new element to the parent
        parent.appendChild(g);

    }
}
