/*
 * Copyright (C) 1999
 * CSIRO Mathematical and Information Sciences Australia
 * Spatial Information Systems Project
 *
 * http://www.cmis.csiro.au/svg
 */


/**
 * GradientElement.java
 *
 * $Id: GradientElement.java,v 1.1.1.1 1999/12/11 18:51:28 gjb Exp $
 *
 */


package org.csiro.svgv.reader;

import java.util.Vector;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import java.awt.GradientPaint;
import java.awt.Paint;
import java.awt.geom.Rectangle2D;

/**
 * The GradientElement abstract class describes the common features of SVG
 * lineargradient and radialgradient elements.
 */
public abstract class GradientElement {

    protected static final String defaultGradientUnits = "userSpace";

    // list of children GradientStopElements
    protected Vector _stops;

    // gradient attributes common to both linear and radial gradients
    protected String _id;
    protected String _gradientUnits;
    protected String _gradientTransform;

    /**
     * Creates a new GradientElement with no gradient stops.
     */
    public GradientElement() {
        _stops = new Vector();
    }

    /**
     * Returns the id of this element.
     * @return The id of this element.
     */
    public String getId() {
        return _id;
    }

    /**
     * Sets the id of this element to the specified value.
     * @param id The value to use when setting the id.
     */
    public void setId(String id) {
        _id = id;
    }

    /**
     * Adds a gradient stop to this gradient.
     * @param stop The gradient stop to add.
     */
    public void addStop(StopElement stop) {
        if (stop != null)
            _stops.addElement(stop);
    }

    /**
     * Returns the gradient stop at the specified index.
     * @param index The index of the gradient stop required.
     * @return The gradient stop at the specified index. Note: will be null
     * if the index is invalid.
     */
    public StopElement getStop(int index) {
        if (index >= 0 && index < _stops.size())
            return (StopElement)_stops.elementAt(index);
        return null;
    }

    public String getGradientUnits() {
        if (_gradientUnits != null)
            return _gradientUnits;
        else
            return defaultGradientUnits;
    }

    public String getGradientTransform() {
        return _gradientTransform;
    }


    /**
     * Returns the GradientPaint object that represents this gradient element.
     * @param The shape that will be coloured in by the GradientPaint. This is
     * needed to determine the start and end points of the GradientPaint object.
     * @return The GradientPaint object that represents this gradient element.
     */
    public abstract Paint getPaint(Rectangle2D userBounds, Rectangle2D objectBounds);

    /**
     * Reads the standard gradient attributes and children gradient stops.
     * @param elem The XML element that contains the data to be read.
     */
    public void readStdGradientAttributes(Element elem) {

        String att;

        // read id attribute
        att = elem.getAttribute("id");
        if (att.length() > 0)
            _id = att;

        // read gradient units attribute
        _gradientUnits = elem.getAttribute("gradientUnits");

        // read gradient transform attribute
        att = elem.getAttribute("gradientTransform");
        if (att.length() > 0)
            _gradientTransform = att;

         // read each gradientstop
         NodeList childNodes = elem.getChildNodes();
         int numChildren = childNodes.getLength();
         for (int i = 0; i < numChildren; i++) {
             Node child = childNodes.item(i);
             if (child.getNodeType() == Node.ELEMENT_NODE) { // it is an element
                Element childElement = (Element)child;
                if (childElement.getTagName().equals("stop")) {
                    StopElement stop = new StopElement();
                    stop.read(childElement);
                    addStop(stop);
                }
            }
         }
    }

    /**
     * Inserts standard gradient attributes and gradient stops into the given
     * Document at the specified position.
     * @param doc The Document to write to.
     * @param elem The gradient XML element to write into.
     */
    public void writeStdGradientAttributes(Document doc, Element elem) {

        // set the standard gradient attributes
         if (_id != null)
            elem.setAttribute("id", _id);
         if (_gradientUnits != null)
            elem.setAttribute("gradientUnits", _gradientUnits);
         if (_gradientTransform != null)
            elem.setAttribute("gradientTransform", _gradientTransform);

         // write the children gradientStops
        int numStops = _stops.size();
        for (int i = 0; i < numStops; i++) {
            StopElement child = getStop(i);
            child.write(doc, elem);
        }
     }

}