/*
 * Copyright (C) 1999
 * CSIRO Mathematical and Information Sciences Australia
 * Spatial Information Systems Project
 *
 * http://www.cmis.csiro.au/svg
 */


/**
 * ImageElement.java
 *
 * $Id: ImageElement.java,v 1.2 2000/05/04 23:24:21 jwnichls Exp $
 *
 */

package org.csiro.svgv.reader;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.csiro.svgv.graphic.*;
import java.awt.Toolkit;
import java.awt.MediaTracker;
import java.awt.Canvas;
import java.awt.geom.AffineTransform;
import java.net.URL;
import java.net.MalformedURLException;

/**
 * The ImageElement class describes an SVG image element.
 */
public class ImageElement extends SvgStdGraphic implements Definable {

    private TitleElement _title = null;
    private DescElement _desc = null;

    // image attributes
    private String _x = null;
    private String _y = null;
    private String _width;
    private String _height;
    private String _href;

    /**
     * Constructs an ImageElement with the href attribute initialised to "".
     */
    public ImageElement() {
        super();
        _width = "0";     // width, height and href attributes are required so initialize it to a default
        _height = "0";
        _href = "";
    }

    private SvgDocument m_svg_document;
    public ImageElement(SvgDocument svg_document) {
	this();
	m_svg_document = svg_document;
    }

    /**
     * Constructs an ImageElement with the specified href attribute.
     * @param href The value to use when initialising the href attribute.
     */
    public ImageElement(String href, String width, String height) {
        super();
        setHref(href);
        setWidth(width);
        setHeight(height);
    }

    /**
     * Constructs and initialises an ImageElement with the specified href and
     * location attributes.
     * @param href The value to use when initialising the href attribute.
     * @param x The X coordinate of the image location.
     * @param y The Y coordinate of the image location.
     * @param width The width of the image.
     * @param y The height of the image.
     */
    public ImageElement(String href, String x, String y, String width, String height) {
        super();
        setHref(href);
        _x = x;
        _y = y;
        setWidth(width);
        setHeight(height);
    }

     /**
     * Constructs and initialises an ImageElement with the specified href and
     * location attributes.
     * @param href The value to use when initialising the href attribute.
     * @param x The X coordinate of the image location.
     * @param y The Y coordinate of the image location.
     * @param width The width of the image.
     * @param y The height of the image.
     */
    public ImageElement(String href, float x, float y, float width, float height) {
        super();
        _href = href;
        _x = Float.toString(x);
        _y = Float.toString(y);
        _width = Float.toString(width);
        _height = Float.toString(height);
    }

    /**
     * Sets the title element.
     * @param title The value to use when setting the title element.
     */
    public void setTitle(TitleElement title) {
        _title = title;
    }

    /**
     * Sets the desc element.
     * @param desc The value to use when setting the desc element.
     */
    public void setDesc(DescElement desc) {
        _desc = desc;
    }

    /**
     * Sets the href attribute.
     * @param href The value to use when setting the href attribute.
     */
    public void setHref(String href) {
        if (href != null)
            _href = href;
        else
            _href = "";
    }

    /**
     * Sets the X coordinate of the image's location to the specified value.
     * @param x The value to use when setting the X coordinate.
     */
    public void setX(String x) {
        _x = x;
    }

    /**
     * Sets the Y coordinate of the image's location to the specified value.
     * @param y The value to use when setting the Y coordinate.
     */
    public void setY(String y) {
        _y = y;
    }

    /**
     * Sets the X coordinate of the image's location to the specified value.
     * @param x The value to use when setting the X coordinate.
     */
    public void setX(float x) {
        _x = Float.toString(x);
    }

    /**
     * Sets the Y coordinate of the image's location to the specified value.
     * @param y The value to use when setting the Y coordinate.
     */
    public void setY(float y) {
        _y = Float.toString(y);
    }


    /**
     * Sets the width of the image to the specified value.
     * @param width The value to use when setting the WIDTH.
     */
    public void setWidth(String width) {
        if (width == null)
            _width = "0";
        else
            _width = width;
    }

    /**
     * Sets the height of the image to the specified value.
     * @param y The value to use when setting the height.
     */
    public void setHeight(String height) {
        if (height == null)
            _height = "0";
        else
            _height = height;
    }

    /**
     * Sets the width of the imageto the specified value.
     * @param width The value to use when setting the width.
     */
    public void setWidth(float width) {
        _width = Float.toString(width);
    }

    /**
     * Sets the height of the image to the specified value.
     * @param y The value to use when setting the height.
     */
    public void setHeight(float height) {
        _height = Float.toString(height);
    }

    /**
     * Sets the location of the image.
     * @param x The value to use when setting the X coordinate.
     * @param y The value to use when setting the Y coordinate.
     */
    public void setLocation(float x, float y) {
        _x = Float.toString(x);
        _y = Float.toString(y);
    }

    public String getX() {
        if (_x == null)
            return "0";
        return _x;
    }

    public String getY() {
        if (_y == null)
            return "0";
        return _y;
    }

    public String getWidth() {
        return _width;
    }

    public String getHeight() {
        return _height;
    }


    /**
     * Returns the Graphic object that represents this image.
     * Doesn't handle svg files yet.
     * @return The Graphic object that represents this image.
     */
    public Graphic getGraphic() {

        setStyleProperties();

        org.csiro.svgv.graphic.Image image = new org.csiro.svgv.graphic.Image();
        image.setX(Float.parseFloat(getX()));
        image.setY(Float.parseFloat(getY()));
        image.setWidth(Float.parseFloat(getWidth()));
        image.setHeight(Float.parseFloat(getHeight()));

        // if the href starts with something like a URL
        // we use the URL image constructor - rather than
        // the normal filename string based one

	// however the href might not look like a URL,
	// but it could be a relative one - look at the 
	// SVG file path as well

	String imageRef = new String(_href);

	// aboslute whopper of a conditional
	if (imageRef != null && !imageRef.startsWith("http://") &&
	    !imageRef.startsWith("ftp://") && SvgDocument.svgFilePath != null
	    && (SvgDocument.svgFilePath.startsWith("http://") ||
		SvgDocument.svgFilePath.startsWith("ftp://"))) {
	  imageRef = SvgDocument.svgFilePath + imageRef;
	}

        if (imageRef != null &&
            (imageRef.startsWith("http://") ||
            imageRef.startsWith("ftp://"))) {
            try {
                URL url = new URL(imageRef);
                java.awt.Image im = Toolkit.getDefaultToolkit().getImage(url);
                image.setImage(im);
                if (im != null) {
                    MediaTracker tracker = new MediaTracker(new Canvas());
                    tracker.addImage(im,0);
                    try {
                        tracker.waitForAll();
                    } catch (InterruptedException e) {}
                }
            } catch (MalformedURLException e) {
                System.out.println("Bad URL in image element : " + imageRef);
                image.setImage(null);
            }

        } else {
            java.awt.Image im = Toolkit.getDefaultToolkit().getImage(SvgDocument.svgFilePath + imageRef);
            image.setImage(im);
            if (im != null) {
                MediaTracker tracker = new MediaTracker(new Canvas());
                tracker.addImage(im,0);
                try {
                    tracker.waitForAll();
                } catch (InterruptedException e) {}
            } else {
                System.out.println("Bad filename in image element : " + imageRef);
            }
        }

        // get transform matrix and apply to graphic
        AffineTransform transformMatrix = getTransformMatrix(getTransform());
        if (transformMatrix != null)
            image.transform(transformMatrix);

        setGraphicStyle(image);

        return image;
    }

    // XmlElement stuff

    /**
     * Returns the tag that identifies an image element in an SVG/XML file.
     * @return The tag that identifies an image element in an SVG/XML file.
     */
    public String getTag() {
        return "image";
    }

    /**
     * Reads in ImageElement data from the given XML Element.
     * @param elem The XML Element that contains data to be read.
     */
    public void read(Element elem) {

        // read the standard graphic attributes
        readStandardGraphicAttributes(elem);

        String att = elem.getAttribute("x");
        if (att.length() > 0)
            _x = att;

        att = elem.getAttribute("y");
        if (att.length() > 0)
            _y = att;

        _width = elem.getAttribute("width");
        _height = elem.getAttribute("height");
        _href = elem.getAttribute("xlink:href");

        // read each child
        NodeList childNodes = elem.getChildNodes();
        int numChildren = childNodes.getLength();
        for (int i = 0; i < numChildren; i++) {
            Node child = childNodes.item(i);
            if (child.getNodeType() == Node.ELEMENT_NODE) { // it is an element
                Element childElement = (Element)child;
                String childTag = childElement.getTagName();

                if (childTag.equals("title")) {
                    _title = new TitleElement();
                    _title.read(childElement);

                } else if (childTag.equals("desc")) {
                    _desc = new DescElement();
                    _desc.read(childElement);
                }
             }
         }
    }

    /**
     * Inserts data into the given Document.
     * The data will become a child of the given parent element.
     *
     * @param doc The Document to write to.
     * @param parent Indicates the parent element of the data to be written.
     */
    public void write(Document doc, Node parent) {

        // create new element for this image element
        Element image = doc.createElement(getTag());

        // set the standard graphic attributes
        setStandardGraphicAttributes(image);

        if (_x != null)
            image.setAttribute("x", _x);
        if (_y != null)
            image.setAttribute("y", _y);

        image.setAttribute("width", _width);
        image.setAttribute("height", _height);
        image.setAttribute("xlink:href", _href);

        if (_desc != null)
            _desc.write(doc, image);
        if (_title != null)
            _title.write(doc, image);

    }

}
