/*
 * Copyright (C) 1999
 * CSIRO Mathematical and Information Sciences Australia
 * Spatial Information Systems Project
 *
 * http://www.cmis.csiro.au/svg
 */


/**
 * LineElement.java
 *
 * $Id: LineElement.java,v 1.2 2000/04/29 06:04:56 will Exp $
 *
 */

package org.csiro.svgv.reader;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import java.awt.geom.GeneralPath;
import org.csiro.svgv.graphic.*;
import java.awt.geom.AffineTransform;

/**
 * The LineElement class describes an SVG line element.
 */
public class LineElement extends SvgStdGraphic implements Definable {

    private DescElement _desc = null;
    private TitleElement _title = null;

    // default attribute values
    private static final String defaultX1 = "0";
    private static final String defaultX2 = "0";
    private static final String defaultY1 = "0";
    private static final String defaultY2 = "0";

    // line attributes
    private String _x1 = null;
    private String _x2 = null;
    private String _y1 = null;
    private String _y2 = null;

	private SvgDocument m_svg_document;

    /**
     * Creates a new LineElement.
     */
    public LineElement(SvgDocument svg_document) {
        super();

	m_svg_document = svg_document;
    }

    /**
     * Creates a new LineElement.
     */
    public LineElement() {
        super();
    }

    /**
     * Creates and initialises a new LineElement with the specified endpoints.
     * @param x1 The X Coordinate of the line's first end point.
     * @param y1 The Y Coordinate of the line's first end point.
     * @param x2 The X Coordinate of the line's second end point.
     * @param y2 The Y Coordinate of the line's second end point.
     */
    public LineElement(String x1, String y1, String x2, String y2) {
        super();
        _x1 = x1;
        _x2 = x2;
        _y1 = y1;
        _y2 = y2;
    }

    /**
     * Creates and initialises a new LineElement with the specified endpoints.
     * @param x1 The X Coordinate of the line's first end point.
     * @param y1 The Y Coordinate of the line's first end point.
     * @param x2 The X Coordinate of the line's second end point.
     * @param y2 The Y Coordinate of the line's second end point.
     */
    public LineElement(float x1, float y1, float x2, float y2) {
        super();
        _x1 = Float.toString(x1);
        _x2 = Float.toString(x2);
        _y1 = Float.toString(y1);
        _y2 = Float.toString(y2);
    }

    /**
     * Sets the desc element.
     * @param desc The value to use when setting the desc element.
     */
    public void setDesc(DescElement desc) {
        _desc = desc;
    }

    /**
     * Sets the title element.
     * @param title The value to use when setting the title element.
     */
    public void setTitle(TitleElement title) {
        _title = title;
    }

    /**
     * Returns the desc element.
     * @return The desc element.
     */
    public DescElement getDesc() {
        return _desc;
    }

    /**
     * Returns the title element.
     * @return The title element.
     */
    public TitleElement getTitle() {
        return _title;
    }


    /**
     * Returns the value of the x1 attribute.
     * @return The value of the x1 attribute.
     */
    public String getX1() {
        if (_x1 == null) {
            return defaultX1;
        }
        return m_svg_document.handlePossibleVariable(_x1);
    }

    /**
     * Returns the value of the x2 attribute.
     * @return The value of the x2 attribute.
     */
    public String getX2() {
        if (_x2 == null) {
            return defaultX2;
        }
        return m_svg_document.handlePossibleVariable(_x2);
    }

    /**
     * Returns the value of the y1 attribute.
     * @return The value of the y1 attribute.
     */
    public String getY1() {
        if (_y1 == null) {
            return defaultY1;
        }
        return m_svg_document.handlePossibleVariable(_y1);
    }

    /**
     * Returns the value of the y2 attribute.
     * @return The value of the y2 attribute.
     */
    public String getY2() {
        if (_y2 == null) {
            return defaultY2;
        }
        return m_svg_document.handlePossibleVariable(_y2);
    }

    /**
     * Sets the X coordinate of the line's first end point.
     * @param x1 The value to use when setting the x1 attribute.
     */
    public void setX1(String x1) {
        _x1 = x1;
    }

    /**
     * Sets the X coordinate of the line's second end point.
     * @param x2 The value to use when setting the x2 attribute.
     */
    public void setX2(String x2) {
        _x2 = x2;
    }

    /**
     * Sets the Y coordinate of the line's first end point.
     * @param y1 The value to use when setting the y1 attribute.
     */
    public void setY1(String y1) {
        _y1 = y1;
    }

    /**
     * Sets the Y coordinate of the line's second end point.
     * @param y2 The value to use when setting the y2 attribute.
     */
    public void setY2(String y2) {
        _y2 = y2;
    }

    /**
     * Sets the X coordinate of the line's first end point.
     * @param x1 The value to use when setting the x1 attribute.
     */
    public void setX1(float x1) {
        _x1 = Float.toString(x1);
    }

    /**
     * Sets the X coordinate of the line's second end point.
     * @param x2 The value to use when setting the x2 attribute.
     */
    public void setX2(float x2) {
        _x2 = Float.toString(x2);
    }

    /**
     * Sets the Y coordinate of the line's first end point.
     * @param y1 The value to use when setting the y1 attribute.
     */
    public void setY1(float y1) {
        _y1 = Float.toString(y1);
    }

    /**
     * Sets the Y coordinate of the line's second end point.
     * @param y2 The value to use when setting the y2 attribute.
     */
    public void setY2(float y2) {
        _y2 = Float.toString(y2);
    }

    /**
     * Sets the coordinates of the line's end points.
     * @param x1 The value to use when setting the x1 attribute.
     * @param y1 The value to use when setting the y1 attribute.
     * @param x2 The value to use when setting the x2 attribute.
     * @param y2 The value to use when setting the y2 attribute.
     */
    public void setCoordinates(float x1, float y1, float x2, float y2) {
        _x1 = Float.toString(x1);
        _x2 = Float.toString(x2);
        _y1 = Float.toString(y1);
        _y2 = Float.toString(y2);
    }

    /**
     * Returns the Graphic object that represents this line.
     * @return The Graphic object that represents this line.
     */
    public Graphic getGraphic() {

        setStyleProperties();

        Path graphic = new Path();
        if (getId() != null)
            graphic.setId(getId());

        GeneralPath path = new GeneralPath();
        float x1 = Float.valueOf(getX1()).floatValue();
        float y1 = Float.valueOf(getY1()).floatValue();
        float x2 = Float.valueOf(getX2()).floatValue();
        float y2 = Float.valueOf(getY2()).floatValue();
        path.moveTo(x1,y1);
        path.lineTo(x2,y2);

        graphic.setGeneralPath(path);

        // get transform matrix and apply to graphic
        AffineTransform transformMatrix = getTransformMatrix(getTransform());
        if (transformMatrix != null)
            graphic.transform(transformMatrix);
            
        setGraphicStyle(graphic);

        return graphic;
    }


    
    // XmlElement stuff

    /**
     * Returns the tag that identifies a line element in an SVG/XML file.
     * @return The tag that identifies a line element in an SVG/XML file.
     */
    public String getTag() {
        return "line";
    }

    /**
     * Reads in LineElement data from the given XML Element.
     * @param elem The XML Element that contains data to be read.
     */
    public void read(Element elem) {

         // read the standard graphic attributes
         readStandardGraphicAttributes(elem);

         // read x1 attribute
         _x1 = elem.getAttribute("x1");

         // read x2 attribute
         _x2 = elem.getAttribute("x2");

         // read y1 attribute
         _y1 = elem.getAttribute("y1");

         // read y2 attribute
         _y2 = elem.getAttribute("y2");

          // read each child
        NodeList childNodes = elem.getChildNodes();
        int numChildren = childNodes.getLength();
        for (int i = 0; i < numChildren; i++) {
            Node child = childNodes.item(i);
            if (child.getNodeType() == Node.ELEMENT_NODE) { // it is an element
                Element childElement = (Element)child;
                String childTag = childElement.getTagName();

                if (childTag.equals("desc")) {
                    _desc = new DescElement();
                    _desc.read(childElement);

                } else if (childTag.equals("title")) {
                    _title = new TitleElement();
                    _title.read(childElement);
                }
             }
         }
    }

    /**
     * Inserts data into the given Document.
     * The data will become a child of the given parent element.
     *
     * @param doc The Document to write to.
     * @param parent Indicates the parent element of the data to be written.
     */
    public void write(Document doc, Node parent) {

        // create new element for this line
        Element line = doc.createElement(getTag());

        // set the standard graphic attributes
        setStandardGraphicAttributes(line);

        // set the attributes, only set ones that have defaults if they are not null
        if (_x1 != null)
            line.setAttribute("x1", _x1);
        if (_x2 != null)
            line.setAttribute("x2", _x2);
        if (_y1 != null)
            line.setAttribute("y1", _y1);
        if (_y2 != null)
            line.setAttribute("y2", _y2);

        if (_desc != null)
            _desc.write(doc, line);
        if (_title != null)
            _title.write(doc, line);

        // append new element to the parent element
        parent.appendChild(line);

    }




}
