/*
 * Copyright (C) 1999
 * CSIRO Mathematical and Information Sciences Australia
 * Spatial Information Systems Project
 *
 * http://www.cmis.csiro.au/svg
 */


/**
 * LinearGradientElement.java
 *
 * $Id: LinearGradientElement.java,v 1.1.1.1 1999/12/11 18:51:28 gjb Exp $
 *
 */

package org.csiro.svgv.reader;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import java.awt.GradientPaint;
import java.awt.Paint;
import java.awt.Shape;
import java.awt.geom.Rectangle2D;
import java.awt.geom.Point2D;
import java.awt.Color;
import java.lang.Math;
import java.awt.geom.AffineTransform;

/**
 * The LinearGradientElement class describes an SVG lineargradient element.
 */
public class LinearGradientElement extends GradientElement implements Definable, XmlElement{

    // default attribute values
    private static String defaultX1 = "0%";
    private static String defaultY1 = "0%";
    private static String defaultX2 = "100%";
    private static String defaultY2 = "100%";        // in degrees
    private static String defaultSpreadMethod = "pad";

    // linear gradient attributes
    private String _x1;
    private String _y1;
    private String _x2;
    private String _y2;
    private String _spreadMethod;

    /**
     * Constructs a LinearGradientElement with no gradient stops.
     */
    public LinearGradientElement() {
        super();
    }

    /**
     * Returns the value of the x1 attribute.
     * @return The value of the x1 attribute.
     */
    public String getX1() {
        if (_x1 != null)
            return _x1;
        else
            return defaultX1;
    }

    /**
     * Returns the value of the y1 attribute.
     * @return The value of the y1 attribute.
     */
    public String getY1() {
        if (_y1 != null)
            return _y1;
        else
            return defaultY1;
    }

     /**
     * Returns the value of the x2 attribute.
     * @return The value of the x2 attribute.
     */
    public String getX2() {
        if (_x2 != null)
            return _x2;
        else
            return defaultX2;
    }

    /**
     * Returns the value of the y2 attribute.
     * @return The value of the y2 attribute.
     */
    public String getY2() {
        if (_y2 != null)
            return _y2;
        else
            return defaultY2;
    }


    /**
     * Returns the value of the spread-method attribute.
     * @return The value of the spread-method attribute.
     */
    public String getSpreadMethod() {
        if (_spreadMethod != null)
            return _spreadMethod;
        else
            return defaultSpreadMethod;
    }


    /**
     * Returns the GradientPaint object that represents this lineargradient element.
     * @param userBounds The bounds of the current user space. This may be needed
     * to determine the start and end points of the GradientPaint object.
     * @param objectBounds The bounds of the shape that will be coloured in by the GradientPaint.
     * This may be needed to determine the start and end points of the GradientPaint object.
     * @return The GradientPaint object that represents this lineargradient element.
     */
    public Paint getPaint(Rectangle2D userBounds, Rectangle2D objectBounds) {

        //*** this only looks at the first two gradient stops because
        // the java GradientPaint class only handles 2 colors


        Color color1 = null;
        Color color2 = null;
        Point2D point1 = null;
        Point2D point2 = null;
        boolean cyclic = true;
        if (!getSpreadMethod().equals("repeat"))
            cyclic = false;

        StopElement stop1 = getStop(0);
        StopElement stop2 = getStop(1);
        if (stop1 != null && stop2 != null) {
             color1 = stop1.getColor();
             color2 = stop2.getColor();
        }

        float x1;
        float y1;
        float x2;
        float y2;
        float minX;
        float minY;
        float width;
        float height;

        if (getGradientUnits().equals("objectBoundingBox")) {
            minX = (float)objectBounds.getMinX();
            minY = (float)objectBounds.getMinY();
            width = (float)objectBounds.getWidth();
            height = (float)objectBounds.getHeight();
        } else { // userSpace
            minX = (float)userBounds.getMinX();
            minY = (float)userBounds.getMinY();
            width = (float)userBounds.getWidth();
            height = (float)userBounds.getHeight();
        }

        if (getX1().indexOf('%') != -1) // is a percentage
            x1 = minX + Float.valueOf(getX1().substring(0,getX1().length()-1)).floatValue() / 100 * width;
        else {
           if (getGradientUnits().equals("objectBoundingBox"))   // is a number between 0 and 1
               x1 = minX + Float.valueOf(getX1()).floatValue() * width;
           else // is a plain coordinate
               x1 = Float.valueOf(getX1()).floatValue();
        }

        if (getY1().indexOf('%') != -1) // is a percentage
            y1 = minY + Float.valueOf(getY1().substring(0,getY1().length()-1)).floatValue() / 100 * height;
        else {
           if (getGradientUnits().equals("objectBoundingBox"))  // is a number between 0 and 1
               y1 = minY + Float.valueOf(getY1()).floatValue() * height;
           else // is a plain coordinate
               y1 = Float.valueOf(getY1()).floatValue();
        }

        if (getX2().indexOf('%') != -1) // is a percentage
            x2 = minX + Float.valueOf(getX2().substring(0,getX2().length()-1)).floatValue() / 100 * width;
        else  {
            if (getGradientUnits().equals("objectBoundingBox"))   // is a number between 0 and 1
                x2 = minX + Float.valueOf(getX2()).floatValue() * width;
            else  // is a plain coordinate
                x2 = Float.valueOf(getX2()).floatValue();
        }

        if (getY2().indexOf('%') != -1) // is a percentage
            y2 = minY + Float.valueOf(getY2().substring(0,getY2().length()-1)).floatValue() / 100 * height;
        else {
            if (getGradientUnits().equals("objectBoundingBox"))  // is a number between 0 and 1
                y2 = minY + Float.valueOf(getY2()).floatValue() * height;
            else  // is a plain coordinate
                y2 = Float.valueOf(getY2()).floatValue();
        }

        point1 = new Point2D.Float(x1, y1);
        point2 = new Point2D.Float(x2, y2);

        // do transformation if there is one
        AffineTransform transformMatrix = SvgGraphic.getTransformMatrix(getGradientTransform());
        if (transformMatrix != null) {
            point1 = transformMatrix.transform(point1, null);
            point2 = transformMatrix.transform(point2, null);
        }

        if (color1 != null && color2 != null && point1 != null && point2 != null)
            return new GradientPaint(point1, color1, point2, color2, cyclic);
        return null;
    }

     // XmlElement stuff

    /**
     * Returns the tag that identifies a lineargradient element in an SVG/XML file.
     * @return The tag that identifies a lineargradient element in an SVG/XML file.
     */
    public String getTag() {
        return "linearGradient";
    }

    /**
     * Reads in LinearGradientElement data from the given XML Element.
     * @param elem The XML Element that contains data to be read.
     */
    public void read(Element elem) {

        readStdGradientAttributes(elem);

        String att = elem.getAttribute("x1");
        if (att.length() > 0)
            _x1 = att;

        att = elem.getAttribute("y1");
        if (att.length() > 0)
            _y1 = att;

        att = elem.getAttribute("x2");
        if (att.length() > 0)
            _x2 = att;

        att = elem.getAttribute("y2");
        if (att.length() > 0)
            _y2 = att;

        _spreadMethod  = elem.getAttribute("spreadMethod");
    }

    /**
     * Inserts data into the given Document.
     * The data will become a child of the given parent element.
     *
     * @param doc The Document to write to.
     * @param parent Indicates the parent element of the data to be written.
     */
    public void write(Document doc, Node parent) {

        // create new element for this lineargradient element
        Element linearGradient = doc.createElement(getTag());

        // write the standard gradient attributes and gradient stops
        writeStdGradientAttributes(doc, linearGradient);

         // set the attributes
        if (_x1 != null)
            linearGradient.setAttribute("x1", _x1);
        if (_y1 != null)
            linearGradient.setAttribute("y1", _y1);
        if (_x2 != null)
            linearGradient.setAttribute("x2", _x2);
        if (_y2 != null)
            linearGradient.setAttribute("y2", _y2);
        if (_spreadMethod != null)
            linearGradient.setAttribute("spreadMethod", _spreadMethod);

        // append new element to the parent element
        parent.appendChild(linearGradient);

    }

} 