/*
 * Copyright (C) 1999
 * CSIRO Mathematical and Information Sciences Australia
 * Spatial Information Systems Project
 *
 * http://www.cmis.csiro.au/svg
 */


/**
 * MarkerElement.java
 *
 * $Id: MarkerElement.java,v 1.4 2000/05/04 23:24:21 jwnichls Exp $
 *
 */

package org.csiro.svgv.reader;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.csiro.svgv.graphic.*;
import java.util.Vector;
/**
 * The MarkerElement class describes an SVG marker element.
 */
public class MarkerElement extends SvgGraphic implements Definable  {

     // children elements
    private DefsElement _defs = null;
    private DescElement _desc = null;
    private TitleElement _title = null;
    private Vector _children;

    // default attribute values
    private static final String defaultMarkerWidth = "3";
    private static final String defaultMarkerHeight = "3";
    private static final String defaultMarkerUnits = "stroke-width";
    private static final String defaultOrient = "0";

    // marker attributes
    private String _markerWidth = null;
    private String _markerHeight = null;
    private String _markerUnits = null;
    private String _orient = null;

    // symbol ref attributes
    private String _refX = null;
    private String _refY = null;

    private String _viewBox = null;
    private String _preserveAspectRatio = null;

    /**
     * Constructs a MarkerElement with no children graphic elements.
     */
    public MarkerElement() {
       _children = new Vector();
    }

    private SvgDocument m_svg_document;
    public MarkerElement(SvgDocument svg_document) {
	this();
	m_svg_document = svg_document;
    }

     /**
     * Sets the defs element of this marker.
     * @param defs The value to use when setting the defs element.
     */
    public void setDefs(DefsElement defs) {
        _defs = defs;
    }

    /**
     * Sets the desc element of this marker.
     * @param desc The value to use when setting the desc element.
     */
    public void setDesc(DescElement desc) {
        _desc = desc;
    }

    /**
     * Sets the title element of this marker.
     * @param title The value to use when setting the title element.
     */
    public void setTitle(TitleElement title) {
        _title = title;
    }



    /**
     * Sets the reference point to the specified location.
     * @param refX The X coordinate of the reference point.
     * @param refY The Y coordinate of the reference point.
     */
    public void setReferencePoint(float refX, float refY) {
        _refX = Float.toString(refX);
        _refY = Float.toString(refY);
    }

    /**
     * Add a child graphic to this marker.
     * @param graphic The child graphic element to add.
     */
    public void addChild(SvgGraphic graphic) {
        if (graphic != null) {
            _children.addElement(graphic);
            graphic.setParent(this);
        }
    }

    /**
     * Returns the child graphic element at the specified index.
     * @param The index of the required child element.
     * @return The child graphic element at the specified index. Note: will be
     * null if the index is invalid.
     */
    public SvgGraphic getChild(int index) {
        if (index >= 0 && index < _children.size())
            return (SvgGraphic)_children.elementAt(index);
        return null;
    }

    /**
     * Returns the defs element.
     * @return The defs element.
     */
    public DefsElement getDefs() {
        return _defs;
    }

    /**
     * Returns the title element.
     * @return The title element.
     */
    public TitleElement getTitle() {
        return _title;
    }

    /**
     * Returns the desc element.
     * @return The desc element.
     */
    public DescElement getDesc() {
        return _desc;
    }

    public String getMarkerWidth() {
        if (_markerWidth != null)
            return _markerWidth;
        return defaultMarkerWidth;
    }

    public String getMarkerHeight() {
        if (_markerHeight != null)
            return _markerHeight;
        return defaultMarkerHeight;
    }

    public String getMarkerUnits() {
        if (_markerUnits != null)
            return _markerUnits;
        return defaultMarkerUnits;
    }

     public String getOrient() {
        if (_orient != null)
            return _orient;
        return defaultOrient;
    }

    public String getRefX() {
        return _refX;
    }
    public String getRefY() {
        return _refY;
    }


    /**
     * Returns the Graphic object that represents this symbol element.
     * @return The Graphic object that represents this symbol element.
     */
    public Graphic getGraphic() {
        setStyleProperties();
        Group group = new Group();
        int numChildren = _children.size();
        for (int i = 0; i < numChildren; i++) {
            SvgGraphic child = getChild(i);
            Graphic graphic = child.getGraphic();
            if (graphic != null)
                group.addElement(graphic);
        }
        setGraphicStyle(group);
        return group;
     }


    // XmlElement stuff

    /**
     * Returns the tag that identifies a symbol element in an SVG/XML file.
     * @return The tag that identifies a symbol element in an SVG/XML file.
     */
    public String getTag() {
        return "marker";
    }

    /**
     * Reads in SymbolElement data from the given XML Element.
     * @param elem The XML Element that contains data to be read.
     */
    public void read(Element elem) {

        // read attributes
        String att = elem.getAttribute("id");
        if (att.length() > 0)
            _id = att;

        att = elem.getAttribute("xml:lang");
        if (att.length() > 0)
            _xmlLang = att;

        att = elem.getAttribute("xml:space");
        if (att.length() > 0)
            _xmlSpace = att;

        att = elem.getAttribute("class");
        if (att.length() > 0)
            _class = att;

        att = elem.getAttribute("style");
        if (att.length() > 0)
            _style = att;

        att = elem.getAttribute("refX");
        if (att.length() > 0)
            _refX = att;

        att = elem.getAttribute("refY");
        if (att.length() > 0)
            _refY = att;

        att = elem.getAttribute("viewBox");
        if (att.length() > 0)
            _viewBox = att;

        _preserveAspectRatio = elem.getAttribute("preserveAspectRatio");
        _markerUnits = elem.getAttribute("markerUnits");
        _markerWidth = elem.getAttribute("markerWidth");
        _markerHeight = elem.getAttribute("markerHeight");
        _orient = elem.getAttribute("orient");

         // read each child
         NodeList childNodes = elem.getChildNodes();
         int numChildren = childNodes.getLength();
         for (int i = 0; i < numChildren; i++) {
             Node child = childNodes.item(i);
             if (child.getNodeType() == Node.ELEMENT_NODE) { // it is an element
                Element childElement = (Element)child;
                String childTag = childElement.getTagName();

                if (childTag.equals("defs")) {
                    _defs = new DefsElement(m_svg_document);
                    _defs.read(childElement);

                } else if (childTag.equals("desc")) {
                    _desc = new DescElement();
                    _desc.read(childElement);

                } else if (childTag.equals("title")) {
                    _title = new TitleElement();
                    _title.read(childElement);

                } else if (childTag.equals("path")) {
                    PathElement path = new PathElement(m_svg_document);
                    path.read(childElement);
                    addChild(path);

                } else if (childTag.equals("text")) {
                    TextElement text = new TextElement(m_svg_document);
                    text.read(childElement);
                    addChild(text);

                } else if (childTag.equals("rect")) {
                    RectElement rect = new RectElement(m_svg_document);
                    rect.read(childElement);
                    addChild(rect);

                } else if (childTag.equals("circle")) {
                    CircleElement circle = new CircleElement(m_svg_document);
                    circle.read(childElement);
                    addChild(circle);

                } else if (childTag.equals("ellipse")) {
                    EllipseElement ellipse = new EllipseElement(m_svg_document);
                    ellipse.read(childElement);
                    addChild(ellipse);

                } else if (childTag.equals("line")) {
                    LineElement line = new LineElement(m_svg_document);
                    line.read(childElement);
                    addChild(line);

                } else if (childTag.equals("polyline")) {
                    PolylineElement polyline = new PolylineElement(m_svg_document);
                    polyline.read(childElement);
                    addChild(polyline);

                } else if (childTag.equals("polygon")) {
                    PolygonElement polygon = new PolygonElement(m_svg_document);
                    polygon.read(childElement);
                    addChild(polygon);

                } else if (childTag.equals("use")) {
                    UseElement use = new UseElement(m_svg_document);
                    use.read(childElement);
                    addChild(use);

                } else if (childTag.equals("image")) {
                    ImageElement image = new ImageElement(m_svg_document);
                    image.read(childElement);
                    addChild(image);

                } else if (childTag.equals("svg")) {
                    SvgElement svg = new SvgElement(m_svg_document);
                    svg.read(childElement);
                    addChild(svg);

                } else if (childTag.equals("g")) {
                    GElement g = new GElement(m_svg_document);
                    g.read(childElement);
                    addChild(g);

                } else if (childTag.equals("switch")) {
                    // <switch> element not yet supported

                } else if (childTag.equals("a")) {
                    // <a> element not yet supported

                } else if (childTag.equals("animate")) {
                    // <animate> element not yet supported

                } else if (childTag.equals("animateTransform")) {
                    // <animateTransform> element not yet supported

                } else if (childTag.equals("animateColor")) {
                    // <animateColor> element not yet supported

                } else if (childTag.equals("constraint")) {
					System.out.println("in MarkerElement::read with childTag == constraint");

				}
            }
         }
    }

    /**
     * Inserts data into the given Document.
     * The data will become a child of the given parent element.
     *
     * @param doc The Document to write to.
     * @param parent Indicates the parent element of the data to be written.
     */
    public void write(Document doc, Node parent) {

        // create new element for this marker element
        Element marker = doc.createElement(getTag());

         // set attributes
         if (_id != null)
            marker.setAttribute("id", _id);
        if (_xmlLang != null)
            marker.setAttribute("xml:lang", _xmlLang);
        if (_xmlSpace != null)
            marker.setAttribute("xml:space", _xmlSpace);
        if (_class != null)
            marker.setAttribute("class", _class);
        if (_style != null)
            marker.setAttribute("style", _style);
        if (_refX != null)
            marker.setAttribute("refX", _refX);
        if (_refY != null)
            marker.setAttribute("refY", _refY);
        if (_viewBox != null)
            marker.setAttribute("viewBox", _viewBox);
        if (_preserveAspectRatio != null)
            marker.setAttribute("preserveAspectRatio", _preserveAspectRatio);
        if (_markerWidth != null)
            marker.setAttribute("markerWidth", _markerWidth);
        if (_markerHeight != null)
            marker.setAttribute("markerHeight", _markerHeight);
        if (_markerUnits != null)
            marker.setAttribute("markerUnits", _markerUnits);
        if (_orient != null)
            marker.setAttribute("orient", _orient);
        

        // write defs, title and desc elements
        if (_defs != null)
            _defs.write(doc, marker);
        if (_desc != null)
            _desc.write(doc, marker);
        if (_title != null)
            _title.write(doc, marker);

        // write children elements
        int numChildren = _children.size();
        for (int i = 0; i < numChildren; i++) {
            SvgGraphic child = getChild(i);
            child.write(doc, marker);
        }

        // append new element to the parent
        parent.appendChild(marker);

    }




}
