/*
 * Copyright (C) 1999
 * CSIRO Mathematical and Information Sciences Australia
 * Spatial Information Systems Project
 *
 * http://www.cmis.csiro.au/svg
 */


/**
 * PathElement.java
 *
 * $Id: PathElement.java,v 1.5 2000/05/04 23:57:37 will Exp $
 *
 */

package org.csiro.svgv.reader;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import java.util.Vector;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;
import java.awt.geom.GeneralPath;
import java.awt.geom.Point2D;
import java.awt.geom.Arc2D;
import java.awt.geom.AffineTransform;
import java.lang.Math;
import org.csiro.svgv.graphic.*;

/**
 * The PathElement class describes an SVG path element.
 */
public class PathElement extends SvgStdGraphic implements Definable {

    private DescElement _desc = null;
    private TitleElement _title = null;

    // path attributes
    private String _d;
    private String _flatness = null;
    private String _nominalLength = null;

    /**
     * Constructs a new PathElement with no children data elements.
     * The d attribute is initialized to "".
     */
    public PathElement() {
        super();
        _d = "";    // the d attribute is required so initialize it with a default value
    }

    private SvgDocument m_svg_document;
    public PathElement(SvgDocument svg_document) {
	this();
	m_svg_document = svg_document;
    }

    /**
     * Constructs and initializes a new PathElement with no children
     * data elements.  The d attribute is initialized to the specified
     * value.
     * @param d The value to use when initializing the d attribute.
     */
    public PathElement(String d) {
        super();
        if (d != null)
            _d = d;
        else
            _d = "";
    }

    /**
     * Constructs and initializes a new PathElement with no children
     * data elements.  The d, flatness and nominal-length attributes
     * are initialized to the specified values.
     * @param d The value to use when initializing the d attribute.
     * @param flatness The value to use when initializing the flatness attribute.
     * @param nominalLength The value to use when initializing the nominal-length attribute.
     */
    public PathElement(String d, String flatness, String nominalLength) {
        super();
        if (d != null)
            _d = d;
        else
            _d = "";
        _flatness = flatness;
        _nominalLength = nominalLength;
    }

    /**
     * Sets the desc element.
     * @param desc The value to use when setting the desc element.
     */
    public void setDesc(DescElement desc) {
        _desc = desc;
    }

    /**
     * Sets the title element.
     * @param title The value to use when setting the title element.
     */
    public void setTitle(TitleElement title) {
        _title = title;
    }

    /**
     * Returns the desc element.
     * @return The desc element.
     */
    public DescElement getDesc() {
        return _desc;
    }

    /**
     * Returns the title element.
     * @return The title element.
     */
    public TitleElement getTitle() {
        return _title;
    }


    public String getD() {
        return _d;
    }
    public String getFlatness() {
        return _flatness;
    }
    public String getNominalLength() {
        return _nominalLength;
    }

    public void setD(String d) {
        if (d != null)
            _d = d;
        else
            _d = "";
    }
    public void setFlatness(String flatness) {
        _flatness = flatness;
    }
    public void setNominalLength(String nominalLength) {
        _nominalLength = nominalLength;
    }

    public void appendMoveTo(boolean absolute, float x, float y) {
        String pathCommand = "";
        if (absolute)
            pathCommand += "M ";
        else
            pathCommand += "m ";
        pathCommand += x + " " + y + " ";
        appendPathCommand(pathCommand);
    }

    public void appendLineTo(boolean absolute, float x, float y) {
        String pathCommand = "";
        if (absolute)
            pathCommand += "L ";
        else
            pathCommand += "l ";
        pathCommand += x + " " + y + " ";
        appendPathCommand(pathCommand);
    }

    public void appendCurveTo(boolean absolute, float x1, float y1,
                                float x2, float y2, float x, float y) {
        String pathCommand = "";
        if (absolute)
            pathCommand += "C ";
        else
            pathCommand += "c ";
        pathCommand += x1 + " " + y1 + " " + x2 + " " + y2 + " " + x + " " + y + " ";
        appendPathCommand(pathCommand);
    }

    public void appendClosePath() {
        appendPathCommand("z ");
    }

    public void appendSmoothCurveTo(boolean absolute, float x2, float y2,
                                    float x, float y) {
        String pathCommand = "";
        if (absolute)
            pathCommand += "S ";
        else
            pathCommand += "s ";
        pathCommand += x2 + " " + y2 + " " + x + " " + y + " ";
        appendPathCommand(pathCommand);
    }
    public void appendHorizontalLineTo(boolean absolute, float x) {
        String pathCommand = "";
        if (absolute)
            pathCommand += "H ";
        else
            pathCommand += "h ";
        pathCommand += x + " ";
        appendPathCommand(pathCommand);
    }
    public void appendVerticalLineTo(boolean absolute, float y) {
        String pathCommand = "";
        if (absolute)
            pathCommand += "V ";
        else
            pathCommand += "v ";
        pathCommand += y + " ";
        appendPathCommand(pathCommand);
    }
    public void appendEllipticalArcAngle(boolean absolute, float rx, float ry, float rotation,
                             boolean largeArcFlag, boolean sweepFlag, float x, float y) {
        String pathCommand = "";
        if (absolute)
            pathCommand += "A ";
        else
            pathCommand += "a ";
        pathCommand += rx + " " + ry + " " + rotation + " ";
        if (largeArcFlag)
            pathCommand += "1 ";
        else
            pathCommand += "0 ";
        if (sweepFlag)
            pathCommand += "1 ";
        else
            pathCommand += "0 ";
        pathCommand += x + " " + y + " ";
        appendPathCommand(pathCommand);
    }
    public void appendQuadraticBezierCurveTo(boolean absolute, float x1, float y1,
                                                float x, float y) {
        String pathCommand = "";
        if (absolute)
            pathCommand += "Q ";
        else
            pathCommand += "q ";
        pathCommand += x1 + " " + y1 + " " + x + " " + y + " ";
        appendPathCommand(pathCommand);
    }
    public void appendTruetypeQuadraticBezierCurveTo(boolean absolute,
                                                     CoordinateList points) {
        int numPoints = points.size();
        if (numPoints > 0) {
            String pathCommand = "";
            if (absolute)
                pathCommand += "T ";
            else
                pathCommand += "t ";
            for (int i = 0; i < numPoints; i++) {
                Coordinate c = points.coordinateAt(i);
                pathCommand += c.x + " " + c.y + " ";
            }
            appendPathCommand(pathCommand);
        }
    }

    // append command to appropriate d attribute
    private void appendPathCommand(String pathCommand) {
        int dLengthBefore = _d.length();
        int dLengthAfter = dLengthBefore + pathCommand.length();
        int nextNewLine = ((int)Math.floor(dLengthBefore / 255) + 1) * 255;
        if (dLengthAfter > nextNewLine)
            _d += System.getProperty("line.separator");
        _d += pathCommand;
    }


    public Graphic getGraphic() {

        setStyleProperties();

        Group group = new Group();  // group consists of path and a set of markers (if there are any)
        Path graphicPath = new Path();
        group.addElement(graphicPath);
        if (getId() != null)
            group.setId(getId());

        GeneralPath path = new GeneralPath();
        MarkerElement startMarker = getStartMarker();
        MarkerElement endMarker = getEndMarker();
        MarkerElement midMarker = getMidMarker();

        boolean firstPoint = true;

        // parse d attribute of Path element
        StringTokenizer st = new StringTokenizer(_d, "MmLlCcZzSsHhVvQqTtAa", true);
        while (st.hasMoreTokens()) {
            // draw marker on current point
            if (path.getCurrentPoint() != null) {
                if (firstPoint) {
                    if (startMarker != null) {
                        Graphic markerGraphic = startMarker.getGraphic();
                        if (markerGraphic != null) {
                            Point2D currentPoint = path.getCurrentPoint();
                            String markerOrient = startMarker.getOrient();
                            if (!markerOrient.equals("auto")) {
                                float rotateAngle = Float.valueOf(markerOrient).floatValue();
                                markerGraphic.rotate(Math.toRadians(rotateAngle));
                            } else {
                                // need to determine current path direction and rotate to match it
                            }
                            if (startMarker.getRefX() != null && startMarker.getRefY() != null) {
                                float refX = Float.valueOf(startMarker.getRefX()).floatValue();
                                float refY = Float.valueOf(startMarker.getRefY()).floatValue();
                                markerGraphic.translate(currentPoint.getX() - refX, currentPoint.getY() - refY);
                            } else {
                                markerGraphic.translateTo(currentPoint.getX(), currentPoint.getY());
                            }
                            group.addElement(markerGraphic);
                        }
                    }
                    firstPoint = false;
                } else  {
                    if (midMarker != null) {
                        Graphic markerGraphic = midMarker.getGraphic();
                        if (markerGraphic != null) {
                            Point2D currentPoint = path.getCurrentPoint();
                            String markerOrient = midMarker.getOrient();
                            if (!markerOrient.equals("auto")) {
                                float rotateAngle = Float.valueOf(markerOrient).floatValue();
                                markerGraphic.rotate(Math.toRadians(rotateAngle));
                            } else {
                                // need to determine current path direction and rotate to match it
                            }
                            if (midMarker.getRefX() != null && midMarker.getRefY() != null) {
                                float refX = Float.valueOf(midMarker.getRefX()).floatValue();
                                float refY = Float.valueOf(midMarker.getRefY()).floatValue();
                                markerGraphic.translate(currentPoint.getX() - refX, currentPoint.getY() - refY);
                            } else {
                                markerGraphic.translateTo(currentPoint.getX(), currentPoint.getY());
                            }
                            group.addElement(markerGraphic);
                        }
                    }
                }
            }
            // do next command
            String command = st.nextToken();
            if (command.equals("Z") || command.equals("z")) {
                path.closePath();
            } else {
                if (st.hasMoreTokens()) {
                    String parameters = st.nextToken();

			// check to see if we got all of the parameters

			boolean got_all_of_parameters = false;

			while (! got_all_of_parameters)
			{
				final String VARIABLE_PREFIX = "$";

				final String VARIABLE_DELIMETER = " ";

				int variable_prefix_index = parameters.lastIndexOf(VARIABLE_PREFIX);

				int variable_delimeter_index = parameters.lastIndexOf(VARIABLE_DELIMETER);

				boolean have_not_seen_space_after_prefix 	= variable_prefix_index		
										> variable_delimeter_index;

				boolean not_at_end_of_d_attribute = st.hasMoreTokens();

				boolean tokenizer_is_in_middle_of_variable 	= have_not_seen_space_after_prefix
										&& not_at_end_of_d_attribute;

				if (tokenizer_is_in_middle_of_variable)
				{
					String next_part_of_parameters = st.nextToken();

					parameters += next_part_of_parameters;
				}
				else
				{
					got_all_of_parameters = true;
				}
			}


                    doCommand(path, command, parameters, _d);
                }
            }
        }

        // put in the end marker if there is one
        if (endMarker != null) {
            Graphic markerGraphic = endMarker.getGraphic();
            if (markerGraphic != null) {
                Point2D currentPoint = path.getCurrentPoint();
                String markerOrient = endMarker.getOrient();
                if (!markerOrient.equals("auto")) {
                    float rotateAngle = Float.valueOf(markerOrient).floatValue();
                    markerGraphic.rotate(Math.toRadians(rotateAngle));
                } else {
                    // need to determine current path direction and rotate to match it
                }
                if (endMarker.getRefX() != null && endMarker.getRefY() != null) {
                    float refX = Float.valueOf(endMarker.getRefX()).floatValue();
                    float refY = Float.valueOf(endMarker.getRefY()).floatValue();
                    markerGraphic.translate(currentPoint.getX() - refX, currentPoint.getY() - refY);
                } else {
                    markerGraphic.translateTo(currentPoint.getX(), currentPoint.getY());
                }
                group.addElement(markerGraphic);
            }
        }
        graphicPath.setGeneralPath(path);

        // get transform matrix and apply to graphic
        AffineTransform transformMatrix = getTransformMatrix(getTransform());
        if (transformMatrix != null)
            group.transform(transformMatrix);

        setGraphicStyle(graphicPath);

        setGraphicStyle(group);

        return group;
    }

    private MarkerElement getStartMarker() {
        if (_style == null || _style.indexOf("marker") == -1)
            return null;
        String searchString = "marker-start";
        if (_style.indexOf("marker-start") == -1)
             searchString = "marker";

        return findMarker(searchString);
    }

    private MarkerElement getEndMarker() {
        if (_style == null || _style.indexOf("marker") == -1)
            return null;
        String searchString = "marker-end";
        if (_style.indexOf("marker-end") == -1)
             searchString = "marker";

        return findMarker(searchString);
    }

    private MarkerElement getMidMarker() {
        if (_style == null || _style.indexOf("marker") == -1)
            return null;
        String searchString = "marker-mid";
        if (_style.indexOf("marker-mid") == -1)
             searchString = "marker";

        return findMarker(searchString);
    }

    private MarkerElement findMarker(String searchString) {

        StringTokenizer st = new StringTokenizer(_style, ";", false);
        while (st.hasMoreTokens()) {
            String styleSetting = st.nextToken();
            int colonIndex = styleSetting.indexOf(':');
            if (colonIndex != -1) {
                String styleAttribute = styleSetting.substring(0,colonIndex).trim();
                String styleValue = styleSetting.substring(colonIndex+1).trim();
                if (styleAttribute.equals(searchString)) {
                    int hashIndex = styleValue.indexOf('#');
                    if (hashIndex != -1) {
                        String id = styleValue.substring(hashIndex+1, styleValue.length()-1).trim();
                        // find marker in defs section

                        SvgGraphic parent = getParent();
                        while (parent != null) {
                            // does parent have a defs element
                            DefsElement defs = null;
                            if (parent instanceof SvgElement) {
                                defs = ((SvgElement)parent).getDefs();
                            } else if (parent instanceof GElement) {
                                defs = ((GElement)parent).getDefs();
                            } else if (parent instanceof SymbolElement) {
                                defs = ((SymbolElement)parent).getDefs();
                            } else if (parent instanceof MarkerElement) {
                                defs = ((MarkerElement)parent).getDefs();
                            }
                            if (defs != null) {
                                // found a defs element, see if it contains a marker with
                                // the id we want
                                Definable ref = defs.getDefinedElement(id);
                                if (ref != null && ref instanceof MarkerElement)
                                    return (MarkerElement)ref;
                            }
                            parent = parent.getParent();
                        }
                    }
                }
            }
        }
        return null;
    }


    



    private void doCommand(GeneralPath path, String command, String parameters, String data) {
        if (command.equals("M") || command.equals("m")) {
            doMoveTo(path, command, parameters);
        } else if (command.equals("L") || command.equals("l")) {
            doLineTo(path, command, parameters);
        } else if (command.equals("C") || command.equals("c")) {
            doCurveTo(path, command, parameters);
        } else if (command.equals("S") || command.equals("s")) {
            doSmoothCurveTo(path, command, parameters);
        } else if (command.equals("H") || command.equals("h")) {
            doHorizontalLineTo(path, command, parameters);
        } else if (command.equals("V") || command.equals("v")) {
            doVerticalLineTo(path, command, parameters);
        } else if (command.equals("Q") || command.equals("q")) {
            doQuadraticBezierCurveTo(path, command, parameters);
        } else if (command.equals("T") || command.equals("t")) {
            doTruetypeQuadraticBezierCurveTo(path, command, parameters);
        } else if (command.equals("A") || command.equals("a")) {
            doEllipticArc(path, command, parameters);
        } else {
            System.out.println("Unrecognised path command: '" + command + "' in path: " + data);
        }
    }


    private void doMoveTo(GeneralPath path, String command, String parameters) {
        boolean absolute = true;
        boolean firstPoint = true;

        if (command.equals("m"))
            absolute = false;

        parameters = parameters.trim();

        StringTokenizer st = new StringTokenizer(parameters, " ,", false);

        while (st.hasMoreTokens()) {

            // get x coordinate
            String token = st.nextToken();
            float x = Float.valueOf(m_svg_document.handlePossibleVariable(token)).floatValue();

            // get y coordinate
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float y = Float.valueOf(m_svg_document.handlePossibleVariable(token)).floatValue();

            // add new coordinate to path
            Point2D currentPoint = path.getCurrentPoint();
            if (currentPoint == null || absolute) { // the first point in the path will always be absolute
                if (firstPoint) {
                    path.moveTo(x, y);
                    firstPoint = false;
                } else {
                    path.lineTo(x,y);
                }
            } else { // relative coordinate
                if (firstPoint) {
                    path.moveTo((float)currentPoint.getX() + x, (float)currentPoint.getY() + y);
                    firstPoint = false;
                } else {
                    path.lineTo((float)currentPoint.getX() + x, (float)currentPoint.getY() + y);
                }
            }
        }
        lastControlPoint = null;
    }


    private void doLineTo(GeneralPath path, String command, String parameters) {
        boolean absolute = true;

        if (command.equals("l"))
            absolute = false;
        parameters = parameters.trim();
        StringTokenizer st = new StringTokenizer(parameters, " ,", false);

        while (st.hasMoreTokens()) {

            // get x coordinate
            String token = st.nextToken();
            float x = Float.valueOf(m_svg_document.handlePossibleVariable(token)).floatValue();

            // get y coordinate
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float y = Float.valueOf(m_svg_document.handlePossibleVariable(token)).floatValue();

            // add new coordinate to path
            Point2D currentPoint = path.getCurrentPoint();
            if (currentPoint == null) { // something wrong, all paths should start with a moveTo
                path.moveTo(x,y);
            } else if (absolute) {
                path.lineTo(x,y);
            } else { // relative coordinate
                path.lineTo((float)currentPoint.getX() + x, (float)currentPoint.getY() + y);
            }
        }
        lastControlPoint = null;
    }
    private Coordinate lastControlPoint = null;   // used for smoothCurveTo

    private void doCurveTo(GeneralPath path, String command, String parameters) {
        boolean absolute = true;

        if (command.equals("c"))
            absolute = false;
        parameters = parameters.trim();
        StringTokenizer st = new StringTokenizer(parameters, " ,", false);

        while (st.hasMoreTokens()) {

            // get x1 coordinate
            String token = st.nextToken();
            float x1 = Float.valueOf(m_svg_document.handlePossibleVariable(token)).floatValue();

            // get y1 coordinate
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float y1 = Float.valueOf(m_svg_document.handlePossibleVariable(token)).floatValue();

            // get x2 coordinate
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float x2 = Float.valueOf(m_svg_document.handlePossibleVariable(token)).floatValue();

            // get y2 coordinate
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float y2 = Float.valueOf(m_svg_document.handlePossibleVariable(token)).floatValue();

            // get x coordinate
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float x = Float.valueOf(m_svg_document.handlePossibleVariable(token)).floatValue();

            // get y coordinate
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float y = Float.valueOf(m_svg_document.handlePossibleVariable(token)).floatValue();

            // add new curve to path
            Point2D currentPoint = path.getCurrentPoint();
            if (currentPoint == null) { // something wrong, all paths should start with a moveTo
                path.moveTo(x,y);
                lastControlPoint = null;
            } else if (absolute) {
                path.curveTo(x1, y1, x2, y2, x, y);
                lastControlPoint = new Coordinate(x2, y2);
            } else { // relative coordinate
                float currX = (float)currentPoint.getX();
                float currY = (float)currentPoint.getY();
                path.curveTo(currX + x1, currY + y1,
                             currX + x2, currY + y2,
                             currX + x, currY + y);
                lastControlPoint = new Coordinate(currX + x2, currY + y2);
            }
        }
    }
    private void doSmoothCurveTo(GeneralPath path, String command, String parameters) {
        boolean absolute = true;

        if (command.equals("s"))
            absolute = false;
        parameters = parameters.trim();
        StringTokenizer st = new StringTokenizer(parameters, " ,", false);

        while (st.hasMoreTokens()) {

            // get x2 coordinate
            String token = st.nextToken();
            float x2 = Float.valueOf(m_svg_document.handlePossibleVariable(token)).floatValue();

            // get y2 coordinate
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float y2 = Float.valueOf(m_svg_document.handlePossibleVariable(token)).floatValue();

            // get x coordinate
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float x = Float.valueOf(m_svg_document.handlePossibleVariable(token)).floatValue();

            // get y coordinate
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float y = Float.valueOf(m_svg_document.handlePossibleVariable(token)).floatValue();

            // add new curve to path
            Point2D currentPoint = path.getCurrentPoint();
           
            if (currentPoint == null) { // something wrong, all paths should start with a moveTo
                path.moveTo(x,y);
            } else {
                float currX = (float)currentPoint.getX();
                float currY = (float)currentPoint.getY();
                if (lastControlPoint == null)
                    lastControlPoint = new Coordinate(currX, currY);
                if (absolute) {
                    path.curveTo(2*currX - lastControlPoint.x, 2*currY - lastControlPoint.y, x2, y2, x, y);
                    lastControlPoint = new Coordinate(x2, y2);
                } else { // relative coordinate
                    path.curveTo(2*currX - lastControlPoint.x, 2*currY - lastControlPoint.y,
                                currX + x2, currY + y2,
                                currX + x, currY + y);
                    lastControlPoint = new Coordinate(currX + x2, currY + y2);
                }
            }
        }
    }


    private void doHorizontalLineTo(GeneralPath path, String command, String parameters) {
        boolean absolute = true;

        if (command.equals("h"))
            absolute = false;
        parameters = parameters.trim();
        StringTokenizer st = new StringTokenizer(parameters, " ,", false);

        while (st.hasMoreTokens()) {

            // get x coordinate
            String token = st.nextToken();
            float x = Float.valueOf(m_svg_document.handlePossibleVariable(token)).floatValue();

            // add new coordinate to path
            Point2D currentPoint = path.getCurrentPoint();
            if (currentPoint != null) {
                if (absolute) {
                    path.lineTo(x,(float)currentPoint.getY());
                } else { // relative coordinate
                    path.lineTo((float)currentPoint.getX() + x, (float)currentPoint.getY());
                }
            }
        }
        lastControlPoint = null;
    }


    private void doVerticalLineTo(GeneralPath path, String command, String parameters) {
        boolean absolute = true;

        if (command.equals("v"))
            absolute = false;
        parameters = parameters.trim();
        StringTokenizer st = new StringTokenizer(parameters, " ,", false);

        while (st.hasMoreTokens()) {

            // get y coordinate
            String token = st.nextToken();
            float y = Float.valueOf(m_svg_document.handlePossibleVariable(token)).floatValue();

            // add new coordinate to path
            Point2D currentPoint = path.getCurrentPoint();
            if (currentPoint != null) {
                if (absolute) {
                    path.lineTo((float)currentPoint.getX(), y);
                } else { // relative coordinate
                    path.lineTo((float)currentPoint.getX(), (float)currentPoint.getY() + y);
                }
            }
        }
        lastControlPoint = null;
    }

 /*  // keep these functions just in case the spec changes back

    private void doEllipticArcAngle(GeneralPath path, String command, String parameters) {

        boolean absolute = true;
        if (command.equals("a"))
            absolute = false;
        parameters = parameters.trim();
        StringTokenizer st = new StringTokenizer(parameters, " ,", false);
         while (st.hasMoreTokens()) {

            // get cx coordinate
            String token = st.nextToken();
            float cx = Float.valueOf(token).floatValue();

            // get cy coordinate
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float cy = Float.valueOf(token).floatValue();

            // get rx
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float rx = Float.valueOf(token).floatValue();

            // get ry
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float ry = Float.valueOf(token).floatValue();

            // get rotation
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float rotation = Float.valueOf(token).floatValue();

            // get start angle
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float startAngle = Float.valueOf(token).floatValue();

            // get sweep angle
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float sweepAngle = Float.valueOf(token).floatValue();

            // get line-to flag
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            int lineTo = Integer.valueOf(token).intValue();

            // add new curve to path
            Point2D currentPoint = path.getCurrentPoint();

            if (absolute || currentPoint == null) {
                Arc2D arc = new Arc2D.Float(cx - rx, cy - ry, rx * 2, ry * 2, startAngle, sweepAngle, Arc2D.OPEN);
                GeneralPath arcPath = new GeneralPath();
                arcPath.append(arc, false);
                arcPath.transform(AffineTransform.getRotateInstance(Math.toRadians(rotation), cx, cy));
                if (lineTo == 1)
                    path.append(arcPath, true);
                else
                    path.append(arcPath, false);
            } else {
                Arc2D arc = new Arc2D.Float((float)currentPoint.getX() + cx - rx,
                                            (float)currentPoint.getY()+ cy - ry,
                                            rx * 2, ry * 2, startAngle, sweepAngle, Arc2D.OPEN);
                GeneralPath arcPath = new GeneralPath();
                arcPath.append(arc, false);
                arcPath.transform(AffineTransform.getRotateInstance(Math.toRadians(rotation),
                                                                    (float)currentPoint.getX() + cx,
                                                                    (float)currentPoint.getY() + cy));
                if (lineTo == 1)
                    path.append(arcPath, true);
                else
                    path.append(arcPath, false);
            }
        }
        lastControlPoint = null;
    }


     private void doEllipticArcVector(GeneralPath path, String command, String parameters) {

        boolean absolute = true;
        if (command.equals("b"))
            absolute = false;
        parameters = parameters.trim();
        StringTokenizer st = new StringTokenizer(parameters, " ,", false);
         while (st.hasMoreTokens()) {

            // get cx coordinate
            String token = st.nextToken();
            float cx = Float.valueOf(token).floatValue();

            // get cy coordinate
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float cy = Float.valueOf(token).floatValue();

            // get rx
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float rx = Float.valueOf(token).floatValue();

            // get ry
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float ry = Float.valueOf(token).floatValue();

            // get rotation
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float rotation = Float.valueOf(token).floatValue();

            // get start x
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float startX = Float.valueOf(token).floatValue();

            // get start y
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float startY = Float.valueOf(token).floatValue();

            // get end x
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float endX = Float.valueOf(token).floatValue();

            // get end y
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float endY = Float.valueOf(token).floatValue();

            // get sweep flag
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            int sweepFlag = Integer.valueOf(token).intValue();

            // get line-to flag
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            int lineTo = Integer.valueOf(token).intValue();

            // start angle = angle from cx,cy to startX,startY
            // do cy - startY and not the other way around to take acount of the
            // y axis being upside down compared to cartesian coords
            float startAngle = (float)Math.toDegrees(Math.atan2((cy - startY),(startX - cx)));

            // end angle = angle from cx,cy to endX,endY
            float endAngle = (float)Math.toDegrees(Math.atan2((cy - endY),(endX - cx)));

            float sweepAngle;
            if (sweepFlag == 1) { // sweep angle positive
                if (startAngle > endAngle)
                    sweepAngle = endAngle - startAngle + 360;
                else
                    sweepAngle = endAngle - startAngle;
            } else {   // sweep angle negative
                if (startAngle > endAngle)
                    sweepAngle = endAngle - startAngle;
                else
                    sweepAngle = endAngle - startAngle - 360;
            }

            // add new curve to path
            Point2D currentPoint = path.getCurrentPoint();

            if (absolute || currentPoint == null) {
                Arc2D arc = new Arc2D.Float(cx - rx, cy - ry, rx * 2, ry * 2, startAngle, sweepAngle, Arc2D.OPEN);
                GeneralPath arcPath = new GeneralPath();
                arcPath.append(arc, false);
                arcPath.transform(AffineTransform.getRotateInstance(Math.toRadians(rotation), cx, cy));
                if (lineTo == 1)
                    path.append(arcPath, true);
                else
                    path.append(arcPath, false);
            } else {
                Arc2D arc = new Arc2D.Float((float)currentPoint.getX() + cx - rx,
                                            (float)currentPoint.getY()+ cy - ry,
                                            rx * 2, ry * 2, startAngle, sweepAngle, Arc2D.OPEN);
                GeneralPath arcPath = new GeneralPath();
                arcPath.append(arc, false);
                arcPath.transform(AffineTransform.getRotateInstance(Math.toRadians(rotation),
                                                                    (float)currentPoint.getX() + cx,
                                                                    (float)currentPoint.getY() + cy));
                if (lineTo == 1)
                    path.append(arcPath, true);
                else
                    path.append(arcPath, false);
            }
        }
        lastControlPoint = null;
    }
   */

    private void doEllipticArc(GeneralPath path, String command, String parameters) {

        boolean absolute = true;
        if (command.equals("a"))
            absolute = false;
        parameters = parameters.trim();
        StringTokenizer st = new StringTokenizer(parameters, " ,", false);
        while (st.hasMoreTokens()) {

            // get rx coordinate
            String token = st.nextToken();
            float rx = Float.valueOf(m_svg_document.handlePossibleVariable(token)).floatValue();

            // get ry coordinate
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float ry = Float.valueOf(m_svg_document.handlePossibleVariable(token)).floatValue();

            // get x-axis-rotation
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float rotation = Float.valueOf(m_svg_document.handlePossibleVariable(token)).floatValue();

            // get large-arc-flag
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            int largeArc = Integer.valueOf(m_svg_document.handlePossibleVariable(token)).intValue();

            // get sweep-flag
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            int sweepFlag = Integer.valueOf(m_svg_document.handlePossibleVariable(token)).intValue();

            // get x coordinate
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float x = Float.valueOf(m_svg_document.handlePossibleVariable(token)).floatValue();

            // get y coordinate
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float y = Float.valueOf(m_svg_document.handlePossibleVariable(token)).floatValue();


            // add new curve to path
            Point2D currentPoint = path.getCurrentPoint();

            if (currentPoint == null) {
                // just do a move to
                path.moveTo(x,y);
            } else {

	      // this is absolutely horrible code (really, I mean it!)
	      // someone please tell me how to do rotated ellipses
	      // this crap only handles circular arcs
	      // - tons of variables included to make sure things
	      //   were working - have to clean this up.
	      
	      double a = currentPoint.getX();
	      double b = currentPoint.getY();
	      double c = x;
	      double d = y;
	      double r = rx;
	      
	      double q = Math.sqrt((a-c)*(a-c) + (b-d)*(b-d));
	      double h = Math.sqrt(r*r - (q * q * 0.25));
	      
	      double x1 = a + 0.5 * (c-a) + (h * (d-b) / q);
	      double y1 = b + 0.5 * (d-b) - (h * (c-a) / q);
	      double x2 = a + 0.5 * (c-a) - (h * (d-b) / q);
	      double y2 = b + 0.5 * (d-b) + (h * (c-a) / q);
	      
	      double xcentre = 0.0;
	      double ycentre = 0.0;
	      double angleextent = 0.0;
		    
	      double anglestart = Math.atan2(b-y1, a-x1);
	      if (anglestart < 0.0) {
		anglestart += 2.0 * Math.PI;
	      }
	      double angleend = Math.atan2(d-y1, c-x1);
	      if (angleend < 0.0) {
		angleend += 2.0 * Math.PI;
	      }

	      // convert to degrees
	      anglestart *= 180.0 / Math.PI;
	      anglestart = 360 - anglestart;	
	      angleend *= 180.0 / Math.PI;
	      angleend = 360 - angleend;
	      //System.out.println("X1 = " + x1 + " Y1 = " + y1);
	      //System.out.println("anglestart = " + anglestart + " End = " + angleend);

	      boolean possweep = sweepFlag == 1;
	      boolean bigsweep = largeArc == 1;

	      if (possweep) {
		if (anglestart > angleend) {
		  angleextent = 360 - anglestart + angleend;
		} else {
		  angleextent = angleend - anglestart;
		}

		//System.out.println("angleextent = " + angleextent  + " bs = " + bigsweep);
		if ((angleextent > 180 && !bigsweep) ||
		    (angleextent < 180 && bigsweep)) {  // tick
		  xcentre = x1;
		  ycentre = y1;
		  angleextent -= 360;
		} else {
		  xcentre = x2;
		  ycentre = y2;
		  anglestart = Math.atan2(b-y2, a-x2);
		  if (anglestart < 0.0) {
		    anglestart += 2.0 * Math.PI;
		  }
		  anglestart *= 180.0 / Math.PI;
		  anglestart = 360 - anglestart;	
		  angleextent *= -1;
		}

	      } else {

		if (anglestart < angleend) {
		  angleextent = 360 - angleend + anglestart;
		} else {
		  angleextent = anglestart - angleend;
		}

		//System.out.println("angleextent = " + angleextent  + " bs = " + bigsweep);

		if ((angleextent > 180 && bigsweep) || //tick
		    (angleextent < 180 && !bigsweep)) {
		  xcentre = x2;
		  ycentre = y2;
		  anglestart = Math.atan2(b-y2, a-x2);
		  if (anglestart < 0.0) {
		    anglestart += 2.0 * Math.PI;
		  }
		  anglestart *= 180.0 / Math.PI;
		  anglestart = 360 - anglestart;	
		} else {
		  xcentre = x1;
		  ycentre = y1;
		  angleextent = 360 - angleextent;		  
		}
	      }

	      //System.out.println("CX = " + xcentre + " CY = " + ycentre);
	      //System.out.println("RX = " + rx + " CY = " + ry);
	      //System.out.println("anglestart = " + anglestart + " Extent = " + angleextent);

	      if (absolute) {
		Arc2D arc = new Arc2D.Double(xcentre - rx, ycentre - ry, rx * 2.0, ry * 2.0, 
					     anglestart, angleextent, Arc2D.OPEN);
		GeneralPath arcPath = new GeneralPath();
		arcPath.append(arc, false);
		path.append(arcPath, true);
		
		path.lineTo(x,y);
		
	      } else {  // relative
		path.lineTo((float)currentPoint.getX() + x, (float)currentPoint.getY() + y);
	      }
            }
        }
        lastControlPoint = null;
    }


    private void doQuadraticBezierCurveTo(GeneralPath path, String command, String parameters) {
        boolean absolute = true;

        if (command.equals("q"))
            absolute = false;
        parameters = parameters.trim();
        StringTokenizer st = new StringTokenizer(parameters, " ,", false);

        while (st.hasMoreTokens()) {

            // get x1 coordinate
            String token = st.nextToken();
            float x1 = Float.valueOf(m_svg_document.handlePossibleVariable(token)).floatValue();

            // get y1 coordinate
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float y1 = Float.valueOf(m_svg_document.handlePossibleVariable(token)).floatValue();

            // get x coordinate
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float x = Float.valueOf(m_svg_document.handlePossibleVariable(token)).floatValue();

            // get y coordinate
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float y = Float.valueOf(m_svg_document.handlePossibleVariable(token)).floatValue();

            // add new curve to path
            Point2D currentPoint = path.getCurrentPoint();
            if (currentPoint == null) { // something wrong, all paths should start with a moveTo
                path.moveTo(x,y);
            } else {
                if (absolute) {
                    path.quadTo(x1, y1, x, y);
                    lastControlPoint = new Coordinate(x1, y1);
                } else { // relative coordinate
                    path.quadTo((float)currentPoint.getX() + x1, (float)currentPoint.getY() + y1,
                                (float)currentPoint.getX() + x, (float)currentPoint.getY() + y);
                    lastControlPoint = new Coordinate((float)currentPoint.getX() + x1, (float)currentPoint.getY() + y1);
                }
            }
        }

    }



    private void doTruetypeQuadraticBezierCurveTo(GeneralPath path, String command, String parameters) {
        boolean absolute = true;

        if (command.equals("t"))
            absolute = false;
        parameters = parameters.trim();
        StringTokenizer st = new StringTokenizer(parameters, " ,", false);

        while (st.hasMoreTokens()) {

            // get x coordinate
            String token = st.nextToken();
            float x = Float.valueOf(m_svg_document.handlePossibleVariable(token)).floatValue();

            // get y coordinate
            try {
                token = st.nextToken();
            } catch (NoSuchElementException e) {
                token = "0";
            }
            float y = Float.valueOf(m_svg_document.handlePossibleVariable(token)).floatValue();

            Point2D currentPoint = path.getCurrentPoint();
            if (currentPoint == null) {
                path.moveTo(x,y);
            } else {

                // if no last control point then make it the current point
                if (lastControlPoint == null)
                    lastControlPoint = new Coordinate((float)currentPoint.getX(), (float)currentPoint.getY());

                // calculate next control point to be reflection of the last
                // control point, relative to current point
                float currX = (float)currentPoint.getX();
                float currY = (float)currentPoint.getY();
                Coordinate nextControlPoint = new Coordinate(2*currX - lastControlPoint.x, 2*currY - lastControlPoint.y);

                // add curve to path
                if (absolute) {
                    path.quadTo(nextControlPoint.x, nextControlPoint.y, x, y);
                } else {
                    path.quadTo(nextControlPoint.x, nextControlPoint.y, currX + x, currY + y);
                }

                lastControlPoint = nextControlPoint;
            }
        }
    }



    // XmlElement stuff

    /**
     * Returns the tag that identifies a path element in an SVG/XML file.
     * @return The tag that identifies a path element in an SVG/XML file.
     */
    public String getTag() {
        return "path";
    }

    /**
     * Reads in PathElement data from the given XML Element.
     * @param elem The XML Element that contains data to be read.
     */
    public void read(Element elem) {

        // read the standard graphic attributes
         readStandardGraphicAttributes(elem);

         // read d attribute
         _d = elem.getAttribute("d");

         // read flatness attribute
         String att = elem.getAttribute("flatness");
         if (att.length() > 0)
            _flatness = att;

         // read nominal-length attribute
         att = elem.getAttribute("nominalLength");
         if (att.length() > 0)
             _nominalLength = att;

        // read each child
        NodeList childNodes = elem.getChildNodes();
        int numChildren = childNodes.getLength();
        for (int i = 0; i < numChildren; i++) {
            Node child = childNodes.item(i);
            if (child.getNodeType() == Node.ELEMENT_NODE) { // it is an element
                Element childElement = (Element)child;
                String childTag = childElement.getTagName();

                if (childTag.equals("desc")) {
                    _desc = new DescElement();
                    _desc.read(childElement);

                } else if (childTag.equals("title")) {
                    _title = new TitleElement();
                    _title.read(childElement);
                }
             }
         }
    }

    /**
     * Inserts data into the given Document.
     * The data will become a child of the given parent element.
     *
     * @param doc The Document to write to.
     * @param parent Indicates the parent element of the data to be written.
     */
    public void write(Document doc, Node parent) {

        // create new element for this path
        Element path = doc.createElement(getTag());

        // set the standard graphic attributes
        setStandardGraphicAttributes(path);

        // set the attributes
        path.setAttribute("d", _d);
        if (_flatness != null)
            path.setAttribute("flatness", _flatness);
        if (_nominalLength != null)
            path.setAttribute("nominalLength", _nominalLength);

        if (_desc != null)
            _desc.write(doc, path);
        if (_title != null)
            _title.write(doc, path);

        // append new element to the parent element
        parent.appendChild(path);
    }


}
