/*
 * Copyright (C) 1999
 * CSIRO Mathematical and Information Sciences Australia
 * Spatial Information Systems Project
 *
 * http://www.cmis.csiro.au/svg
 */


/**
 * PolygonElement.java
 *
 * $Id: PolygonElement.java,v 1.2 2000/05/04 23:24:22 jwnichls Exp $
 *
 */

package org.csiro.svgv.reader;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.csiro.svgv.graphic.*;
import java.awt.geom.GeneralPath;
import java.awt.geom.Arc2D;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;

/**
 * The PolygonElement class describes an SVG polygon element.
 */
public class PolygonElement extends SvgStdGraphic implements Definable {

    private DescElement _desc = null;
    private TitleElement _title = null;

    // points attribute
    private String _points;

    /**
     * Constructs a new polygonElement with no points.
     */
    public PolygonElement() {
        super();
        _points = "";    // points attribute is required so initialise it to
                        // a default value
    }

    private SvgDocument m_svg_document;
    public PolygonElement(SvgDocument svg_document) {
	this();
	m_svg_document = svg_document;
    }

    /**
     * Constructs and initializes a polygonElement with the given points string.
     * @param points The value to use when initializing the points attribute.
     */
    public PolygonElement(String points) {
        super();
        if (points != null)
            _points = points;
        else
            _points = "";
    }

    /**
     * Sets the desc element.
     * @param desc The value to use when setting the desc element.
     */
    public void setDesc(DescElement desc) {
        _desc = desc;
    }

    /**
     * Sets the title element.
     * @param title The value to use when setting the title element.
     */
    public void setTitle(TitleElement title) {
        _title = title;
    }

    /**
     * Returns the desc element.
     * @return The desc element.
     */
    public DescElement getDesc() {
        return _desc;
    }

    /**
     * Returns the title element.
     * @return The title element.
     */
    public TitleElement getTitle() {
        return _title;
    }


    /**
     * Returns the string of pointsthat make up this polygon.
     * @return The points attribute of this polygon.
     */
    public String getPoints() {
        return _points;
    }

    /**
     * Sets the points attribute of this polygon to the specified points string.
     * @param points The string containing the (space or comma seperated) points.
     */
    public void setPoints(String points) {
        if (points != null)
            _points = points;
        else
            _points = "";
    }

    /**
     * Adds the specified point to the points attribute.
     * @param x The X corrdinate of the point to be added.
     * @param y The Y corrdinate of the point to be added.
     */
    public void addPoint(float x, float y) {
        String newPoint = x + "," + y;
        if (_points.length() > 0)
            _points += " " + newPoint;
        else
            _points = newPoint;
    }

    /**
     * Returns the Graphic object that represents this polygon.
     * @return The Graphic object that represents this polygon.
     */
    public Graphic getGraphic() {

        setStyleProperties();

        Path graphic = new Path();
        if (getId() != null)
            graphic.setId(getId());

        GeneralPath path = new GeneralPath();
        StringTokenizer st = new StringTokenizer(_points, " ,", false);
        if (st.countTokens() >= 4) {
            // read first point
            String token = st.nextToken();
            float startX = Float.valueOf(token).floatValue();
            token = st.nextToken();
            float startY = Float.valueOf(token).floatValue();
            path.moveTo(startX, startY);
            // read the other points
            while(st.hasMoreTokens()) {
                try {
                    token = st.nextToken();
                    float x = Float.valueOf(token).floatValue();
                    token = st.nextToken();
                    float y = Float.valueOf(token).floatValue();
                    path.lineTo(x,y);
                } catch (NoSuchElementException e) {}
            }
            // add a line to the first point to close the polygon
            path.closePath();
        }
        graphic.setGeneralPath(path);
       

        // get transform matrix and apply to graphic
        AffineTransform transformMatrix = getTransformMatrix(getTransform());
        if (transformMatrix != null)
            graphic.transform(transformMatrix);

        setGraphicStyle(graphic);

        return graphic;
    }



    // XmlElement stuff

    /**
     * Returns the tag that identifies a polygon element in an SVG/XML file.
     * @return The tag that identifies a polygon element in an SVG/XML file.
     */
    public String getTag() {
        return "polygon";
    }

    /**
     * Reads in polygonElement data from the given XML Element.
     * @param elem The XML Element that contains data to be read.
     */
    public void read(Element elem) {

         // read the standard graphic attributes
         readStandardGraphicAttributes(elem);

         // read points attribute
         _points = elem.getAttribute("points");

          // read each child
        NodeList childNodes = elem.getChildNodes();
        int numChildren = childNodes.getLength();
        for (int i = 0; i < numChildren; i++) {
            Node child = childNodes.item(i);
            if (child.getNodeType() == Node.ELEMENT_NODE) { // it is an element
                Element childElement = (Element)child;
                String childTag = childElement.getTagName();

                if (childTag.equals("desc")) {
                    _desc = new DescElement();
                    _desc.read(childElement);

                } else if (childTag.equals("title")) {
                    _title = new TitleElement();
                    _title.read(childElement);
                }
             }
         }
    }

    /**
     * Inserts data into the given Document.
     * The data will become a child of the given parent element.
     *
     * @param doc The Document to write to.
     * @param parent Indicates the parent element of the data to be written.
     */
    public void write(Document doc, Node parent) {

        // create new element for this polygon
        Element polygon = doc.createElement(getTag());

        // set the standard graphic attributes
        setStandardGraphicAttributes(polygon);

        // set the points attribute
        polygon.setAttribute("points", _points);

        if (_desc != null)
            _desc.write(doc, polygon);
        if (_title != null)
            _title.write(doc, polygon);

        // append new element to the parent element
        parent.appendChild(polygon);

    }




}
