/*
 * Copyright (C) 1999
 * CSIRO Mathematical and Information Sciences Australia
 * Spatial Information Systems Project
 *
 * http://www.cmis.csiro.au/svg
 */


/**
 * RadialGradientElement.java
 *
 * $Id: RadialGradientElement.java,v 1.1.1.1 1999/12/11 18:51:28 gjb Exp $
 *
 */

package org.csiro.svgv.reader;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Document;
import java.awt.Paint;
import java.awt.TexturePaint;
import java.awt.geom.Rectangle2D;
import java.awt.geom.Line2D;
import java.awt.Color;
import java.awt.image.BufferedImage;

/**
 * The RadialGradientElement class describes an SVG radialgradient element.
 */
public class RadialGradientElement extends GradientElement implements Definable, XmlElement {

    // radialgradient attributes
    private String _cx;
    private String _cy;
    private String _r;
    private String _fx;
    private String _fy;

    /**
     * Constructs a RadialGradientElement with no gradient stops.
     */
    public RadialGradientElement() {
        super();
    }

     public RadialGradientElement(String cx, String cy, String r, String fx, String fy) {
        _cx = cx;
        _cy = cy;
        _r = r;
        _fx = fx;
        _fy = fy;
    }


    public RadialGradientElement(float cx, float cy, float r, float fx, float fy) {
        _cx = Float.toString(cx);
        _cy = Float.toString(cy);
        _r = Float.toString(r);
        _fx = Float.toString(fx);
        _fy = Float.toString(fy);
    }


    /**
     * Returns the TexturePaint object that represents this radialgradient element.
     * @param The shape that will be coloured in by the GradientPaint. This is
     * needed to determine the start and end points of the GradientPaint object.
     * @param objectBounds The bounds of the shape that will be coloured in by the GradientPaint.
     * This may be needed to determine the start and end points of the GradientPaint object.
     * @return The GradientPaint object that represents this gradient element.
     */
    public Paint getPaint(Rectangle2D userBounds, Rectangle2D objectBounds) {

        float cx;
        float cy;
        float fx;
        float fy;
        float r;

        float boxMinX;
        float boxMinY;
        float boxWidth;
        float boxHeight;

        if (getGradientUnits().equals("objectBoundingBox")) {
            boxMinX = (float)objectBounds.getMinX();
            boxMinY = (float)objectBounds.getMinY();
            boxWidth = (float)objectBounds.getWidth();
            boxHeight = (float)objectBounds.getHeight();
        } else { // user space
            boxMinX = (float)userBounds.getMinX();
            boxMinY = (float)userBounds.getMinY();
            boxWidth = (float)userBounds.getWidth();
            boxHeight = (float)userBounds.getHeight();
        }

        if (_cx.indexOf('%') != -1) // is a percentage
            cx = boxMinX + Float.parseFloat(_cx.substring(0,_cx.length()-1)) / 100 * boxWidth;
        else {
           if (getGradientUnits().equals("objectBoundingBox"))   // is a number between 0 and 1
               cx = boxMinX + Float.parseFloat(_cx) * boxWidth;
           else // is a plain coordinate
               cx = Float.parseFloat(_cx);
        }

         if (_cy.indexOf('%') != -1) // is a percentage
            cy = boxMinX + Float.parseFloat(_cy.substring(0,_cy.length()-1)) / 100 * boxHeight;
        else {
           if (getGradientUnits().equals("objectBoundingBox"))   // is a number between 0 and 1
               cy = boxMinX + Float.parseFloat(_cy) * boxHeight;
           else // is a plain coordinate
               cy = Float.parseFloat(_cy);
        }

         if (_fx.indexOf('%') != -1) // is a percentage
            fx = boxMinX + Float.parseFloat(_fx.substring(0,_fx.length()-1)) / 100 * boxWidth;
        else {
           if (getGradientUnits().equals("objectBoundingBox"))   // is a number between 0 and 1
               fx = boxMinX + Float.parseFloat(_fx) * boxWidth;
           else // is a plain coordinate
               fx = Float.parseFloat(_fx);
        }

        if (_fy.indexOf('%') != -1) // is a percentage
            fy = boxMinX + Float.parseFloat(_fy.substring(0,_fy.length()-1)) / 100 * boxHeight;
        else {
           if (getGradientUnits().equals("objectBoundingBox"))   // is a number between 0 and 1
               fy = boxMinX + Float.parseFloat(_fy) * boxHeight;
           else // is a plain coordinate
               fy = Float.parseFloat(_fy);
        }

        // if radius is a percentage then make it a percentage of the bbox width
        if (_r.indexOf('%') != -1) // is a percentage
            r = Float.parseFloat(_r.substring(0,_r.length()-1)) / 100 * boxWidth;
        else {
           if (getGradientUnits().equals("objectBoundingBox"))   // is a number between 0 and 1
               r =  Float.parseFloat(_r) * boxWidth;
           else // is a plain coordinate
               r = Float.parseFloat(_r);
        }




        int minX = (int)boxMinX;
        int minY = (int)boxMinY;
        int maxX = (int)(boxMinX + boxWidth);
        int maxY = (int)(boxMinY + boxHeight);


        Color color1 = null;
        Color color2 = null;

        StopElement stop1 = getStop(0);
        StopElement stop2 = getStop(1);
        if (stop1 != null && stop2 != null) {
             color1 = stop1.getColor();
             color2 = stop2.getColor();
        }

        if (color1 == null || color2 == null)
            return null;

        int diffR = color2.getRed() - color1.getRed();
        int diffG = color2.getGreen() - color1.getGreen();
        int diffB = color2.getBlue() - color1.getBlue();
        int diffA = color2.getAlpha() - color1.getAlpha();

        BufferedImage image = new BufferedImage(maxX-minX, maxY-minY,
                                                BufferedImage.TYPE_4BYTE_ABGR);

        // V1 is vector from (fx,fy) to (cx,cy)
        double v1x = cx - fx;
        double v1y = cy - fy;
        double magV1 = Math.sqrt(v1x*v1x + v1y*v1y);

        if (magV1 > r) {
            System.out.println("Radial Gradient focal point outside circle!! ");
            return null;
        }

        for (int x = minX; x < maxX; x++) {
            for (int y = minY; y < maxY; y++) {

                // V2 is vector from (fx,fy) to (x,y)
                double v2x = x - fx;
                double v2y = y - fy;
                double magV2 = Math.sqrt(v2x*v2x + v2y*v2y);

                double v1DotV2 = v1x * v2x + v1y * v2y;
                double angleV1V2 = Math.acos(v1DotV2/(magV1*magV2));

                double b = magV1 * Math.cos(angleV1V2);
                double d = magV1 * Math.sin(angleV1V2);
                double a = Math.sqrt(r*r - d*d);

                double maxDistAtAngleV1V2 = a + b;

                double percentFromF = magV2/maxDistAtAngleV1V2;
                if (percentFromF > 1.0)
                    percentFromF = 1.0;

                int red = color1.getRed() + (int)(percentFromF * diffR);
                int green = color1.getGreen() + (int)(percentFromF * diffG);
                int blue = color1.getBlue() + (int)(percentFromF * diffB);
                int alpha = color1.getAlpha() + (int)(percentFromF * diffA);

                Color c = new Color(red, green, blue, alpha);

                image.setRGB(x-minX, y-minY, c.getRGB());
            }
        }

        TexturePaint paint = new TexturePaint(image, new Rectangle2D.Float(minX, minY, maxX-minX, maxY-minY));

        return paint;
    }




     // XmlElement stuff

    /**
     * Returns the tag that identifies a radialgradient element in an SVG/XML file.
     * @return The tag that identifies a radialgradient element in an SVG/XML file.
     */
    public String getTag() {
        return "radialGradient";
    }

    /**
     * Reads in RadialGradientElement data from the given XML Element.
     * @param elem The XML Element that contains data to be read.
     */
    public void read(Element elem) {

        readStdGradientAttributes(elem);

        String att = elem.getAttribute("cx");
        if (att.length() > 0)
            _cx = att;

        att = elem.getAttribute("cy");
        if (att.length() > 0)
            _cy = att;

        att = elem.getAttribute("r");
        if (att.length() > 0)
            _r = att;

        att = elem.getAttribute("fx");
        if (att.length() > 0)
            _fx = att;

        att = elem.getAttribute("fy");
        if (att.length() > 0)
            _fy = att;


    }

    /**
     * Inserts data into the given Document.
     * The data will become a child of the given parent element.
     *
     * @param doc The Document to write to.
     * @param parent Indicates the parent element of the data to be written.
     */
    public void write(Document doc, Node parent) {

        // create new element for this radialgradient element
        Element radialGradient = doc.createElement(getTag());

        // write the standard gradient attributes and gradient stops
        writeStdGradientAttributes(doc, radialGradient);

         // set the attributes
        if (_cx != null)
            radialGradient.setAttribute("cx", _cx);
        if (_cy != null)
            radialGradient.setAttribute("cy", _cy);
        if (_r != null)
            radialGradient.setAttribute("r", _r);
        if (_fx != null)
            radialGradient.setAttribute("fx", _fx);
        if (_fy!= null)
            radialGradient.setAttribute("fy", _fy);

        // append new element to the parent element
        parent.appendChild(radialGradient);

    }

}