/*
 * Copyright (C) 1999
 * CSIRO Mathematical and Information Sciences Australia
 * Spatial Information Systems Project
 *
 * http://www.cmis.csiro.au/svg
 */


/**
 * RectElement.java
 *
 * $Id: RectElement.java,v 1.3 2000/05/04 23:24:22 jwnichls Exp $
 *
 */

package org.csiro.svgv.reader;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import java.awt.geom.GeneralPath;
import java.awt.geom.RoundRectangle2D;
import org.csiro.svgv.graphic.*;
import java.awt.geom.AffineTransform;

/**
 * The RectElement class describes an SVG rect element.
 */
public class RectElement extends SvgStdGraphic implements Definable {

    private DescElement _desc = null;
    private TitleElement _title = null;

    // default attribute values
    private static final String defaultRX = "0";
    private static final String defaultRY = "0";
    private static final String defaultX = "0";
    private static final String defaultY = "0";

    // rect attributes
    private String _x = null;
    private String _y = null;
    private String _width;
    private String _height;
    private String _rx = null;
    private String _ry = null;

	private SvgDocument m_svg_document;


    /**
     * Constructs a new RectElement. The width and height attributes are initialized
     * to 0.
     */
    public RectElement(SvgDocument svg_document) {
       this();

	m_svg_document = svg_document;
    }

    /**
     * Constructs a new RectElement. The width and height attributes are initialized
     * to 0.
     */
    public RectElement() {
       super();
       _width = "0";       // width and height attributes are required so
       _height = "0";      // initialize to some default values
    }

    /**
     * Constructs and initializes a RectElement with the specified dimensions.
     * @param width The width of the rectangle.
     * @param height The height of the rectangle.
     */
    public RectElement(String width, String height) {
        super();
        if (width != null)
            _width = width;
        else
            _width = "0";
        if (height != null)
            _height = height;
        else
            _height = "0";
    }

    /**
     * Constructs and initializes a RectElement with the specified location
     * and dimensions.
     * @param x The X coordinate of the upper left corner of the rectangle.
     * @param y The Y coordinate of the upper left corner of the rectangle.
     * @param width The width of the rectangle.
     * @param height The height of the rectangle.
     */
    public RectElement(String x, String y, String width, String height) {
        super();
        _x = x;
        _y = y;
        if (width != null)
            _width = width;
        else
            _width = "0";
        if (height != null)
            _height = height;
        else
            _height = "0";
    }

    /**
     * Constructs and initializes a RectElement with the specified dimensions.
     * @param width The width of the rectangle.
     * @param height The height of the rectangle.
     */
    public RectElement(float width, float height) {
        super();
        _width = Float.toString(width);
        _height = Float.toString(height);
    }

    /**
     * Constructs and initializes a RectElement with the specified location
     * and dimensions.
     * @param x The X coordinate of the upper left corner of the rectangle.
     * @param y The Y coordinate of the upper left corner of the rectangle.
     * @param width The width of the rectangle.
     * @param height The height of the rectangle.
     */
    public RectElement(float x, float y, float width, float height) {
        super();
        _x = Float.toString(x);
        _y = Float.toString(y);
        _width = Float.toString(width);
        _height = Float.toString(height);
    }

    /**
     * Constructs and initializes a RectElement with the specified location
     * and dimensions.
     * @param x The X coordinate of the upper left corner of the rectangle.
     * @param y The Y coordinate of the upper left corner of the rectangle.
     * @param width The width of the rectangle.
     * @param height The height of the rectangle.
     * @param rx The x radius of the ellipse used to round the rectangles corners.
     * @param ry The y radius of the ellipse used to round the rectangles corners.
     */
    public RectElement(float x, float y, float width, float height, float rx, float ry) {
        super();
        _x = Float.toString(x);
        _y = Float.toString(y);
        _width = Float.toString(width);
        _height = Float.toString(height);
        _rx = Float.toString(rx);
        _ry = Float.toString(ry);
    }

    /**
     * Sets the desc element.
     * @param desc The value to use when setting the desc element.
     */
    public void setDesc(DescElement desc) {
        _desc = desc;
    }

    /**
     * Sets the title element.
     * @param title The value to use when setting the title element.
     */
    public void setTitle(TitleElement title) {
        _title = title;
    }

    /**
     * Returns the desc element.
     * @return The desc element.
     */
    public DescElement getDesc() {
        return _desc;
    }

    /**
     * Returns the title element.
     * @return The title element.
     */
    public TitleElement getTitle() {
        return _title;
    }

    /**
     * Returns the X coordinate of the upper left corner of the rectangle.
     * @return The X coordinate of the upper left corner of the rectangle.
     */
    public String getX() {
        if (_x == null) {
            return defaultX;
        }
        return m_svg_document.handlePossibleVariable(_x);
    }

    /**
     * Returns the Y coordinate of the upper left corner of the rectangle.
     * @return The Y coordinate of the upper left corner of the rectangle.
     */
    public String getY() {
        if (_y == null) {
            return defaultY;
        }
        return m_svg_document.handlePossibleVariable(_y);
    }

    /**
     * Returns the x radius of the ellipse used to round the rectangles corners.
     * @return The x radius of the ellipse used to round the rectangles corners.
     */
    public String getRx() {
        if (_rx == null && _ry == null)
            return defaultRX;
        if (_rx == null && _ry != null)
            return m_svg_document.handlePossibleVariable(_ry);
        return m_svg_document.handlePossibleVariable(_rx);
    }

    /**
     * Returns the y radius of the ellipse used to round the rectangles corners.
     * @return The y radius of the ellipse used to round the rectangles corners.
     */
    public String getRy() {
        if (_ry == null && _rx == null)
            return defaultRY;
        if (_ry == null && _rx != null)
            return m_svg_document.handlePossibleVariable(_rx);
        return m_svg_document.handlePossibleVariable(_ry);
    }

    /**
     * Returns the width of the rectangle.
     * @return The width of the rectangle.
     */
    public String getWidth() {
        return m_svg_document.handlePossibleVariable(_width);
    }

    /**
     * Returns the height of the rectangle.
     * @return The height of the rectangle.
     */
    public String getHeight() {
        return m_svg_document.handlePossibleVariable(_height);
    }

    /**
     * Sets the X coordinate of the upper left corner of the rectangle to the
     * specfied value.
     * @param x The value to use when setting the X coordinate of the
     * upper left corner of the rectangle.
     */
    public void setX(float x) {
        _x = Float.toString(x);
    }

    /**
     * Sets the Y coordinate of the upper left corner of the rectangle to the
     * specfied value.
     * @param y The value to use when setting the Y coordinate of the
     * upper left corner of the rectangle.
     */
    public void setY(float y) {
        _y = Float.toString(y);
    }

    /**
     * Sets the width of the rectangle to the specified value.
     * @param width The value to use when setting the width of the rectangle.
     */
    public void setWidth(float width) {
        _width = Float.toString(width);
    }

    /**
     * Sets the height of the rectangle to the specified value.
     * @param height The value to use when setting the height of the rectangle.
     */
    public void setHeight(float height) {
        _height = Float.toString(height);
    }

    /**
     * Sets the location of the rectangle to the specition coordinates.
     * @param x The value to use when setting the X coordinate of the
     * upper left corner of the rectangle.
     * @param y The value to use when setting the Y coordinate of the
     * upper left corner of the rectangle.
     */
    public void setLocation(float x, float y) {
        _x = Float.toString(x);
        _y = Float.toString(y);
    }

    /**
     * Sets the dimensions of the rectangle to the specified values.
     * @param width The value to use when setting the width of the rectangle.
     * @param height The value to use when setting the height of the rectangle.
     */
    public void setDimensions(float width, float height) {
        _width = Float.toString(width);
        _height = Float.toString(height);
    }

    /**
     * Returns the Graphic object that represents this rectangle.
     * @return The Graphic object that represents this rectangle.
     */
    public Graphic getGraphic() {

        setStyleProperties();

        Path graphic = new Path();
        if (getId() != null)
            graphic.setId(getId());

        GeneralPath path = new GeneralPath();
        float x = Float.valueOf(getX()).floatValue();
        float y = Float.valueOf(getY()).floatValue();
        float rx = Float.valueOf(getRx()).floatValue();
        float ry = Float.valueOf(getRy()).floatValue();
        float width = Float.valueOf(getWidth()).floatValue();
        float height = Float.valueOf(getHeight()).floatValue();
        RoundRectangle2D rect = new RoundRectangle2D.Float(x, y, width, height, rx * 2, ry * 2);
        path.append(rect, false);

        graphic.setGeneralPath(path);
       

        // get transform matrix and apply to graphic
        AffineTransform transformMatrix = getTransformMatrix(getTransform());
        if (transformMatrix != null)
            graphic.transform(transformMatrix);

        setGraphicStyle(graphic);
        
        return graphic;
    }


    // XmlElement stuff

    /**
     * Returns the tag that identifies a rect element in an SVG/XML file.
     * @return The tag that identifies a rect element in an SVG/XML file.
     */
    public String getTag() {
        return "rect";
    }

    /**
     * Reads in RectElement data from the given XML Element.
     * @param elem The XML Element that contains data to be read.
     */
    public void read(Element elem) {

         // read the standard graphic attributes
         readStandardGraphicAttributes(elem);

         // read x attribute
         _x = elem.getAttribute("x");
         if (_x.equals(""))
            _x = null;

         // read y attribute
         _y = elem.getAttribute("y");
         if (_y.equals(""))
            _y = null;

         // read width attribute
         _width = elem.getAttribute("width");
System.out.println("width = " + _width);
         // read height attribute
         _height = elem.getAttribute("height");

         // read rx attribute
         _rx = elem.getAttribute("rx");
         if (_rx.equals(""))
            _rx = null;

         // read ry attribute
         _ry = elem.getAttribute("ry");
          if (_ry.equals(""))
            _ry = null;

         // read each child
        NodeList childNodes = elem.getChildNodes();
        int numChildren = childNodes.getLength();
        for (int i = 0; i < numChildren; i++) {
            Node child = childNodes.item(i);
            if (child.getNodeType() == Node.ELEMENT_NODE) { // it is an element
                Element childElement = (Element)child;
                String childTag = childElement.getTagName();

                if (childTag.equals("desc")) {
                    _desc = new DescElement();
                    _desc.read(childElement);

                } else if (childTag.equals("title")) {
                    _title = new TitleElement();
                    _title.read(childElement);
                }
             }
         }
    }

    /**
     * Inserts data into the given Document.
     * The data will become a child of the given parent element.
     *
     * @param doc The Document to write to.
     * @param parent Indicates the parent element of the data to be written.
     */
    public void write(Document doc, Node parent) {

        // create new element for this rect
        Element rect = doc.createElement(getTag());

        // set the standard graphic attributes
        setStandardGraphicAttributes(rect);

        // set the attributes, only set ones that have defaults if they are not null
        if (_x != null)
            rect.setAttribute("x", _x);
        if (_y != null)
            rect.setAttribute("y", _y);
        rect.setAttribute("width", _width);
        rect.setAttribute("height", _height);
        if (_rx != null)
            rect.setAttribute("rx", _rx);
        if (_ry != null)
            rect.setAttribute("ry", _ry);

         if (_desc != null)
            _desc.write(doc, rect);
        if (_title != null)
            _title.write(doc, rect);

        // append new element to the parent element
        parent.appendChild(rect);
    }
}
