/*
 * Copyright (C) 1999
 * CSIRO Mathematical and Information Sciences Australia
 * Spatial Information Systems Project
 *
 * http://www.cmis.csiro.au/svg
 */


/**
 * StopElement.java
 *
 * $Id: StopElement.java,v 1.1.1.1 1999/12/11 18:51:28 gjb Exp $
 *
 */

package org.csiro.svgv.reader;


import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import java.awt.Color;
import java.util.StringTokenizer;

/**
 * The StopElement class describes an SVG stop element.
 */
public class StopElement implements XmlElement {

    //  stop attributes
    private String _style = null;
    private String _id = null;
    private String _offset;  // required attribute

    /**
     * Constructs a new StopElement. The offset attribute is
     * initialised to 0 since it is a required attribute.
     */
    public StopElement() {
        _offset = "0";        // offset attribute is required so initialize it
    }

    /**
     * Constructs and initialises a new GradientStopElement.
     * @param offset The value to use when initialising the offset attribute.
     */
    public StopElement(String offset) {
        if (offset != null)
            _offset = offset;
        else
            _offset = "0";
    }

    /**
     * Sets the id attribute.
     * @param id The value to use when setting the id attribute.
     */
    public void setId(String id) {
        _id = id;
    }

    /**
     * Sets the style attribute.
     * @param style The value to use when setting the style attribute.
     */
    public void setStyle(String style) {
        _style = style;
    }

    /**
     * Sets the offset attribute.
     * @param offset The value to use when setting the offset attribute.
     */
    public void setOffset(String offset) {
        if (offset != null)
            _offset = offset;
        else
            _offset = "0";
    }

    /**
     * Returns the value of the id attribute.
     * @return The value of the id attribute.
     */
    public String getId() {
        return _id;
    }

    /**
     * Returns the value of the style attribute.
     * @return The value of the style attribute.
     */
    public String getStyle() {
        return _style;
    }

    /**
     * Returns the value of the offset attribute.
     * @return The value of the offset attribute.
     */
    public String getOffset() {
        return _offset;
    }

    public Color getColor() {

        if (_style == null)
            return Color.black;

        // find the color and opacity parameters in the style attribute
        String colorString = null;
        String opacityString = null;

        StringTokenizer st = new StringTokenizer(_style, ";", false);
        while (st.hasMoreTokens()) {
            String styleSetting = st.nextToken();
            int colonIndex = styleSetting.indexOf(':');
            if (colonIndex > 0) {
                String attribute = styleSetting.substring(0,colonIndex);
                String value = styleSetting.substring(colonIndex+1);
                attribute = attribute.trim();
                value = value.trim();

                if (attribute.equals("color")) {
                    colorString = value;
                } else if (attribute.equals("opacity")) {
                    opacityString = value;
                }
            }
        }

        if (colorString != null) {
            Color color = SvgGraphic.getColor(colorString);
            float opacity = 1.0f;
            if (opacityString != null) {
                try {
                    int index = opacityString.indexOf('%');
                    if (index == -1) { // should be a number between 0 and 1
                        opacity = Float.valueOf(opacityString).floatValue();
                    } else { // must be percentage
                        String percentage = opacityString.substring(0,index);
                        opacity = Float.valueOf(percentage).floatValue() / 100;
                    }
                } catch (NumberFormatException e) {
                    System.out.println("cannot decode opacity: " + opacityString);
                }
            }
            return new Color(color.getRed(), color.getGreen(), color.getBlue(), (int)(opacity * 255));
        }
        return Color.black;
    }



    // XmlElement stuff

    /**
     * Returns the tag that identifies a gradientstop element in an SVG/XML file.
     * @return The tag that identifies a gradientstop element in an SVG/XML file.
     */
    public String getTag() {
        return "stop";
    }

    /**
     * Reads in StopElement data from the given XML Element.
     * @param elem The XML Element that contains data to be read.
     */
    public void read(Element elem) {

        String att = elem.getAttribute("style");
        if (att.length() > 0)
            _style = att;

        att = elem.getAttribute("id");
        if (att.length() > 0)
            _id = att;

        _offset = elem.getAttribute("offset");
    }

    /**
     * Inserts data into the given Document.
     * The data will become a child of the given parent element.
     *
     * @param doc The Document to write to.
     * @param parent Indicates the parent element of the data to be written.
     */
    public void write(Document doc, Node parent) {

        // create new element for this stop element
        Element stop = doc.createElement(getTag());

         // set the attributes
        if (_style != null)
            stop.setAttribute("style", _style);
        if (_id != null)
            stop.setAttribute("id", _id);
        stop.setAttribute("offset", _offset);

        // append new element to the parent element
        parent.appendChild(stop);
    }
}