/*
 * Copyright (C) 1999
 * CSIRO Mathematical and Information Sciences Australia
 * Spatial Information Systems Project
 *
 * http://www.cmis.csiro.au/svg
 */


/**
 * SvgElement.java
 *
 * $Id: SvgElement.java,v 1.13 2000/05/08 03:12:39 will Exp $
 *
 */

package org.csiro.svgv.reader;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import java.util.Vector;
import org.csiro.svgv.graphic.*;
import java.awt.geom.AffineTransform;

import EDU.Washington.grad.gjb.cassowary.*;
/**
 * The SvgElement class describes an SVG svg element.
 */
public class SvgElement extends SvgStdGraphic implements Definable {

  // children elements
  private DefsElement _defs = null;
  private DescElement _desc = null;
  private TitleElement _title = null;
  private Vector _children;

  // attributes
  private static final String _xmlns = "http://www.w3.org/Graphics/SVG/SVG-19990812";

  private String _onunload = null;
  private String _onzoom = null;

  private String _refX = null;
  private String _refY = null;

  private String _x = null;
  private String _y = null;
  private String _width;
  private String _height;

  private String _allowZoomAndPan = null;
  private static final String defaultAllowZoomAndPan = "true";
  private String _viewBox = null;
  private String _preserveAspectRatio = null;
  private String _contentScriptType = null;

  /**
   * Constructs a new SvgElement. The width and height attributes are
   * initialised to 100.
   */
  /*
    public SvgElement() {
    super();
    _children = new Vector();
    _width = "100";         // width and height are required attributes so
    _height = "100";        // initialize them to some default values
    }
  */

  private SvgDocument m_svg_document;
  public SvgElement(SvgDocument svg_document) {
    super();
    _children = new Vector();
    _width = "100";         // width and height are required attributes so
    _height = "100";        // initialize them to some default values        
    m_svg_document = svg_document;    
  }

  /**
   * Constructs a new SvgElement with the specified dimensions.
   * @param width The width of the svg element.
   * @param height The height of the svg element.
   */
  public SvgElement(String width, String height) {
    super();
    _children = new Vector();
    if (width != null)
      _width = width;
    else
      _width = "100";
    if (height != null)
      _height = height;
    else
      _height = "100";

  }

  /**
   * Add a child graphic to this svg element.
   * @param graphic The child graphic element to add.
   */
  public void addChild(SvgGraphic graphic) {
    if (graphic != null) {
      _children.addElement(graphic);
      graphic.setParent(this);
    }
  }

  /**
   * Sets the defs element.
   * @param defs The value to use when setting the defs element.
   */
  public void setDefs(DefsElement defs) {
    _defs = defs;
  }

  /**
   * Sets the title element.
   * @param title The value to use when setting the title element.
   */
  public void setTitle(TitleElement title) {
    _title = title;
  }

  /**
   * Sets the desc element.
   * @param desc The value to use when setting the desc element.
   */
  public void setDesc(DescElement desc) {
    _desc = desc;
  }

  /**
   * Sets the reference point to the specified location.
   * @param refX The X coordinate of the reference point.
   * @param refY The Y coordinate of the reference point.
   */
  public void setReferencePoint(float refX, float refY) {
    _refX = Float.toString(refX);
    _refY = Float.toString(refY);
  }


  /**
   * Returns the child graphic element at the specified index.
   * @param The index of the required child element.
   * @return The child graphic element at the specified index. Note: will be
   * null if the index is invalid.
   */
  public SvgGraphic getChild(int index) {
    if (index >= 0 && index < _children.size())
      return (SvgGraphic) _children.elementAt(index);
    return null;
  }

  /**
   * Returns the Graphic object that represents this svg element.
   * @return The Graphic object that represents this svg element.
   */
  public Graphic getGraphic() {
    setStyleProperties();
    Group group = new Group();
    int numChildren = _children.size();
    for (int i = 0; i < numChildren; i++) {
      SvgGraphic child = getChild(i);
      Graphic graphic = child.getGraphic();
      if (graphic != null)
        group.addElement(graphic);
    }

    // **** transform not part of <svg> element any more ****
    // Need to restructure the class heirachy, since all SvgStdGraphic classes
    // have a transform attribute

    // get transform matrix and apply to graphic
    // AffineTransform transformMatrix = getTransformMatrix(getTransform());
    //  if (transformMatrix != null)
    //    group.transform(transformMatrix);


    // translate to x and y coords (if there are any and this is not the root svg element)
    if (_x != null && _y != null && getParent() != null) {
      group.translate(Double.parseDouble(_x), Double.parseDouble(_y));
    }

    setGraphicStyle(group);
        
    return group;
  }

  /**
   * Returns the width (in pixels) of this svg elemnt.
   * @return The width (in pixels) of this svg element.
   */
  public int getWidth() {
    return getPixelDimension(_width);
  }

  /**
   * Returns the width (in pixels) of this svg elemnt.
   * @return The width (in pixels) of this svg element.
   */
  public int getHeight() {
    return getPixelDimension(_height);
  }

  /**
   * Returns the defs element.
   * @return The defs element.
   */
  public DefsElement getDefs() {
    return _defs;
  }

  /**
   * Returns the title element.
   * @return The title element.
   */
  public TitleElement getTitle() {
    return _title;
  }

  /**
   * Returns the desc element.
   * @return The desc element.
   */
  public DescElement getDesc() {
    return _desc;
  }



  // XmlElement stuff

  /**
   * Returns the tag that identifies an svg element in an SVG/XML file.
   * @return The tag that identifies an svg element in an SVG/XML file.
   */
  public String getTag() {
    return "svg";
  }

  /**
   * Reads in SvgElement data from the given XML Element.
   * @param elem The XML Element that contains data to be read.
   */
  public void read(Element elem) {

    // read the standard graphic attributes
    readStandardGraphicAttributes(elem);

    // set the attributes
    String att = elem.getAttribute("onunload");
    if (att.length() > 0 )
      _onunload = att;

    att = elem.getAttribute("onzoom");
    if (att.length() > 0)
      _onzoom = att;

    att = elem.getAttribute("x");
    if (att.length() > 0)
      _x = att;

    att = elem.getAttribute("y");
    if (att.length() > 0)
      _y = att;

    _width = elem.getAttribute("width");
    _height = elem.getAttribute("height");

    att = elem.getAttribute("refX");
    if (att.length() > 0)
      _refX = att;

    att = elem.getAttribute("refY");
    if (att.length() > 0)
      _refY = att;

    att = elem.getAttribute("viewBox");
    if (att.length() > 0)
      _viewBox = att;

    _preserveAspectRatio = elem.getAttribute("preserveAspectRatio");
    _allowZoomAndPan = elem.getAttribute("allowZoomAndPan");

    att = elem.getAttribute("contentScriptType");
    if (att.length() > 0)
      _contentScriptType = att;

    // read each child
    NodeList childNodes = elem.getChildNodes();
    int numChildren = childNodes.getLength();
    for (int i = 0; i < numChildren; i++) {
      Node child = childNodes.item(i);
      if (child.getNodeType() == Node.ELEMENT_NODE) { // it is an element
        Element childElement = (Element)child;
        String childTag = childElement.getTagName();

        if (childTag.equals("defs")) {
          _defs = new DefsElement(m_svg_document);
          _defs.read(childElement);

        } else if (childTag.equals("desc")) {
          _desc = new DescElement();
          _desc.read(childElement);

        } else if (childTag.equals("title")) {
          _title = new TitleElement();
          _title.read(childElement);

        } else if (childTag.equals("constraint")) {

          String constraint_string = childElement.getAttribute("rule");
          System.err.print("Rule: " + constraint_string);

          String strength = childElement.getAttribute("strength");
          System.err.println(", strength: " + strength);

          // take strength out of contraint_string
          final String REQUIRED_STRENGTH_STRING = "required";

          boolean ends_with_required_strength_string = strength.equals(REQUIRED_STRENGTH_STRING);

          final String STRONG_STRENGTH_STRING = "strong";

          boolean ends_with_strong_strength_string = strength.equals(STRONG_STRENGTH_STRING);

          final String MEDIUM_STRENGTH_STRING = "medium";

          boolean ends_with_medium_strength_string = strength.equals(MEDIUM_STRENGTH_STRING);

          final String WEAK_STRENGTH_STRING = "weak";

          boolean ends_with_weak_strength_string = strength.equals(WEAK_STRENGTH_STRING);

          boolean have_specified_strength	=  ends_with_required_strength_string
            || ends_with_strong_strength_string
            || ends_with_medium_strength_string
            || ends_with_weak_strength_string;

          java.io.StringReader string_reader = new java.io.StringReader(constraint_string);

          Yylex yylex = new Yylex(string_reader);

          if (m_svg_document.m_variable_name_object_hash == null)
            {
              System.err.println("SvgElement doesn't have an accessible hash.");
            }

          yylex.setVariableNameObjectHash(m_svg_document.m_variable_name_object_hash);

          parser constraint_parser = new parser(yylex);

          constraint_parser.setVariableNameObjectHash(m_svg_document.m_variable_name_object_hash);

          if (m_svg_document.m_variable_name_object_hash.isEmpty())
            {
              System.err.println("Variable name object hash is empty!");
            }

          try
            {
              java_cup.runtime.Symbol symbol = constraint_parser.parse();
						
              ClConstraint constraint = (ClConstraint) symbol.value;

              if (have_specified_strength)
                {
                  if (ends_with_required_strength_string)
                    {
                      constraint.changeStrength(ClStrength.required);
                    }
                  else if (ends_with_strong_strength_string)
                    {
                      constraint.changeStrength(ClStrength.strong);
                    }
                  else if (ends_with_medium_strength_string)
                    {
                      constraint.changeStrength(ClStrength.medium);
                    }
                  else if (ends_with_weak_strength_string)
                    {
                      constraint.changeStrength(ClStrength.weak);
                    }
                  else
                    {
                      throw new Error("Illegal Logic.");
                    }
                }
              System.err.println("Parsed as: " + symbol.value.toString());
              m_svg_document.addConstraint(constraint);
            }
          catch (Exception exception)
            {
              System.err.println("Exception: " + exception.getMessage());
              exception.printStackTrace();
            }
          System.err.println("");					
					
        } else if (childTag.equals("path")) {
          PathElement path = new PathElement(m_svg_document);
          path.read(childElement);
          addChild(path);

        } else if (childTag.equals("text")) {
          TextElement text = new TextElement(m_svg_document);
          text.read(childElement);
          addChild(text);

        } else if (childTag.equals("rect")) {
          RectElement rect = new RectElement(m_svg_document);
          rect.read(childElement);
          addChild(rect);

        } else if (childTag.equals("circle")) {
          /*
            System.err.println("Creating a circle element with a svg document");					
            if (m_svg_document == null)
            {
            System.err.println("But svg document is null!");
            }
          */					
          CircleElement circle = new CircleElement(m_svg_document);
          circle.read(childElement);
          addChild(circle);

        } else if (childTag.equals("ellipse")) {
          EllipseElement ellipse = new EllipseElement(m_svg_document);
          ellipse.read(childElement);
          addChild(ellipse);

        } else if (childTag.equals("line")) {
          LineElement line = new LineElement(m_svg_document);
          line.read(childElement);
          addChild(line);

        } else if (childTag.equals("polyline")) {
          PolylineElement polyline = new PolylineElement(m_svg_document);
          polyline.read(childElement);
          addChild(polyline);

        } else if (childTag.equals("polygon")) {
          PolygonElement polygon = new PolygonElement(m_svg_document);
          polygon.read(childElement);
          addChild(polygon);

        } else if (childTag.equals("use")) {
          UseElement use = new UseElement(m_svg_document);
          use.read(childElement);
          addChild(use);

        } else if (childTag.equals("image")) {
          ImageElement image = new ImageElement(m_svg_document);
          image.read(childElement);
          addChild(image);

        } else if (childTag.equals("svg")) {
          SvgElement svg = new SvgElement(m_svg_document);
          svg.read(childElement);
          addChild(svg);

        } else if (childTag.equals("g")) {
          GElement g = new GElement(m_svg_document);
          g.read(childElement);
          addChild(g);

        } else if (childTag.equals("switch")) {
          // <switch> element not yet supported

        } else if (childTag.equals("a")) {
          // <a> element not yet supported

        }
      }
    }
  }

  /**
   * Inserts data into the given Document.
   * The data will become a child of the given parent element.
   *
   * @param doc The Document to write to.
   * @param parent Indicates the parent element of the data to be written.
   */
  public void write(Document doc, Node parent) {

    // create new element for this svg element
    Element svg = doc.createElement(getTag());

    // set the standard graphic attributes
    setStandardGraphicAttributes(svg);

    // set the attributes
    if (_onunload != null)
      svg.setAttribute("onunload", _onunload);
    if (_onzoom != null)
      svg.setAttribute("onzoom", _onzoom);

    if (_x != null)
      svg.setAttribute("x", _x);
    if (_y != null)
      svg.setAttribute("y", _y);

    svg.setAttribute("width", _width);
    svg.setAttribute("height", _height);

    if (_refX != null)
      svg.setAttribute("refX", _refX);
    if (_refY != null)
      svg.setAttribute("refY", _refY);

    if (_viewBox != null)
      svg.setAttribute("viewBox", _viewBox);
    if (_preserveAspectRatio != null)
      svg.setAttribute("preserveAspectRatio", _preserveAspectRatio);
    if (_allowZoomAndPan != null)
      svg.setAttribute("allowZoomAndPan", _allowZoomAndPan);
    if (_contentScriptType != null)
      svg.setAttribute("contentScriptType", _contentScriptType);

    if (_defs != null)
      _defs.write(doc, svg);
    if (_desc != null)
      _desc.write(doc, svg);
    if (_title != null)
      _title.write(doc, svg);

    // write children elements
    int numChildren = _children.size();
    for (int i = 0; i < numChildren; i++) {
      SvgGraphic child = getChild(i);
      child.write(doc, svg);
    }

    // append new element to the parent element
    parent.appendChild(svg);

  }
}
/* Local Variables: */
/* tab-width: 4 */
