/*
 * Copyright (C) 1999
 * CSIRO Mathematical and Information Sciences Australia
 * Spatial Information Systems Project
 *
 * http://www.cmis.csiro.au/svg
 */


/**
 * SvgGraphic.java
 *
 * $Id: SvgGraphic.java,v 1.5 2000/05/06 21:40:58 gjb Exp $
 *
 */

package org.csiro.svgv.reader;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.csiro.svgv.graphic.*;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;
import java.awt.Paint;
import java.awt.Stroke;
import java.awt.BasicStroke;
import java.awt.Color;
import java.util.Vector;
import java.awt.GradientPaint;
import java.awt.Shape;
import java.awt.Font;
import java.awt.geom.AffineTransform;
import java.awt.geom.Rectangle2D;

/**
 * This is an abstract base class for all drawable SVG elements.
 * All extending classes need to implement a getGraphic() function.
 */
public abstract class SvgGraphic implements XmlElement {

  protected String _id = null;
  protected String _xmlLang = null;
  protected String _xmlSpace = null;
  protected String _class = null;
  protected String _style = null;

  protected BasicStroke _stroke = null;

  protected Font _font = null;
  protected String _textAnchor = null;
  protected Color _fillColor = null;
  protected float _fillOpacity;
  protected GradientElement _fillGradient = null;

  protected Color _lineColor = null;
  protected float _lineOpacity;
  protected GradientElement _lineGradient = null;

  protected SvgGraphic _parentGraphic = null;

  public static final String defaultFontFamily = "serif";
  public static final float defaultFontSize = 12f;
  public static final String defaultTextAnchor = "start";
  public static final int defaultFontStyle = Font.PLAIN;
  public static final int defaultFontWeight = Font.PLAIN;
  public static final Color defaultFillColor = null;
  public static final float defaultFillOpacity = 1f;
  public static final Color defaultLineColor = null;
  public static final float defaultLineOpacity = 1f;
  public static final float defaultStrokeWidth = 1f;
  public static final int defaultLineCap = BasicStroke.CAP_BUTT;
  public static final int defaultLineJoin = BasicStroke.JOIN_MITER;
  public static final float defaultMiterLimit = 8f;
  public static final float[] defaultDashArray = null;
  public static final float defaultDashOffset = 0f;


  public abstract Graphic getGraphic();

  public String getId() {
    return _id;
  }
  public String getSvgClass() {
    return _class;
  }
  public String getStyle() {
    return _style;
  }
  public void setId(String id) {
    _id = id;
  }
  public void setSvgClass(String svgClass) {
    _class = svgClass;
  }
  public void setStyle(String style) {
    _style = style;
  }
  protected void setParent(SvgGraphic parent) {
    _parentGraphic = parent;
  }
  protected SvgGraphic getParent() {
    return _parentGraphic;
  }

  public Font getFont() {
    return _font;
  }
  public String getTextAnchor() {
    return _textAnchor;
  }
  public BasicStroke getStroke() {
    return _stroke;
  }
  public Color getLineColor() {
    return _lineColor;
  }
  public float getLineOpacity() {
    return _lineOpacity;
  }
  public Color getFillColor() {
    return _fillColor;
  }
  public float getFillOpacity() {
    return _fillOpacity;
  }
  public GradientElement getFillGradient() {
    return _fillGradient;
  }
  public GradientElement getLineGradient() {
    return _lineGradient;
  }

  public SvgDocument m_svg_document;

  public void setStyleProperties(SvgDocument svg_document)
  {
    m_svg_document = svg_document;
    setStyleProperties();
    m_svg_document = null;
  }


  public void setStyleProperties() {

    String style = getStyle();

    boolean fontFamilySet = false;
    boolean fontSizeSet = false;
    boolean textAnchorSet = false;
    boolean fontStyleSet = false;
    boolean fontWeightSet = false;
    boolean fillColorSet = false;
    boolean fillOpacitySet = false;
    boolean lineColorSet = false;
    boolean lineOpacitySet = false;
    boolean strokeWidthSet = false;
    boolean lineCapSet = false;
    boolean lineJoinSet = false;
    boolean miterLimitSet = false;
    boolean dashArraySet = false;
    boolean dashOffsetSet = false;

    String fontFamily = defaultFontFamily;
    float fontSize = defaultFontSize;
    String textAnchor = defaultTextAnchor;
    int fontStyle = defaultFontStyle;
    int fontWeight = defaultFontWeight;
    Color fillColor = defaultFillColor;
    float fillOpacity = defaultFillOpacity;
    Color lineColor = defaultLineColor;
    float lineOpacity = defaultLineOpacity;
    float strokeWidth = defaultStrokeWidth;
    int lineCap = defaultLineCap;
    int lineJoin = defaultLineJoin;
    float miterLimit = defaultMiterLimit;
    float[] dashArray = defaultDashArray;
    float dashOffset = defaultDashOffset;

    if (style != null) {

      StringTokenizer st = new StringTokenizer(style, ";", false);

      while (st.hasMoreTokens()) {

        String styleSetting = st.nextToken();
        int colonIndex = styleSetting.indexOf(':');
        if (colonIndex > 0) {
          String property = styleSetting.substring(0,colonIndex);
          String value = styleSetting.substring(colonIndex+1);
          property = property.trim();
          value = value.trim();

          //  System.out.println("property: " + property + " " + value);

          if (property.equals("color")) {
            if (value.equals("inherit")) {
              if (getParent() != null) {
                fillColor = getParent().getFillColor();
                lineColor = getParent().getLineColor();
                fillColorSet = true;
                lineColorSet = true;
              }
            } else {
	                        // must be a color
              if (!fillColorSet)
                fillColor = getColor(value);
              if (!lineColorSet)
                lineColor = getColor(value);
              fillColorSet = true;
              lineColorSet = true;
            }

          }

          else if (property.equals("fill")) {
            if (value.equals("none")) {
              fillColor = null;
              fillColorSet = true;

            } else if (value.equals("current-color")) {
              // don't do anything, it will be set by the color property

            } else if (value.indexOf("url") != -1) {
              // is a reference to something in the defs sections
              int hashIndex = value.indexOf('#');
              if (hashIndex != -1) {
                String id = value.substring(hashIndex+1, value.length()-1);
                GradientElement gradientPaint = null;
                                // find defs section that has an element with this id
                SvgGraphic parent = getParent();
                while (parent != null) {
                  // does parent have a defs section?
                  DefsElement defs = null;
                  if (parent instanceof SvgElement) {
                    defs = ((SvgElement)parent).getDefs();
                  } else if (parent instanceof GElement) {
                    defs = ((GElement)parent).getDefs();
                  } else if (parent instanceof SymbolElement) {
                    defs = ((SymbolElement)parent).getDefs();
                  }
                  if (defs != null) {
                    Definable ref = defs.getDefinedElement(id);
                    if (ref != null && ref instanceof GradientElement) {
                      _fillGradient = (GradientElement)ref;
                      fillColor = null;
                      fillColorSet = true;
                      break;
                    }
                  }
                  parent = parent.getParent();
                }
              }
            } else if (value.equals("inherit")) {
              if (getParent() != null) {
                fillColor = getParent().getFillColor();
                fillColorSet = true;
              }
            } else {
	                        // must be a color
              fillColor = getColor(value);
              fillColorSet = true;
            }
          }

          else if (property.equals("fill-opacity")) {
            try {
              int index = value.indexOf('%');
              if (index == -1) { // should be a number between 0 and 1
                fillOpacity = Float.valueOf(value).floatValue();
              } else { // must be percentage
                String percentage = value.substring(0,index);
                fillOpacity = Float.valueOf(percentage).floatValue() / 100;
              }
              fillOpacitySet = true;
            } catch (NumberFormatException e) {
              System.out.println("cannot decode fill-opacity: " + value);
            }
          }

          else if (property.equals("stroke")) {

            if (value.equals("none")) {
              lineColor = null;
              lineColorSet = true;

            } else if (value.equals("current-color")) {
              // don't do anything, it will be set by the color attribute

            } else if (value.indexOf("url") != -1) {
              // is a reference to something in the defs sections,
              int hashIndex = value.indexOf('#');
              if (hashIndex != -1) {
                String id = value.substring(hashIndex+1, value.length()-1);
                GradientElement gradientPaint = null;
                                // find defs section that has an element with this id
                SvgGraphic parent = getParent();
                while (parent != null) {
                  // does parent have a defs section?
                  DefsElement defs = null;
                  if (parent instanceof SvgElement) {
                    defs = ((SvgElement)parent).getDefs();
                  } else if (parent instanceof GElement) {
                    defs = ((GElement)parent).getDefs();
                  } else if (parent instanceof SymbolElement) {
                    defs = ((SymbolElement)parent).getDefs();
                  }
                  if (defs != null) {
                    Definable ref = defs.getDefinedElement(id);
                    if (ref != null && ref instanceof GradientElement) {
                      _lineGradient = (GradientElement)ref;
                      lineColor = null;
                      lineColorSet = true;
                      break;
                    }
                  }
                  parent = parent.getParent();
                }
              }

            } else if (value.equals("inherit")) {
              if (getParent() != null) {
                lineColor = getParent().getLineColor();
                lineColorSet = true;
              }

            } else {
              // must be a color
              lineColor = getColor(value);
              lineColorSet = true;
            }
          }

          else if (property.equals("stroke-width")) {
            if (value.equals("inherit")) {
              if (getParent() != null && getParent().getStroke() != null) {
                strokeWidth = getParent().getStroke().getLineWidth();
                strokeWidthSet = true;
              }
            } else {
              try {
                strokeWidth = Float.valueOf(value).floatValue();
                strokeWidthSet = true;
              } catch (NumberFormatException e) {
                System.out.println("cannot decode stroke width: " + value);
              }
            }
          }

          else if (property.equals("stroke-linecap")) {
            if (value.equals("butt")) {
              lineCap = BasicStroke.CAP_BUTT;
              lineCapSet = true;
            } else if (value.equals("round")) {
              lineCap = BasicStroke.CAP_ROUND;
              lineCapSet = true;
            } else if (value.equals("square")) {
              lineCap = BasicStroke.CAP_SQUARE;
              lineCapSet = true;
            } else if (value.equals("inherit")) {
              if (getParent() != null && getParent().getStroke() != null) {
                lineCap = getParent().getStroke().getEndCap();
                lineCapSet = true;
              }
            }
          }

          else if (property.equals("stroke-linejoin")) {
            if (value.equals("miter")) {
              lineJoin = BasicStroke.JOIN_MITER;
              lineJoinSet = true;
            } else if (value.equals("round")) {
              lineJoin = BasicStroke.JOIN_ROUND;
              lineJoinSet = true;
            } else if (value.equals("bevel")) {
              lineJoin = BasicStroke.JOIN_BEVEL;
              lineJoinSet = true;
            } else if (value.equals("inherit")) {
              if (getParent() != null && getParent().getStroke() != null) {
                lineJoin = getParent().getStroke().getLineJoin();
                lineJoinSet = true;
              }
            }
          }

          else if (property.equals("stroke-miterlimit")) {
            if (value.equals("inherit")) {
              if (getParent() != null && getParent().getStroke() != null) {
                miterLimit = getParent().getStroke().getMiterLimit();
                miterLimitSet = true;
              }
            } else {
              try {
                miterLimit = Float.valueOf(value).floatValue();
                miterLimitSet = true;
              } catch (NumberFormatException e) {
                System.out.println("cannot decode miterlimit: " + value);
              }
            }
          }

          else if (property.equals("stroke-dasharray")) {
            if (value.equals("inherit")) {
              if (getParent() != null && getParent().getStroke() != null) {
                dashArray = getParent().getStroke().getDashArray();
                dashArraySet = true;
              }
            } else {
              StringTokenizer st1 = new StringTokenizer(value, " ,", false);
              Vector dashVector = new Vector();
              while (st1.hasMoreTokens()) {
                String token = st1.nextToken();
                int length = getPixelDimension(token);
                dashVector.addElement(new Float(length));
              }
              int numElements = dashVector.size();
              if (numElements > 0) {
                dashArray = new float[numElements];
                for (int i = 0; i < numElements; i++) {
                  dashArray[i] = ((Float)dashVector.elementAt(i)).floatValue();
                }
                dashArraySet = true;
              }
            }
          }

          else if (property.equals("stroke-dashoffset")) {
            if (value.equals("inherit")) {
              if (getParent() != null && getParent().getStroke() != null) {
                dashOffset = getParent().getStroke().getDashPhase();
                dashOffsetSet = true;
              }
            } else {
              try {
                dashOffset = Float.valueOf(value).floatValue();
                dashOffsetSet = true;
              } catch (NumberFormatException e) {
                System.out.println("cannot decode dashoffset: " + value);
              }
            }
          }

          else if (property.equals("stroke-opacity")) {
            if (value.equals("inherit")) {
              if (getParent() != null) {
                lineOpacity = getParent().getLineOpacity();
                lineOpacitySet = true;
              }
            } else {
              try {
                int index = value.indexOf('%');
                if (index == -1) { // should be a number between 0 and 1
                  lineOpacity = Float.valueOf(value).floatValue();
                } else {   // must be percentage
                  String percentage = value.substring(0,index);
                  lineOpacity = Float.valueOf(percentage).floatValue() / 100;
                }
                lineOpacitySet = true;
              } catch (NumberFormatException e) {
                System.out.println("cannot decode strokeOpacity: " + value);
              }
            }

          }

          else if (property.equals("opacity")) {
            // set both line and fill opacity values
            try {
              int index = value.indexOf('%');

              if (index == -1) { // should be a number between 0 and 1
                float opacity = Float.valueOf(value).floatValue();
                if (!lineOpacitySet)
                  lineOpacity = opacity;
                if (!fillOpacitySet)
                  fillOpacity = opacity;

              } else {   // must be percentage
                String percentage = value.substring(0,index);
                float opacity = Float.valueOf(percentage).floatValue() / 100;
                if (!lineOpacitySet)
                  lineOpacity = opacity;
                if (!fillOpacitySet)
                  fillOpacity = opacity;
              }
              lineOpacitySet = true;
              fillOpacitySet = true;
            } catch (NumberFormatException e) {
              System.out.println("cannot decode opacity: " + value);
            }
          }

          else if (property.equals("font-size")) {
            try {

              String looked_up_value = value;

              if (m_svg_document != null)
                {
                  looked_up_value = m_svg_document.handlePossibleVariable(value);
                }

              fontSize = Float.valueOf(looked_up_value).floatValue();
              fontSizeSet = true;
            } catch (NumberFormatException e) {
              System.out.println("cannot decode font-size: " + value);
              if (m_svg_document == null)
                System.err.println("m_svg_document == null");
            }
          }
          else if (property.equals("text-anchor")) {
            textAnchor = value;
            textAnchorSet = true;
          }

          else if (property.equals("font-family")) {
            fontFamily = value;
            fontFamilySet = true;
          }

          else if (property.equals("font-style")) {
            if (value.equalsIgnoreCase("italic") || value.equalsIgnoreCase("oblique")) {
              fontStyle = Font.ITALIC;
              fontStyleSet = true;
            }
          }

          else if (property.equals("font-weight")) {
            if (value.equalsIgnoreCase("bold") || value.equalsIgnoreCase("bolder")) {
              fontWeight = Font.BOLD;
              fontWeightSet = true;
            }
          }
        }
      }
    }

    // I'm not sure if this bit should be here
    // I think inheritence needs to be defined a bit clearer in the spec
    SvgGraphic parent = getParent();
    if (parent != null) {
      if (!fontFamilySet)
        fontFamily = parent.getFont().getFamily();
      if (!fontSizeSet)
        fontSize = parent.getFont().getSize();
      if (!textAnchorSet)
        textAnchor = parent.getTextAnchor();
      if (!fontStyleSet) {
        boolean italic = parent.getFont().isItalic();
        if (italic)
          fontStyle = Font.ITALIC;
        else
          fontStyle = Font.PLAIN;
      }
      if (!fontWeightSet) {
        boolean bold = parent.getFont().isBold();
        if (bold)
          fontWeight = Font.BOLD;
        else
          fontStyle = Font.PLAIN;
      }
      if (!fillColorSet)
        fillColor = parent.getFillColor();
      if (!fillOpacitySet)
        fillOpacity = parent.getFillOpacity();
      if (!lineColorSet)
        lineColor = parent.getLineColor();
      if (!lineOpacitySet)
        lineOpacity = parent.getLineOpacity();
      if (!strokeWidthSet)
        strokeWidth = parent.getStroke().getLineWidth();
      if (!lineCapSet)
        lineCap = parent.getStroke().getEndCap();
      if (!lineJoinSet)
        lineJoin = parent.getStroke().getLineJoin();
      if (!miterLimitSet)
        miterLimit = parent.getStroke().getMiterLimit();
      if (!dashArraySet)
        dashArray = parent.getStroke().getDashArray();
      if (!dashOffsetSet)
        dashOffset = parent.getStroke().getDashPhase();
    }

    _textAnchor = textAnchor;
    _fillColor = fillColor;
    _fillOpacity = fillOpacity;
    _lineColor = lineColor;
    _lineOpacity = lineOpacity;
    _stroke = new BasicStroke(strokeWidth, lineCap, lineJoin, miterLimit, dashArray, dashOffset);
    _font = new Font(fontFamily, fontStyle+fontWeight, 1);
    _font = _font.deriveFont(fontSize);

  }


  protected void setGraphicStyle(Graphic graphic) {

    graphic.setStroke(_stroke);
    graphic.setFont(_font);

    if (!(graphic instanceof Path || graphic instanceof Text)) {

      Color fillPaint = null;
      if (_fillColor != null)
        fillPaint = new Color(_fillColor.getRed(), _fillColor.getGreen(),
                              _fillColor.getBlue(), (int)(_fillOpacity*255));
      graphic.setFillPaint(fillPaint);

      Color linePaint = null;
      if (_lineColor != null)
        linePaint = new Color(_lineColor.getRed(), _lineColor.getGreen(),
                              _lineColor.getBlue(), (int)(_lineOpacity*255));
      graphic.setLinePaint(linePaint);

    } else {  // graphic is a Path or Text so there may be a gradient to use somewhere

      if (_fillColor != null) {
        Color fillPaint = new Color(_fillColor.getRed(), _fillColor.getGreen(),
                                    _fillColor.getBlue(), (int)(_fillOpacity*255));
        graphic.setFillPaint(fillPaint);

      } else {
        // see if there is a gradient to be used
        Paint gradientPaint = null;
        if (_fillGradient != null) {
          gradientPaint = _fillGradient.getPaint(getUserBounds(), graphic.getMBR());
        } else {
          SvgGraphic parent = getParent();
          while (parent != null) {
            if (parent.getFillGradient() != null) {
              gradientPaint = parent.getFillGradient().getPaint(getUserBounds(),graphic.getMBR());
              break;
            }
            parent = parent.getParent();
          }
        }
        graphic.setFillPaint(gradientPaint);
      }

      if (_lineColor != null) {
        Color linePaint = new Color(_lineColor.getRed(), _lineColor.getGreen(),
                                    _lineColor.getBlue(), (int)(_lineOpacity*255));
        graphic.setLinePaint(linePaint);
      } else {
        // see if there is a gradient to be used
        Paint gradientPaint = null;
        if (_lineGradient != null) {
          gradientPaint = _lineGradient.getPaint(getUserBounds(), graphic.getMBR());
        } else {
          SvgGraphic parent = getParent();
          while (parent != null) {
            if (parent.getLineGradient() != null) {
              gradientPaint = parent.getLineGradient().getPaint(getUserBounds(), graphic.getMBR());
              break;
            }
            parent = parent.getParent();
          }
        }
        graphic.setLinePaint(gradientPaint);
      }
    }
  }


  /**
   * Converts the length into a pixel dimension.
   * @param length The string containing the dimension the be converted.
   * @return The dimension of length in pixels.
   */
  public static int getPixelDimension(String length) {
    // find index of first non-digit
    String len = length.trim();
    int index = -1;
    for (int i = 0; i < len.length(); i++) {
      if (!Character.isDigit(len.charAt(i))) {
        index = i;
        break;
      }
    }
    if (index == -1) {  // no unit specified
      return (Integer.valueOf(len).intValue());
    } else {
      // note: where an absolute measure is given, eg. cm or in, the
      // value returned will probably be incorrect.  They should be
      // approximately correct for a 19 inch monitor running at 1280x1024.
      String unit = len.substring(index);
      float value = Float.valueOf(len.substring(0,index)).floatValue();
      if (unit.equals("px")) {
        return (int)value;
      } else if (unit.equals("cm")) {
        return (int) (value * 33.7);
      } else if (unit.equals("mm")) {
        return (int) (value * 3.37);
      } else if (unit.equals("in")) {
        return (int) (value * 33.7 * 2.54);
      } else if (unit.equals("pt")) {
        return (int) (value * 33.7 * 2.54 / 72.0);
      } else if (unit.equals("pc")) {
        return (int) (value * 33.7 * 2.54 / 72.0 * 12.0);
      } else {
        // not sure what other units mean
        // just return the value
        return (int)value;
      }
    }
  }


  /**
   * Returns a Clor object that represents the color string specified.
   * @param color A string representation of the color required. Can be in one
   * of the following formats: #f00, #ff0000, rgb(255,0,0), rgb(100%, 0%,0%), red.
   * @return The Color object required.
   */
  public static Color getColor(String color) {
    color = color.trim();

    if (color.indexOf('#') == 0) {

      // if the color specification is seven characters, then we
      // can call decode directly
      if (color.length() == 7) {
        try {
          return Color.decode(color);
        } catch (NumberFormatException e) {}
      } else if (color.length() == 4) {
        StringBuffer newcolor = new StringBuffer(color);
        newcolor.insert(1, color.charAt(1));
        newcolor.insert(3, color.charAt(2));
        newcolor.insert(6, color.charAt(3));
        try {
          return Color.decode(newcolor.toString());
        } catch (NumberFormatException e) {}
      }


    } else if (color.indexOf("rgb") == 0) {

      int bracket1index = color.indexOf('(');
      int bracket2index = color.indexOf(')');
      String temp = color.substring(bracket1index+1, bracket2index);
      StringTokenizer st = new StringTokenizer(temp, ", ", false);
      String r;
      String g;
      String b;
      try {
        r = st.nextToken();
        g = st.nextToken();
        b = st.nextToken();
      } catch (NoSuchElementException e) {
        System.out.println("Could not decode color: " + color);
        return Color.black;
      }
      int rValue;
      int gValue;
      int bValue;

      if (r.indexOf('%') == -1) { // it is a number between 0 and 255
        rValue = Integer.valueOf(r).intValue();
      } else {  // it a percentage between 0 and 100
        int percentIndex = r.indexOf('%');
        String tempR = r.substring(0,percentIndex);
        rValue = (int)(Float.valueOf(tempR).floatValue() * 255f / 100f );
      }
      if (g.indexOf('%') == -1) { // it is a number between 0 and 255
        gValue = Integer.valueOf(g).intValue();
      } else {  // it a percentage between 0 and 100
        int percentIndex = g.indexOf('%');
        String tempG = g.substring(0,percentIndex);
        gValue = (int)(Float.valueOf(tempG).floatValue() * 255f / 100f );
      }
      if (b.indexOf('%') == -1) { // it is a number between 0 and 255
        bValue = Integer.valueOf(b).intValue();
      } else {  // it a percentage between 0 and 100
        int percentIndex = b.indexOf('%');
        String tempB = b.substring(0,percentIndex);
        bValue = (int)(Float.valueOf(tempB).floatValue() * 255f / 100f );
      }

      return new Color(rValue, gValue, bValue);


    } else {

      if (color.equalsIgnoreCase("black"))
        return Color.decode("#000000");
      if (color.equalsIgnoreCase("silver"))
        return Color.decode("#C0C0C0");
      if (color.equalsIgnoreCase("gray"))
        return Color.decode("#808080");
      if (color.equalsIgnoreCase("white"))
        return Color.decode("#FFFFFF");
      if (color.equalsIgnoreCase("maroon"))
        return Color.decode("#800000");
      if (color.equalsIgnoreCase("red"))
        return Color.decode("#FF0000");
      if (color.equalsIgnoreCase("purple"))
        return Color.decode("#800080");
      if (color.equalsIgnoreCase("fuchsia"))
        return Color.decode("#FF00FF");
      if (color.equalsIgnoreCase("green"))
        return Color.decode("#008000");
      if (color.equalsIgnoreCase("lime"))
        return Color.decode("#00FF00");
      if (color.equalsIgnoreCase("olive"))
        return Color.decode("#808000");
      if (color.equalsIgnoreCase("yellow"))
        return Color.decode("#FFFF00");
      if (color.equalsIgnoreCase("navy"))
        return Color.decode("#000080");
      if (color.equalsIgnoreCase("blue"))
        return Color.decode("#0000FF");
      if (color.equalsIgnoreCase("teal"))
        return Color.decode("#008080");
      if (color.equalsIgnoreCase("aqua"))
        return Color.decode("#00FFFF");
    }
    System.out.println("Could not decode color: " + color);
    return Color.black;
  }

  /**
   * Returns the bounding box of this graphic's parent svg element.
   * @return The bounding box of this graphic's parent svg element.
   */
  protected Rectangle2D getUserBounds() {
    if (this instanceof SvgElement) {
      return new Rectangle2D.Float(0,0,(float)((SvgElement)this).getWidth(),
                                   (float)((SvgElement)this).getHeight());
    } else {
      // find parent svg element
      SvgGraphic parent = getParent();
      while (parent != null) {
        if (parent instanceof SvgElement) {
          return new Rectangle2D.Float(0,0,(float)((SvgElement)parent).getWidth(),
                                       (float)((SvgElement)parent).getHeight());
        }
        parent = parent.getParent();
      }
    }
    return new Rectangle2D.Float();  // should never get here
  }



  /**
   * Parses the transform attribute value and returns an appropriate
   * AffineTransform object that can be applied to this graphic.
   * @param transformString The string that contains the transformation commands.
   * @return The AffineTransform that represents the transfromString.
   */
  public static AffineTransform getTransformMatrix(String transformString) {

    if (transformString == null)
      return null;

    transformString = transformString.trim();
    AffineTransform transformMatrix = new AffineTransform();
    StringTokenizer st = new StringTokenizer(transformString, "()", false);

    while (st.hasMoreTokens()) {

      String transformType = st.nextToken().trim();
      if (!st.hasMoreTokens()) break;
      String transformArgs = st.nextToken().trim();

      if (transformType.equals("matrix")) {
        StringTokenizer st1 = new StringTokenizer(transformArgs, ", ", false);
        int numArgs = st1.countTokens();
        if (numArgs == 6) {
          double a1 = Double.valueOf(st1.nextToken()).doubleValue();;
          double a2 = Double.valueOf(st1.nextToken()).doubleValue();;
          double a3 = Double.valueOf(st1.nextToken()).doubleValue();;
          double a4 = Double.valueOf(st1.nextToken()).doubleValue();;
          double a5 = Double.valueOf(st1.nextToken()).doubleValue();;
          double a6 = Double.valueOf(st1.nextToken()).doubleValue();;
          AffineTransform transform
            = new AffineTransform(a1, a2, a3, a4, a5, a6);
          transformMatrix.concatenate(transform);
        }
      } else if (transformType.equals("translate")) {
        StringTokenizer st1 = new StringTokenizer(transformArgs, ", ", false);
        int numArgs = st1.countTokens();
        double tx = 0;
        double ty = 0;
        if (numArgs == 1) {
          tx = Double.valueOf(st1.nextToken()).doubleValue();
        } else if (numArgs == 2) {
          tx = Float.valueOf(st1.nextToken()).doubleValue();
          ty = Float.valueOf(st1.nextToken()).doubleValue();
        }
        transformMatrix.translate(tx, ty);

      } else if (transformType.equals("scale")) {
        StringTokenizer st1 = new StringTokenizer(transformArgs, ", ", false);
        int numArgs = st1.countTokens();
        double sx = 0;
        double sy = 0;
        if (numArgs == 1) {
          sx = Double.valueOf(st1.nextToken()).doubleValue();
          sy = sx;
        } else if (numArgs == 2) {
          sx = Float.valueOf(st1.nextToken()).doubleValue();
          sy = Float.valueOf(st1.nextToken()).doubleValue();
        }
        transformMatrix.scale(sx, sy);

      } else if (transformType.equals("rotate")) {
        float angle = Float.valueOf(transformArgs).floatValue();
        transformMatrix.rotate(Math.toRadians(angle));

      } else if (transformType.equals("skewX")) {
        float skewAngle = Float.valueOf(transformArgs).floatValue();
        AffineTransform skewTransform
          = new AffineTransform(1, 0, Math.tan(Math.toRadians(skewAngle)), 1, 0, 0);
        transformMatrix.concatenate(skewTransform);

      } else if (transformType.equals("skewY")) {
        float skewAngle = Float.valueOf(transformArgs).floatValue();
        AffineTransform skewTransform
          = new AffineTransform(1, Math.tan(Math.toRadians(skewAngle)), 0, 1, 0, 0);
        transformMatrix.concatenate(skewTransform);

      } else {
        System.out.println("Invavid transform: " + transformType + "(" + transformArgs + ")" );
      }
    }
    return transformMatrix;
  }

}
