/*
 * Copyright (C) 1999
 * CSIRO Mathematical and Information Sciences Australia
 * Spatial Information Systems Project
 *
 * http://www.cmis.csiro.au/svg
 */


/**
 * TextElement.java
 *
 * $Id: TextElement.java,v 1.5 2000/05/06 21:41:31 gjb Exp $
 *
 */

package org.csiro.svgv.reader;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.csiro.svgv.graphic.*;
import java.awt.geom.AffineTransform;
import java.awt.Font;
import java.awt.font.*;
import java.awt.geom.Rectangle2D;

/**
 * The TextElement class describes an SVG text element.
 */
public class TextElement extends SvgStdGraphic implements Definable {

    // default attribute values
    private static final String defaultX = "0";
    private static final String defaultY = "0";

    // text attributes
    private String _text = null;
    private String _x = null;
    private String _y = null;

    /**
     * Constructs a new TextElement.
     */
    public TextElement() {
        super();
    }

    private SvgDocument m_svg_document;
    public TextElement(SvgDocument svg_document)
    {
	this();
	
	m_svg_document = svg_document;
    }


    /**
     * Constructs a new TextElement initializing the display text to the
     * specified string.
     * @param text The text to display.
     */
    public TextElement(String text) {
        _text = text;
    }

    /**
     * Constructs a new TextElement initializing the display text and location
     * to the spcified values.
     * @param text The text to display.
     * @param x The X coordinate of the text to display.
     * @param y The Y coordinate of the text to display.
     */
    public TextElement(String text, String x, String y) {
        _text = text;
        _x = x;
        _y = y;
    }

    /**
     * Constructs a new TextElement initializing the display text and location
     * to the spcified values.
     * @param text The text to display.
     * @param x The X coordinate of the text to display.
     * @param y The Y coordinate of the text to display.
     */
    public TextElement(String text, float x, float y) {
        _text = text;
        _x = Float.toString(x);
        _y = Float.toString(y);
    }

    /**
     * Returns the text to display for this text element.
     * @return The text to display for this text element.
     */
    public String getText() {
        if (_text == null)
            return "";
        return _text;
    }

    /**
     * Returns the X coordinate of the text's location.
     * @return The X coordinate of the text's location.
     */
    public String getX() {
        if (_x != null)
            return m_svg_document.handlePossibleVariable(_x);
        return defaultX;
    }

    /**
     * Returns the Y coordinate of the text's location.
     * @return The Y coordinate of the text's location.
     */
    public String getY() {
        if (_y != null)
            return m_svg_document.handlePossibleVariable(_y);
        return defaultY;
    }

    /**
     * Sets the text to display to the specified string.
     * @param text The text to display.
     */
    public void setText(String text) {
        _text = text;
    }

    /**
     * Sets the X coordinate of the text's location to the specified value.
     * @param x The X coordinate of the text's location.
     */
    public void setX(String x) {
        _x = x;
    }
    /**
     * Sets the Y coordinate of the text's location to the specified value.
     * @param y The Y coordinate of the text's location.
     */
    public void setY(String y) {
        _y = y;
    }

    /**
     * Sets the X coordinate of the text's location to the specified value.
     * @param x The X coordinate of the text's location.
     */
    public void setX(float x) {
        _x = Float.toString(x);
    }

    /**
     * Sets the Y coordinate of the text's location to the specified value.
     * @param y The Y coordinate of the text's location.
     */
    public void setY(float y) {
      _y = Float.toString(y);
    }

    /**
     * Sets the text's location to the specified location.
     * @param x The X coordinate of the text's location.
     * @param y The Y coordinate of the text's location.
     */
    public void setLocation(float x, float y) {
      _x = Float.toString(x);
      _y = Float.toString(y);
    }

    protected double anchorToOffset(String anchor, double width) {
      if (anchor.equals("start"))
        return 0;
      else if (anchor.equals("middle"))
        return width/2;
      else if (anchor.equals("end"))
        return width;
      else 
        return 0;
    }

    /**
     * Returns the Graphic object that represents this text element.
     * @return The Graphic object that represents this text element.
     */
    public Graphic getGraphic() {

      setStyleProperties(m_svg_document);

      Text text = new Text();
      text.setX(Double.parseDouble(getX()));
      text.setY(Double.parseDouble(getY()));
      text.setString(getText());
      text.setFont(_font);
      text.createShape();
      TextLayout tl = text.getLayout();
      Rectangle2D bounds = tl.getBounds();
      String anchor = getTextAnchor();
      
      System.err.println("Text has width = " + bounds.getWidth() + ", text-anchor =" + anchor);
      double dx = anchorToOffset(anchor,bounds.getWidth());

      text.translate(Double.parseDouble(getX()) - dx, Double.parseDouble(getY()));

      // get transform matrix and apply to graphic
      AffineTransform transformMatrix = getTransformMatrix(getTransform());
      if (transformMatrix != null)
        text.transform(transformMatrix);

      setGraphicStyle(text);

      return text;
    }

     /**
     * Returns the Graphic object that represents this text element.
     * @return The Graphic object that represents this text element.
     */
    public Graphic getGraphic(Font font) {

      setStyleProperties();

      Text text = new Text();
      text.setString(getText());
      text.setFont(font);
      text.createShape();

      text.translate(Double.parseDouble(getX()), Double.parseDouble(getY()));

      // get transform matrix and apply to graphic
      AffineTransform transformMatrix = getTransformMatrix(getTransform());
      if (transformMatrix != null)
        text.transform(transformMatrix);

      setGraphicStyle(text);

      return text;
    }


    // XmlElement stuff

    /**
     * Returns the tag that identifies a text element in an SVG/XML file.
     * @return The tag that identifies a text element in an SVG/XML file.
     */
    public String getTag() {
        return "text";
    }

    /**
     * Reads in TextElement data from the given XML Element.
     * @param elem The XML Element that contains data to be read.
     */
    public void read(Element elem) {

        // read the standard graphic attributes
        readStandardGraphicAttributes(elem);

        // set the attributes
        String att = elem.getAttribute("x");
        if (att.length() > 0)
            _x = att;

        att = elem.getAttribute("y");
        if (att.length() > 0)
            _y = att;

        // read each child
        NodeList childNodes = elem.getChildNodes();
        int numChildren = childNodes.getLength();
        for (int i = 0; i < numChildren; i++) {
            Node child = childNodes.item(i);

            if (child.getNodeType() == Node.TEXT_NODE) { // it is #PCDATA
                if (_text == null)
                    _text = "";
                _text += child.getNodeValue();

            } else if (child.getNodeType() == Node.ELEMENT_NODE) { // it is an element
                Element childElement = (Element)child;
                String childTag = childElement.getTagName();

                if (childTag.equals("tspan")) {
                    // <tspan> element not yet supported

                } else if (childTag.equals("textPath")) {
                   // <textPath> element not yet supported
                }
            }
        }

    }

    /**
     * Inserts data into the given Document.
     * The data will become a child of the given parent element.
     *
     * @param doc The Document to write to.
     * @param parent Indicates the parent element of the data to be written.
     */
    public void write(Document doc, Node parent) {

        // create new element for this text element
        Element text = doc.createElement(getTag());

         // set the standard graphic attributes
        setStandardGraphicAttributes(text);

        // set attributes
        if (_x != null)
            text.setAttribute("x", _x);
        if (_y != null)
            text.setAttribute("y", _y);

        // set text to display
        if (_text != null)
            text.appendChild(doc.createTextNode(_text));

        // append new element to the parent
        parent.appendChild(text);
    }

}

