/*
 * Copyright (C) 1999
 * CSIRO Mathematical and Information Sciences Australia
 * Spatial Information Systems Project
 *
 * http://www.cmis.csiro.au/svg
 */


/**
 * TitleElement.java
 *
 * $Id: TitleElement.java,v 1.1.1.1 1999/12/11 18:51:28 gjb Exp $
 *
 */

package org.csiro.svgv.reader;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * The TitleElement class describes an SVG title element.
 */
public class TitleElement implements XmlElement {

    // title attributes
    private String _titleText = null;
    private String _class = null;
    private String _id = null;
    private String _lang = null;
    private String _style = null;
    private static final String _content = "structured text";

    /**
     * Constructs a new TitleElement.
     */
    public TitleElement() {
    }

    /**
     * Constructs a new TitleElement initializing the title text with the
     * specified string.
     * @param titleText The string to use when initializing the title text.
     */
    public TitleElement(String titleText) {
        _titleText = titleText;
    }

    /**
     * Sets the title text to the specified string.
     * @param titleText The string to use when setting the title text.
     */
    public void setTitleText(String titleText) {
         _titleText = titleText;
    }

    /**
     * Returns the title text.
     * @return The title text.
     */
    public String getTitleText() {
        return _titleText;
    }


    // XmlElement stuff

    /**
     * Returns the tag that identifies a title element in an SVG/XML file.
     * @return The tag that identifies a title element in an SVG/XML file.
     */
    public String getTag() {
        return "title";
    }

    /**
     * Reads in TitleElement data from the given XML Element.
     * @param elem The XML Element that contains data to be read.
     */
    public void read(Element elem) {

        // read text for title
        if (elem.getFirstChild() != null) {
            _titleText = elem.getFirstChild().getNodeValue();
        }

        // read attributes
        String att = elem.getAttribute("xml:lang");
        if (att.length() > 0)
            _lang = att;

        att = elem.getAttribute("id");
        if (att.length() > 0)
            _id = att;

        att = elem.getAttribute("class");
        if (att.length() > 0)
            _class = att;

        att = elem.getAttribute("style");
        if (att.length() > 0)
            _style = att;


    }

    /**
     * Inserts data into the given Document.
     * The data will become a child of the given parent element.
     *
     * @param doc The Document to write to.
     * @param parent Indicates the parent element of the data to be written.
     */
    public void write(Document doc, Node parent) {

        // create new element for this title element
        Element title = doc.createElement(getTag());

        // set attributes
        if (_class != null)
            title.setAttribute("class", _class);
        if (_id != null)
            title.setAttribute("id", _id);
        if (_lang != null)
            title.setAttribute("lang", _lang);
        if (_style != null)
            title.setAttribute("style", _style);

        // set text for title
        if (_titleText != null)
            title.appendChild(doc.createTextNode(_titleText));

        // append new element to the parent
        parent.appendChild(title);
    }
}