/*
 * Copyright (C) 1999
 * CSIRO Mathematical and Information Sciences Australia
 * Spatial Information Systems Project
 *
 * http://www.cmis.csiro.au/svg
 */


/**
 * UseElement.java
 *
 * $Id: UseElement.java,v 1.2 2000/05/04 23:24:22 jwnichls Exp $
 *
 */

package org.csiro.svgv.reader;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.csiro.svgv.graphic.*;
import java.awt.geom.AffineTransform;

/**
 * The UseElement class describes an SVG use element.
 */
public class UseElement extends SvgStdGraphic implements Definable {

    private DescElement _desc = null;
    private TitleElement _title = null;

    // use attributes
    private String _x = null;
    private String _y= null;
    private String _width = null;
    private String _height = null;
    private String _href;

    /**
     * Constructs a new UseElement with the href attribute initialised to "".
     */
    public UseElement() {
        super();
        _href = ""; // href is a required attribute so initialize it to a default
    }

    private SvgDocument m_svg_document;
    public UseElement(SvgDocument svg_document) {
	this();
	m_svg_document = svg_document;
    }

    /**
     * Constructs a new UseElement with the specified href attribute value.
     * @param href The value to use when initialising the href attribute.
     */
    public UseElement(String href) {
        super();
        if (href != null)
            _href = href;
        else
            _href = "";
    }

    /**
     * Constructs and initialises a UseElement with the specified href,
     * location and dimension attribute values.
     * @param href The value to use when initialising the href attribute.
     * @param x The X coordinate of the use element.
     * @param y The Y coordinate of the use element.
     * @param width The width of the use element.
     * @param height The height of the use element.
     */
    public UseElement(String href, String x, String y, String width, String height) {
        super();
        if (href != null)
            _href = href;
        else
            _href = "";
        _x = x;
        _y = y;
        _width = width;
        _height = height;
    }

    /**
     * Constructs and initialises a UseElement with the specified href,
     * location and dimension attribute values.
     * @param href The value to use when initialising the href attribute.
     * @param x The X coordinate of the use element.
     * @param y The Y coordinate of the use element.
     * @param width The width of the use element.
     * @param height The height of the use element.
     */
    public UseElement(String href, float x, float y, float width, float height) {
        super();
        if (href != null)
            _href = href;
        else
            _href = "";
        _x = Float.toString(x);
        _y = Float.toString(y);
        _width = Float.toString(width);
        _height = Float.toString(height);
    }

    /**
     * Sets the desc element.
     * @param desc The value to use when setting the desc element.
     */
    public void setDesc(DescElement desc) {
        _desc = desc;
    }

    /**
     * Sets the title element.
     * @param title The value to use when setting the title element.
     */
    public void setTitle(TitleElement title) {
        _title = title;
    }

    /**
     * Returns the desc element.
     * @return The desc element.
     */
    public DescElement getDesc() {
        return _desc;
    }

    /**
     * Returns the title element.
     * @return The title element.
     */
    public TitleElement getTitle() {
        return _title;
    }

    /**
     * Returns the href attribute value.
     * @return The href attribute value.
     */
    public String getHref() {
        return _href;
    }

    /**
     * Returns the x attribute value.
     * @return The x attribute value.
     */
    public String getX() {
        return _x;
    }

    /**
     * Returns the y attribute value.
     * @return The y attribute value.
     */
    public String getY() {
        return _y;
    }

    /**
     * Returns the width attribute value.
     * @return The width attribute value.
     */
    public String getWidth() {
        return _width;
    }

    /**
     * Returns the height attribute value.
     * @return The height attribute value.
     */
    public String getHeight() {
        return _height;
    }

    /**
     * Sets the href attribute.
     * @param href The value to use when setting the href attribute.
     */
    public void setHref(String href) {
        if (href != null)
            _href = href;
        else
            _href = "";
    }

    /**
     * Sets the location of the use element to the specified location.
     * @param x The value to use when setting the X coordinate.
     * @param y The value to use when setting the Y coordinate.
     */
    public void setLocation(float x, float y) {
       _x = Float.toString(x);
       _y = Float.toString(y);
    }

    /**
     * Sets the dimensions of the use element to the spcified values.
     * @param x The value to use when setting the X coordinate.
     * @param y The value to use when setting the Y coordinate.
     */
    public void setDimensions(float width, float height) {
        _width = Float.toString(width);
        _height = Float.toString(height);
    }

    /**
     * Returns the Graphic object that represents this use element.
     * @return The Graphic object that represents this use element.
     */
    public Graphic getGraphic() {

        setStyleProperties();

        String href = getHref();
        SvgGraphic svgGraphic = null;
        if (href.length() == 0)
            return null;
        else {
            int index = href.indexOf('#');
            if (index != -1) {
                String id = href.substring(index+1);
                id = id.trim();

                // find defs element that has an element with this id
                SvgGraphic parent = getParent();
                while (parent != null) {
                    // does parent have a defs element
                    DefsElement defs = null;
                    if (parent instanceof SvgElement) {
                         defs = ((SvgElement)parent).getDefs();
                    } else if (parent instanceof GElement) {
                        defs = ((GElement)parent).getDefs();
                    } else if (parent instanceof SymbolElement) {
                        defs = ((SymbolElement)parent).getDefs();
                    }
                    if (defs != null) {
                        // found a defs element, see if it contains an element with
                        // the id we want
                        Definable ref = defs.getDefinedElement(id);
                        if (ref != null && ref instanceof SvgGraphic) {
                            svgGraphic = (SvgGraphic)ref;
                            break;
                        }
                    }
                    parent = parent.getParent();
                }

                if (svgGraphic != null) {

                    Graphic graphic;
                    if (svgGraphic instanceof TextElement && _style != null && _style.indexOf("font") != -1)
                        // if the use element has specified a font then use its rather than the text's one
                        graphic = ((TextElement)svgGraphic).getGraphic(_font);
                    else
                        graphic = svgGraphic.getGraphic();
                    
                    try {
                        if (_width != null && _height != null) {
                            graphic.scaleTo(Double.valueOf(_width).doubleValue(), Double.valueOf(_height).doubleValue());
                        }
                        if (_x != null && _y != null) {
                            graphic.translateTo(Double.valueOf(_x).doubleValue(), Double.valueOf(_y).doubleValue());
                        }
                    } catch (NumberFormatException e) {
                        System.out.println("Use element " + id + " has bad location or dimension");
                    }

                    // get transform matrix and apply to graphic
                    AffineTransform transformMatrix = getTransformMatrix(getTransform());
                    if (transformMatrix != null)
                        graphic.transform(transformMatrix);
                    setGraphicStyle(graphic);
                    
                    return graphic;
                }
            }
        }
        return null;
    }



    // XmlElement stuff

    /**
     * Returns the tag that identifies a use element in an SVG/XML file.
     * @return The tag that identifies a use element in an SVG/XML file.
     */
    public String getTag() {
        return "use";
    }

    /**
     * Reads in UseElement data from the given XML Element.
     * @param elem The XML Element that contains data to be read.
     */
    public void read(Element elem) {

        // read the standard graphic attributes
        readStandardGraphicAttributes(elem);

        String att = elem.getAttribute("x");
        if (att.length() > 0)
            _x = att;

        att = elem.getAttribute("y");
        if (att.length() > 0)
            _y = att;

        att = elem.getAttribute("width");
        if (att.length() > 0)
            _width = att;

        att = elem.getAttribute("height");
        if (att.length() > 0)
            _height = att;

        _href = elem.getAttribute("xlink:href");

         // read each child
        NodeList childNodes = elem.getChildNodes();
        int numChildren = childNodes.getLength();
        for (int i = 0; i < numChildren; i++) {
            Node child = childNodes.item(i);
            if (child.getNodeType() == Node.ELEMENT_NODE) { // it is an element
                Element childElement = (Element)child;
                String childTag = childElement.getTagName();

                if (childTag.equals("desc")) {
                    _desc = new DescElement();
                    _desc.read(childElement);

                } else if (childTag.equals("title")) {
                    _title = new TitleElement();
                    _title.read(childElement);
                }
             }
         }
    }

    /**
     * Inserts data into the given Document.
     * The data will become a child of the given parent element.
     *
     * @param doc The Document to write to.
     * @param parent Indicates the parent element of the data to be written.
     */
    public void write(Document doc, Node parent) {

        // create new element for this use element
        Element use = doc.createElement(getTag());

        // set the standard graphic attributes
        setStandardGraphicAttributes(use);

         // set the attributes
        use.setAttribute("xlink:href", _href);
        if (_x != null)
            use.setAttribute("x", _x);
        if (_y != null)
            use.setAttribute("y", _y);
        if (_width != null)
            use.setAttribute("width", _width);
        if (_height != null)
            use.setAttribute("height", _height);

         if (_desc != null)
            _desc.write(doc, use);
        if (_title != null)
            _title.write(doc, use);

        // append new element to the parent element
        parent.appendChild(use);

    }

}
